proc initCanvas {} {
	global var
	
	set var(CONNECTED) {}
	set var(popup) [menu .popupMenu -tearoff 0 -disabledforeground red]
	set var(dialogState) initial
	set var(showconn) 1

	set var(areaX1) 0
	set var(areaY1) 0
	set var(areaX2) 0
	set var(areaY2) 0
}

proc resetCanvas {} {
	global var widget statusMsg

	$widget(canvas) delete all
	set var(CONNECTED) {}
	set var(dialogState) initial
	set var(dragged) ""
}

proc createCanvas {} {
	global widget

	# create widget
	set w $widget(mainframe)
	set c $w.frame.c
	set widget(canvas) $c
	frame $w.frame
	pack $w.frame -side top -fill both -expand yes
	
	canvas $c -scrollregion {0c 0c 30c 24c} -width 15c -height 10c \
		-relief sunken -borderwidth 2 \
		-xscrollcommand "$w.frame.hscroll set" \
		-yscrollcommand "$w.frame.vscroll set"
	scrollbar $w.frame.vscroll -command "$c yview" -width 12
	scrollbar $w.frame.hscroll -orient horiz -command "$c xview" -width 12
	
	grid $c -in $w.frame \
	    -row 0 -column 0 -rowspan 1 -columnspan 1 -sticky news
	grid $w.frame.vscroll \
	    -row 0 -column 1 -rowspan 1 -columnspan 1 -sticky news
	grid $w.frame.hscroll \
	    -row 1 -column 0 -rowspan 1 -columnspan 1 -sticky news
	grid rowconfig    $w.frame 0 -weight 1 -minsize 0
	grid columnconfig $w.frame 0 -weight 1 -minsize 0

	# Set up event bindings for canvas:
	$c bind item <Any-Enter> "itemEnter $c"
	$c bind item <Any-Leave> "itemLeave $c"
	$c bind connection <Any-Enter> "itemEnter $c"
	$c bind connection <Any-Leave> "itemLeave $c"
	#bind $c <2> "resetDialogState $c"
	#bind $c <2> "$c scan mark %x %y"
	#bind $c <B2-Motion> "$c scan dragto %x %y"
	#bind $c <3> "itemMark $c %x %y"
	bind $c <2> "itemMark $c %x %y"
	bind $c <B2-Motion> "itemStroke $c %x %y"
	#bind $c <Control-f> "itemsUnderArea $c"
	bind $c <1> "itemStartDrag $c %x %y"
	bind $c <B1-Motion> "itemDrag $c %x %y"
	bind $c <ButtonRelease-1> "itemStopDrag $c"
	
	# Arrange for the menu to pop up when the label is clicked
	bind $c <3> "createContextMenu $c %X %Y"
	bind $c <Double-1> "jumpToEditProperties $c %X %Y"
}

# calc unique tag
proc uniqTag {tag} {
	if {[string range $tag 0 4]!="uniq:"} {
		return uniq:$tag
	} else {
		return $tag
	}
}

# calc bbox tag
proc bboxTag {tag} {
	if {[string range $tag 0 4]!="bbox:"} {
		return bbox:$tag
	} else {
		return $tag
	}
}

# calc name tag
proc nameTag {tag} {
	if {[string range $tag 0 4]=="uniq:"} {
		return [string range $tag 5 end]
	} else {
		return $tag
	}
}

# proc to create one rectangle
proc createRectangle {c xpos ypos xsize ysize color font fontcolor name text endstate vxmlcode endstatesize} {
	global var

#debug "$xpos $ypos $xsize $ysize $color $name $font $fontcolor $text"

	$c create rectangle \
		$xpos \
		$ypos \
		[expr $xpos+$xsize-1] \
		[expr $ypos+$ysize-1] \
		-fill $color \
		-tags [list item uniq:$name]

	if {$endstate} {
		$c create polygon \
			[expr $xpos+$xsize-1] [expr $ypos+$ysize-1] \
			[expr $xpos+$xsize-$endstatesize] [expr $ypos+$ysize-1] \
			[expr $xpos+$xsize-1] [expr $ypos+$ysize-$endstatesize] \
			[expr $xpos+$xsize-1] [expr $ypos+$ysize-1] \
			-fill black \
			-tags [list item uniq:$name]
	}

	if {$vxmlcode} {
		$c create polygon \
			$xpos [expr $ypos+$ysize-1] \
			[expr $xpos+$endstatesize] [expr $ypos+$ysize-1] \
			$xpos [expr $ypos+$ysize-$endstatesize] \
			$xpos [expr $ypos+$ysize-1] \
			-fill blue \
			-tags [list item uniq:$name]
	}

	$c create text \
		[expr $xpos+$xsize/2] \
		[expr $ypos+$ysize/2] \
		-text $text \
		-fill $fontcolor \
		-font $font \
		-anchor center -justify center \
		-tags [list item uniq:$name]
}

proc createBox {c xpos ypos color name htext btext endstate vxmlcode {factor 1}} {
	# move coordinates to grid
	set xpos [expr $xpos/[ini::Str Graphics.gridX]*[ini::Str Graphics.gridX]]
	set ypos [expr $ypos/[ini::Str Graphics.gridY]*[ini::Str Graphics.gridY]]
	
	# draw header and box
	createRectangle $c [expr round($factor*$xpos)] [expr round($factor*$ypos)] \
		[expr round($factor*[ini::Str Graphics.boxSizeX])] \
		[expr round($factor*[ini::Str Graphics.headerBoxSizeY])] \
		[ini::Str Graphics.colorHeader] \
		[ini::Str Graphics.fontHeader] \
		[ini::Str Graphics.colorHeaderFont] \
		$name $htext \
		0 0 0
	createRectangle $c \
		[expr round($factor*$xpos)] \
		[expr round($factor*([ini::Str Graphics.headerBoxSizeY]+$ypos))] \
		[expr round($factor*[ini::Str Graphics.boxSizeX])] \
		[expr round($factor*[ini::Str Graphics.boxSizeY])] \
		$color \
		[ini::Str Graphics.fontBody] \
		[ini::Str Graphics.colorBodyFont] \
		$name $btext \
		$endstate $vxmlcode [expr round(9*$factor)]
	
	# draw a surrounding box for coordinate calculation
	$c create rectangle \
		[expr round($factor*$xpos)] \
		[expr round($factor*$ypos)] \
		[expr round($factor*($xpos+[ini::Str Graphics.boxSizeX])-1)] \
		[expr round($factor*($ypos+[ini::Str Graphics.boxSizeY]+[ini::Str Graphics.headerBoxSizeY])-1)] \
		-outline black \
		-tags [list item uniq:$name bbox:$name]
}

proc lineColor {a b} {
	global plugin
	if {[util::value plugin(O,$a,$b)]!=""} {return [ini::Str Graphics.connColorAttributed]}
	return [ini::Str Graphics.connColorNoAttribute]
}

proc connectRectangles {c a b {arrow none} {color black} {width default} {factor 1}} {
	global var

	set conn {}

	if {$a!="" && $b!=""} {
		if {$var(showconn)} {
			# remove uniq: if necessary
			set a [nameTag $a]
			set b [nameTag $b]
			
			if {$a!=$b} {
				set ca [$c coords [bboxTag $a]]
				set cb [$c coords [bboxTag $b]]
	
				if {$ca=="" || $cb==""} {
					#debug "a=$a;b=$b"
				} else {
					#$c create rectangle $ca -outline yellow
					#$c create rectangle $cb -outline yellow
				}
			
				# give names
				set ax1 [lindex $ca 0]
				set ay1 [lindex $ca 1]
				set ax2 [lindex $ca 2]
				set ay2 [lindex $ca 3]
				
				set bx1 [lindex $cb 0]
				set by1 [lindex $cb 1]
				set bx2 [lindex $cb 2]
				set by2 [lindex $cb 3]
			
				# calculate the middle coordinates
				set mxa [expr round(($ax1+$ax2)/2)]
				set mya [expr round(($ay1+$ay2)/2)]
				set mxb [expr round(($bx1+$bx2)/2)]
				set myb [expr round(($by1+$by2)/2)]
			
				set val1 [expr $mya+($mxa-$mxb)]
				set val2 [expr $mya-($mxa-$mxb)]
			
				# determine which connection point to use
				if {$myb<=$val1 && $myb>=$val2} {
					if {[ini::Str Graphics.connType]=="direct"} {
						set conn [list $ax1 $mya $bx2 $myb]
					} else {
						set tmpx [expr ([ini::Str Graphics.connFactX1]*$ax1+[ini::Str Graphics.connFactX2]*$bx2)/([ini::Str Graphics.connFactX1]+[ini::Str Graphics.connFactX2])]
						set conn [list $ax1 $mya $tmpx $mya $tmpx $myb $bx2 $myb]
					}
				} elseif {$myb<=$val1 && $myb<=$val2} {
					if {[ini::Str Graphics.connType]=="direct"} {
						set conn [list $mxa $ay1 $mxb $by2]
					} else {
						set tmpy [expr ([ini::Str Graphics.connFactY1]*$ay1+[ini::Str Graphics.connFactY2]*$by2)/([ini::Str Graphics.connFactY1]+[ini::Str Graphics.connFactY2])]
						set conn [list $mxa $ay1 $mxa $tmpy $mxb $tmpy $mxb $by2]
					}
				} elseif {$myb>=$val1 && $myb>=$val2} {
					if {[ini::Str Graphics.connType]=="direct"} {
						set conn [list $mxa $ay2 $mxb $by1]
					} else {
						set tmpy [expr ([ini::Str Graphics.connFactY1]*$ay2+[ini::Str Graphics.connFactY2]*$by1)/([ini::Str Graphics.connFactY1]+[ini::Str Graphics.connFactY2])]
						set conn [list $mxa $ay2 $mxa $tmpy $mxb $tmpy $mxb $by1]
					}
				} else {
					if {[ini::Str Graphics.connType]=="direct"} {
						set conn [list $ax2 $mya $bx1 $myb]
					} else {
						set tmpx [expr ([ini::Str Graphics.connFactX1]*$ax2+[ini::Str Graphics.connFactX2]*$bx1)/([ini::Str Graphics.connFactX1]+[ini::Str Graphics.connFactX2])]
						set conn [list $ax2 $mya $tmpx $mya $tmpx $myb $bx1 $myb]
					}
				}
			} else {
				# connect a box with itself
				set ca [$c coords [uniqTag $a]]
				set ax1 [lindex $ca 0]
				set ay1 [lindex $ca 1]
				set ax2 [lindex $ca 2]
				set ay2 [lindex $ca 3]
				
				set tmpx [expr ([ini::Str Graphics.selfConnFactX1]*$ax1+[ini::Str Graphics.selfConnFactX2]*$ax2)/([ini::Str Graphics.selfConnFactX1]+[ini::Str Graphics.selfConnFactX2])]
				set tmpy [expr ([ini::Str Graphics.selfConnFactY1]*$ay1+[ini::Str Graphics.selfConnFactY2]*$ay2)/([ini::Str Graphics.selfConnFactY1]+[ini::Str Graphics.selfConnFactY2])]
				
				set conn [list $tmpx $ay1 \
											 $tmpx [expr $ay1-round($factor*[ini::Str Graphics.selfConnDistY])]\
											 [expr $ax2+round($factor*[ini::Str Graphics.selfConnDistX])] [expr $ay1-round($factor*[ini::Str Graphics.selfConnDistY])] \
											 [expr $ax2+round($factor*[ini::Str Graphics.selfConnDistX])] $tmpy \
											 $ax2 $tmpy]
			}

			# determine width
			if {$width=="default"} {set width [ini::Str Graphics.connWidth]}

			$c create line $conn -arrow $arrow -tags [list connection connection:$a:$b] \
				-joinstyle round -smooth [ini::Str Graphics.connSmooth] \
				-width $width -fill $color
		}
		lappend var(CONNECTED) [list $c $a $b $arrow]
		set var(CONNECTED) [lsort -uniq $var(CONNECTED)]
	}
}

proc restoreConnections {c items} {
	global var

	foreach conn $var(CONNECTED) {
		if {[lindex $conn 0]==$c} {
			set ra [lindex $conn 1]
			set rb [lindex $conn 2]

			if {[lsearch -exact $items $ra]!=-1 || [lsearch -exact $items $rb]!=-1} {
				$c delete connection:$ra:$rb
				connectRectangles $c $ra $rb [lindex $conn 3] [lineColor $ra $rb]
			}
		}
	}
}

proc createConnectionTo {c from} {
	global var statusMsg
	set var(dialogState) create_connect
	set var(dialogState,from) $from
	set var(dialogState,c) $c
	set statusMsg [res::Str status_createconnect]
}

proc removeConnectionTo {c from} {
	global var statusMsg
	set var(dialogState) remove_connect
	set var(dialogState,from) $from
	set var(dialogState,c) $c
	set statusMsg [res::Str status_removeconnect]
}

proc editProperties {c tag} {
	global widget var
	
	resetDialogState $c
	valList2Curobj $tag
	if {[isPrompt $tag]} {
		set w "property,p"
	} elseif {[isConcept $tag]} {
		set w "property,c"
	} elseif {[isSampleCall $tag]} {
		move::init "s_call"
		createSampleCallWidgets $tag $widget(s_call)
		set w "property,s"
	} else {
		set w "property,q"
	}
	::vTcl:Toplevel:WidgetProc $widget($w) startmodal
}

proc editPropertiesFromList {{id ""}} {
	global widget var

	if {$id==""} {
		set tmp [$widget(textlist) curselection]
		set choice {}
		foreach c $tmp {
			lappend choice [string trim [lindex [split [$widget(textlist) get $c] "|"] 0]]
		}
	} else {
		set choice [list $id]
	}
		
	foreach tag $choice {
		if {[string first ">" $tag]==-1} {
			valList2Curobj $tag
		} else {
			set tag [split $tag ">"]
			connection2Curobj [string trim [lindex $tag 0]] [string trim [lindex $tag 1]]
		}

		if {[isPrompt $tag]} {
			set w "property,p"
		} elseif {[isConcept $tag]} {
			set w "property,c"
		} elseif {[isQuestion $tag]} {
			set w "property,q"
		} elseif {[isSampleCall $tag]} {
			move::init "s_call"
			createSampleCallWidgets $tag $widget(s_call)
			set w "property,s"
		} else {
			set w "property,o"
		}
		::vTcl:Toplevel:WidgetProc $widget($w) startmodal
	}
}

proc editConnectionProperties {c tag_from tag_to} {
	global widget plugin

	resetDialogState $c
	connection2Curobj $tag_from $tag_to
	::vTcl:Toplevel:WidgetProc $widget(property,o) startmodal

	# check if the other direction exists as well
	if {$tag_to!=$tag_from && [isConnected $tag_to $tag_from]} {
		connection2Curobj $tag_to $tag_from
		::vTcl:Toplevel:WidgetProc $widget(property,o) startmodal
	}
}

proc addConceptToQuestion {tag} {
	global var plugin

	set redo_cmd "set plugin($tag,q_concepts) [list [concat $plugin($tag,q_concepts) $var(newid)]]"
	set undo_cmd "set plugin($tag,q_concepts) [list $plugin($tag,q_concepts)]"

	move::add $redo_cmd $undo_cmd [res::Str undo_add_concept $var(newid) $tag]
	prjFlags "-" 1 "-"
}

proc scExportAudio {tag} {
  global statusMsg widget plugin

	set audio [createSoundObjectFromSampleCall $tag 0]

	if {$audio!=""} {
		if {[ini::Str Files.playAudio]==1} {$audio play}

	  set ft [list \
	  	[list [ini::Str files.audType] [ini::Str files.audExt]] \
	  	[list [ini::Str files.allType] [ini::Str files.allExt]] \
	  ]
	  set fn "$tag [correctFN $plugin($tag,s_name)][ini::Str files.audExt]"
	  set fn [tk_getSaveFile \
	    -filetypes $ft \
	    -initialfile $fn \
	    -defaultextension [ini::Str files.audExt] \
	    -initialdir [ini::Str files.expDir]]
	
	  if {$fn!={}} {
	    if {[catch {$audio write $fn} statusMsg]} {
				tk_messageBox -message $statusMsg -type ok
	    } else {
	      log [res::Str log_export "Audio from sample call $tag" $fn]
	    }
	  }

		$audio destroy
	}
}

proc createPcontextMenu {c tag graphical x y} {
	global plugin var

	$var(popup) delete 0 end

	catch {eval destroy [winfo children $var(popup)]}

	set conn_menu [menu $var(popup).connections -tearoff 0]
	set samp_menu [menu $var(popup).samplecalls -tearoff 0]
	set fill_menu [menu $var(popup).fillfields -tearoff 0]

	$var(popup) add command -label $tag -state disabled

	$var(popup) add command -label [res::Str item_popup_clone] -command "copyObj $tag"
	$var(popup) add command -label [res::Str item_popup_delete] -command "deleteObj $tag"
	$var(popup) add separator

	$var(popup) add cascade -label [res::Str item_popup_connections] -menu $conn_menu
	if {$plugin($tag,p_follows)!={}} {
		foreach co $plugin($tag,p_follows) {
			$conn_menu add command -label "$co - [util::oneLine [util::value plugin(O,$tag,$co)] [ini::Str Graphics.connectionDisplayLength]]" -command "editConnectionProperties [list $c] $tag $co"
		}
		$conn_menu add separator
	}

	if {$graphical} {
		$conn_menu add command -label [res::Str item_popup_connect_to] -command "createConnectionTo [list $c] $tag"
		$conn_menu add command -label [res::Str item_popup_remove_conn_to] -command "removeConnectionTo [list $c] $tag"
	}
	$conn_menu add command -label [res::Str item_popup_remove_conn_all] -command "deleteAllConnections $tag"
	$conn_menu add command -label [res::Str item_popup_remove_conn_in] -command "deleteAllIncomingConnections $tag"
	$conn_menu add command -label [res::Str item_popup_remove_conn_out] -command "deleteAllOutgoingConnections $tag"

	$var(popup) add cascade -label [res::Str item_popup_samplecalls] -menu $samp_menu
	foreach co [usedInSampleCalls $tag] {
		$samp_menu add command -label "$co - [util::value plugin($co,s_name) [ini::Str Graphics.sampleCallDisplayLength]]" -command "editProperties [list $c] $co"
	}
	$var(popup) add cascade -label [res::Str item_popup_fill_fields] -menu $fill_menu
	foreach co [allEntries P] {
		if {$co!=$tag} {
			$fill_menu add command -label "$co - [util::oneLine [util::value plugin($co,p_title)] [ini::Str Graphics.connectionDisplayLength]]" -command "fillEmptyFields $tag $co"
		}
	}
	$var(popup) add separator

	$var(popup) add command -label [res::Str item_popup_properties] -command "editProperties [list $c] $tag"
	tk_popup $var(popup) $x $y
}

proc createQcontextMenu {c tag graphical x y} {
	global plugin var

	$var(popup) delete 0 end

	catch {eval destroy [winfo children $var(popup)]}

	set conn_menu [menu $var(popup).connections -tearoff 0]
	set conc_menu [menu $var(popup).concepts -tearoff 0]
	set samp_menu [menu $var(popup).samplecalls -tearoff 0]
	set fill_menu [menu $var(popup).fillfields -tearoff 0]

	$var(popup) add command -label $tag -state disabled

	$var(popup) add command -label [res::Str item_popup_clone] -command "copyObj $tag"
	$var(popup) add command -label [res::Str item_popup_delete] -command "deleteObj $tag"
	$var(popup) add separator

	$var(popup) add cascade -label [res::Str item_popup_connections] -menu $conn_menu
	if {$plugin($tag,q_follows)!={}} {
		foreach co $plugin($tag,q_follows) {
			$conn_menu add command -label "$co - [util::oneLine [util::value plugin(O,$tag,$co)] [ini::Str Graphics.connectionDisplayLength]]" -command "editConnectionProperties [list $c] $tag $co"
		}
		$conn_menu add separator
	}
	if {$graphical} {
		$conn_menu add command -label [res::Str item_popup_connect_to] -command "createConnectionTo [list $c] $tag"
		$conn_menu add command -label [res::Str item_popup_remove_conn_to] -command "removeConnectionTo [list $c] $tag"
	}
	$conn_menu add command -label [res::Str item_popup_remove_conn_all] -command "deleteAllConnections $tag"
	$conn_menu add command -label [res::Str item_popup_remove_conn_in] -command "deleteAllIncomingConnections $tag"
	$conn_menu add command -label [res::Str item_popup_remove_conn_out] -command "deleteAllOutgoingConnections $tag"

	$var(popup) add cascade -label [res::Str item_popup_concepts] -menu $conc_menu
	foreach co $plugin($tag,q_concepts) {
		$conc_menu add command -label "$co - [util::value plugin($co,c_title) [ini::Str Graphics.conceptDisplayLength]]" -command "editProperties [list $c] $co"
	}
	$conc_menu add separator
	$conc_menu add command -label [res::Str item_popup_add_concept] -command "menuCommand edit,concept; addConceptToQuestion $tag"

	$var(popup) add cascade -label [res::Str item_popup_samplecalls] -menu $samp_menu
	foreach co [usedInSampleCalls $tag] {
		$samp_menu add command -label "$co - [util::value plugin($co,s_name) [ini::Str Graphics.sampleCallDisplayLength]]" -command "editProperties [list $c] $co"
	}
	$var(popup) add cascade -label [res::Str item_popup_fill_fields] -menu $fill_menu
	foreach co [allEntries Q] {
		if {$co!=$tag} {
			$fill_menu add command -label "$co - [util::oneLine [util::value plugin($co,q_title)] [ini::Str Graphics.connectionDisplayLength]]" -command "fillEmptyFields $tag $co"
		}
	}
	$var(popup) add separator
	
	$var(popup) add command -label [res::Str item_popup_properties] -command "editProperties [list $c] $tag"
	tk_popup $var(popup) $x $y
}

proc createCcontextMenu {c tag graphical x y} {
	global plugin var

	$var(popup) delete 0 end

	catch {eval destroy [winfo children $var(popup)]}

	set used_menu [menu $var(popup).concepts -tearoff 0]
	set fill_menu [menu $var(popup).fillfields -tearoff 0]

	$var(popup) add command -label $tag -state disabled

	$var(popup) add command -label [res::Str item_popup_clone] -command "copyObj $tag"
	$var(popup) add command -label [res::Str item_popup_delete] -command "deleteObj $tag"
	$var(popup) add separator

	$var(popup) add cascade -label [res::Str item_popup_used_concepts] -menu $used_menu
	foreach co [calcUsedBy $tag] {
		$used_menu add command -label "$co - $plugin($co,q_title)" -command "editProperties [list $c] $co"
	}
	$var(popup) add cascade -label [res::Str item_popup_fill_fields] -menu $fill_menu
	foreach co [allEntries C] {
		if {$co!=$tag} {
			$fill_menu add command -label "$co - [util::oneLine [util::value plugin($co,c_title)] [ini::Str Graphics.connectionDisplayLength]]" -command "fillEmptyFields $tag $co"
		}
	}
	$var(popup) add separator

	$var(popup) add command -label [res::Str item_popup_properties] -command "editProperties [list $c] $tag"
	tk_popup $var(popup) $x $y
}

proc createScontextMenu {c tag graphical x y} {
	global plugin var

	$var(popup) delete 0 end

	catch {eval destroy [winfo children $var(popup)]}

	set uses_menu [menu $var(popup).concepts -tearoff 0]
	set fill_menu [menu $var(popup).fillfields -tearoff 0]

	$var(popup) add command -label $tag -state disabled

	$var(popup) add command -label [res::Str item_popup_clone] -command "copyObj $tag"
	$var(popup) add command -label [res::Str item_popup_delete] -command "deleteObj $tag"
	$var(popup) add command -label [res::Str item_popup_analyze] -command "scAnalyze $tag"
	$var(popup) add command -label [res::Str item_popup_sync] -command "scSync $tag"
	$var(popup) add command -label [res::Str item_popup_expaudio] -command "scExportAudio $tag"
	$var(popup) add separator

	$var(popup) add cascade -label [res::Str item_popup_scall_uses] -menu $uses_menu
	foreach co [sampleCallUses $tag] {
		if {[objectExists $co]} {
			if {[isPrompt $co]} {set tmp p_title} else {set tmp q_title}
			$uses_menu add command -label "$co - $plugin($co,$tmp)" -command "editProperties [list $c] $co"
		}
	}
	$var(popup) add cascade -label [res::Str item_popup_fill_fields] -menu $fill_menu
	foreach co [allEntries S] {
		if {$co!=$tag} {
			$fill_menu add command -label "$co - [util::oneLine [util::value plugin($co,s_name)] [ini::Str Graphics.connectionDisplayLength]]" -command "fillEmptyFields $tag $co"
		}
	}
	$var(popup) add separator

	$var(popup) add command -label [res::Str item_popup_properties] -command "editProperties [list $c] $tag"
	tk_popup $var(popup) $x $y
}

proc createOcontextMenu {c tag1 tag2 graphical x y} {
	global plugin var

	$var(popup) delete 0 end

	catch {eval destroy [winfo children $var(popup)]}

	set samp_menu [menu $var(popup).samplecalls -tearoff 0]

	$var(popup) add command -label "$tag1 > $tag2" -state disabled

	$var(popup) add command -label [res::Str conn_popup_delete] -command "deleteConnection $tag1 $tag2"
	$var(popup) add separator

	$var(popup) add cascade -label [res::Str item_popup_samplecalls] -menu $samp_menu
	set used [connUsedInSampleCalls $tag1 $tag2]
	if {[isConnected $tag2 $tag1]} {
		set used [lsort -uniq [concat $used [connUsedInSampleCalls $tag2 $tag1]]]
	}
	foreach co $used {
		$samp_menu add command -label "$co - [util::value plugin($co,s_name) [ini::Str Graphics.sampleCallDisplayLength]]" -command "editProperties [list $c] $co"
	}
	$var(popup) add separator
	
	$var(popup) add command -label [res::Str conn_popup_properties] -command "editConnectionProperties [list $c] $tag1 $tag2"
	tk_popup $var(popup) $x $y
}

proc createListContextMenu {x y {id ""}} {
	global widget var plugin

	if {$id==""} {
		set tmp [$widget(textlist) curselection]
		set choice {}
		foreach c $tmp {
			lappend choice [string trim [lindex [split [$widget(textlist) get $c] "|"] 0]]
		}
	} else {
		set choice [list $id]
	}

	foreach tag $choice {
		if {[string first ">" $tag]!=-1} {
			set tag [split $tag ">"]
			createOcontextMenu "" [string trim [lindex $tag 0]] [string trim [lindex $tag 1]] 0 [expr [winfo rootx $widget(textlist)]+$x] [expr [winfo rooty $widget(textlist)]+$y]
		} elseif {[isPrompt $tag]} {
			createPcontextMenu "" $tag 0 [expr [winfo rootx $widget(textlist)]+$x] [expr [winfo rooty $widget(textlist)]+$y]
		} elseif {[isConcept $tag]} {
			createCcontextMenu "" $tag 0 [expr [winfo rootx $widget(textlist)]+$x] [expr [winfo rooty $widget(textlist)]+$y]
		} elseif {[isQuestion $tag]} {
			createQcontextMenu "" $tag 0 [expr [winfo rootx $widget(textlist)]+$x] [expr [winfo rooty $widget(textlist)]+$y]
		} elseif {[isSampleCall $tag]} {
			createScontextMenu "" $tag 0 [expr [winfo rootx $widget(textlist)]+$x] [expr [winfo rooty $widget(textlist)]+$y]
		}
	}
}

proc createContextMenu {c x y} {
	global var plugin

	# prompt or question
	resetDialogState $c
	set tag [getCurrentNameTag $c]
	if {$tag!=""} {
		if {[isPrompt $tag]} {
			createPcontextMenu $c $tag 1 $x $y
		} elseif {[isQuestion $tag]} {
			createQcontextMenu $c $tag 1 $x $y
		}
		return
	}
	
	# connection
	set tag [getConnectionTags $c]
	if {$tag!={}} {
		createOcontextMenu $c [lindex $tag 0] [lindex $tag 1] 1 $x $y
		return
	}

	# right click on background	(no object)
	$var(popup) delete 0 end

	catch {destroy $var(popup).connections}
	catch {destroy $var(popup).concepts}
	catch {destroy $var(popup).delconcepts}
	catch {destroy $var(popup).copyconcepts}
	catch {destroy $var(popup).samplecalls}
	catch {destroy $var(popup).delscall}
	catch {destroy $var(popup).copyscall}

	set conc_menu [menu $var(popup).concepts -tearoff 0]
	set delconc_menu [menu $var(popup).delconcepts -tearoff 0]
	set copyconc_menu [menu $var(popup).copyconcepts -tearoff 0]
	set samp_menu [menu $var(popup).samplecalls -tearoff 0]
	set delscall_menu [menu $var(popup).delscall -tearoff 0]
	set copyscall_menu [menu $var(popup).copyscall -tearoff 0]

	$var(popup) add command -label [res::Str none_popup_refresh] -command "menuCommand edit,refresh"
	$var(popup) add separator

	$var(popup) add cascade -label [res::Str none_popup_concept_clone] -menu $copyconc_menu
	foreach co [allEntries C] {
		$copyconc_menu add command -label "$co - [util::value plugin($co,c_title) [ini::Str Graphics.conceptDisplayLength]]" -command "copyObj $co"
	}
	$var(popup) add cascade -label [res::Str none_popup_concept_del] -menu $delconc_menu
	foreach co [allEntries C] {
		$delconc_menu add command -label "$co - [util::value plugin($co,c_title) [ini::Str Graphics.conceptDisplayLength]]" -command "deleteObj $co"
	}
	$var(popup) add cascade -label [res::Str none_popup_concept_prop] -menu $conc_menu
	foreach co [allEntries C] {
		$conc_menu add command -label "$co - [util::value plugin($co,c_title) [ini::Str Graphics.conceptDisplayLength]]" -command "editProperties $c $co"
	}
	$var(popup) add separator

	$var(popup) add cascade -label [res::Str none_popup_scall_clone] -menu $copyscall_menu
	foreach co [allEntries S] {
		$copyscall_menu add command -label "$co - [util::value plugin($co,s_name) [ini::Str Graphics.sampleCallDisplayLength]]" -command "copyObj $co"
	}
	$var(popup) add cascade -label [res::Str none_popup_scall_del] -menu $delscall_menu
	foreach co [allEntries S] {
		$delscall_menu add command -label "$co - [util::value plugin($co,s_name) [ini::Str Graphics.sampleCallDisplayLength]]" -command "deleteObj $co"
	}
	$var(popup) add cascade -label [res::Str none_popup_scall_prop] -menu $samp_menu
	foreach co [allEntries S] {
		$samp_menu add command -label "$co - [util::value plugin($co,s_name) [ini::Str Graphics.sampleCallDisplayLength]]" -command "editProperties $c $co"
	}

	tk_popup $var(popup) $x $y
}

proc jumpToEditProperties {c x y} {
	global var
	
	resetDialogState $c
	set tag [getCurrentNameTag $c]
	if {$tag!=""} {
		editProperties $c $tag
		return
	}
	set tag [getConnectionTags $c]
	if {$tag!=""} {
		editConnectionProperties $c [lindex $tag 0] [lindex $tag 1]
	}
}

# Utility procedures for highlighting the item under the pointer
proc itemEnter {c} {
  global var

  if {[winfo depth $c] == 1} {
		set var(restoreCmd) {}
		return
  }
  set type [$c type current]
  if {$type == "window"} {
		set var(restoreCmd) {}
		return
  }
	if {$type=="line"} {
		$c raise current
	}
  if {$type == "bitmap"} {
		set bg [lindex [$c itemconf current -background] 4]
		set var(restoreCmd) [list $c itemconfig current -background $bg]
		$c itemconfig current -background SteelBlue2
		return
  }
  set fill [lindex [$c itemconfig current -fill] 4]
  if {(($type == "rectangle") || ($type == "oval") || ($type == "arc"))
    && ($fill == "")} {
		set outline [lindex [$c itemconfig current -outline] 4]
		set var(restoreCmd) "$c itemconfig current -outline $outline"
		$c itemconfig current -outline SteelBlue2
  } else {
		set var(restoreCmd) "$c itemconfig current -fill $fill"
		$c itemconfig current -fill SteelBlue2
  }
}

proc itemLeave {c} {
    global var plugin

    eval $var(restoreCmd)
}

proc itemStopDrag {c} {
	global var plugin
	
	if {$var(dragged)!=""} {
		set ut [uniqTag $var(dragged)]
		set ca [$c coords $ut]

		if {$plugin($var(dragged),x)!=[lindex $ca 0] ||
				$plugin($var(dragged),y)!=[lindex $ca 1]} {
			set undo_cmd "set plugin($var(dragged),x) $plugin($var(dragged),x);
									  set plugin($var(dragged),y) $plugin($var(dragged),y);"
			set redo_cmd "set plugin($var(dragged),x) [lindex $ca 0];
										set plugin($var(dragged),y) [lindex $ca 1];"

			set moved $var(dragged)
			
			# check if there's a dragging area
			set area [$c bbox area]
			if {$area!={}} {
				# find the touched items
				set tmp [$c find overlapping [lindex $area 0] [lindex $area 1] \
																		 [lindex $area 2] [lindex $area 3]]
				set touched {}
				foreach x $tmp {
					set tags [$c gettags $x]
					if {[lsearch -exact $tags "item"]!=-1 && \
					    [lsearch -exact $tags $ut]==-1} {
						lappend touched $x
					}
				}

				# move the touched items
				set xmove [expr [lindex $ca 0]-$plugin($var(dragged),x)]
				set ymove [expr [lindex $ca 1]-$plugin($var(dragged),y)]
				foreach x $touched {
					$c move $x $xmove $ymove
					set tags [$c gettags $x]
					set p [lsearch -regexp $tags "^uniq:.+"]
					set tag [nameTag [lindex $tags $p]]
					lappend moved $tag
					append undo_cmd "set plugin($tag,x) $plugin($tag,x);
											  	 set plugin($tag,y) $plugin($tag,y);"
					append redo_cmd "set plugin($tag,x) [expr $plugin($tag,x)+$xmove];
													 set plugin($tag,y) [expr $plugin($tag,y)+$ymove];"
				}
			}
			move::add $redo_cmd $undo_cmd [res::Str undo_move [lsort -uniq $moved]]
			prjFlags "-" 1 "-"
			updateList
		}
		set var(dragged) ""
	}
}

# Utility procedures for stroking out a rectangle and printing what's
# underneath the rectangle's area.

proc itemMark {c x y} {
  global var
  set var(arrx1) [$c canvasx $x]
  set var(arry1) [$c canvasy $y]
  $c delete area
}

proc itemStroke {c x y} {
	global var
	set x [$c canvasx $x]
	set y [$c canvasy $y]
	if {($var(arrx1) != $x) && ($var(arry1) != $y)} {
		$c delete area
		$c addtag area withtag [$c create rect $var(arrx1) $var(arry1) $x $y -outline black]
		set var(arrx2) $x
		set var(arry2) $y
	}
}

proc itemsUnderArea {c} {
  global var
  set area [$c find withtag area]
  set items ""
  foreach i [$c find enclosed $var(areaX1) $var(areaY1) $var(areaX2) $var(areaY2)] {
		if {[lsearch [$c gettags $i] item] != -1} {
	    lappend items $i
		}
  }
  puts stdout "Items enclosed by area: $items"
  set items ""
  foreach i [$c find overlapping $var(areaX1) $var(areaY1) $var(areaX2) $var(areaY2)] {
		if {[lsearch [$c gettags $i] item] != -1} {
		    lappend items $i
		}
  }
  puts stdout "Items overlapping area: $items"
}

# Utility procedures to support dragging of items.

proc resetDialogState {c} {
	global var statusMsg
	
	if {$var(dialogState)!="initial"} {
		set var(dialogState) "initial"
		set statusMsg [res::Str status_opcancelled]
	}
	
	if {$c!=""} {$c delete area}
}

proc itemStartDrag {c x y} {
  global var plugin statusMsg

	if {$var(dialogState)=="create_connect"} {
		set tag [getCurrentNameTag $c]
		connectRectangles $c $var(dialogState,from) $tag last [lineColor $var(dialogState,from) $tag]
		set var(dialogState) initial

		if {[isPrompt $var(dialogState,from)]} {
			set redo_cmd "set plugin($var(dialogState,from),p_follows) \
				[list [lsort -uniq [concat $plugin($var(dialogState,from),p_follows) $tag]]]"
			set undo_cmd "set plugin($var(dialogState,from),p_follows) \
				$plugin($var(dialogState,from),p_follows)"
			move::add $redo_cmd $undo_cmd [res::Str undo_add_conn $var(dialogState,from) $tag]
		} else {
			set redo_cmd "set plugin($var(dialogState,from),q_follows) \
				[list [lsort -uniq [concat $plugin($var(dialogState,from),q_follows) $tag]]]"
			set undo_cmd "set plugin($var(dialogState,from),q_follows) \
				$plugin($var(dialogState,from),q_follows)"
			move::add $redo_cmd $undo_cmd [res::Str undo_add_conn $var(dialogState,from) $tag]
		}

		set statusMsg ""
		set plugin(prj,changed) 1
		updateList
	} elseif {$var(dialogState)=="remove_connect"} {
		set tag [getCurrentNameTag $c]
		deleteConnection $var(dialogState,from) $tag
		updateList
	}

	set var(lastX) [expr round([$c canvasx $x]/[ini::Str Graphics.gridX])*[ini::Str Graphics.gridX]]
	set var(lastY) [expr round([$c canvasy $y]/[ini::Str Graphics.gridY])*[ini::Str Graphics.gridY]]
}

proc getCurrentNameTag {c} {
	set tag ""
	foreach t [$c gettags current] {
		if {[string range $t 0 4]=="uniq:"} {
			set tag [nameTag $t]
			break
		}
	}
	return $tag
}

proc getConnectionTags {c} {
	foreach t [$c gettags current] {
		if {[string range $t 0 10]=="connection:"} {
			set tag [string range $t 11 end]
			set p [string first ":" $tag]
			if {$p>-1} {
				return [list [string range $tag 0 [expr $p-1]] [string range $tag [expr $p+1] end]]
			}
		}
	}
	return {}
}

proc itemDrag {c x y} {
  global var plugin
  
  set x [$c canvasx $x]
  set y [$c canvasy $y]

	set x [expr round($x/[ini::Str Graphics.gridX])*[ini::Str Graphics.gridX]]
	set y [expr round($y/[ini::Str Graphics.gridY])*[ini::Str Graphics.gridY]]

	set tag [getCurrentNameTag $c]

	if {$tag!=""} {	
	  $c move [uniqTag $tag] [expr {$x-$var(lastX)}] [expr {$y-$var(lastY)}]
	  
	  set var(lastX) $x
	  set var(lastY) $y

		set var(dragged) $tag
		
		restoreConnections $c $tag
	}
}
