proc getChildren {id} {
	global plugin
	
	if {$id==""} {return [concat [allEntries P] [allEntries Q]]}
	if {[isPrompt $id]} {return $plugin($id,p_follows)}
	return $plugin($id,q_follows)
}

proc layoutInit {} {
  global VIEW
    
  array set VIEW {
    opt,overwriteExistingMinLevel	0
    opt,overwriteExistingSuccLevel	1
    opt,setSuccLevel1			0
		opt,center 1
    opt,dx				120
    opt,dy				80
  }
}

proc layout {getChildren} {
  # visualize a nearly-tree structure in an integer matrix.
  # getChildren is a proc invoked with an ID, returning a list of
  # children ID's (empty if there are none). When getChildren is
  # invoked with the empty string, it must return all elements that are needed
  # to reach each node
  global VIEW plugin

	array unset VIEW
	layoutInit

  # store the structural information inside local data structures
  eval {set examSet [$getChildren {}]}
  set examSet [lsort -uniq $examSet]
  set VIEW(nodes) {}

  while {$examSet!={}} {
    # get the first element and remove it
    set e [lindex $examSet 0]
    set examSet [lreplace $examSet 0 0]
    
    # get the children of that element and store them
    lappend VIEW(nodes) $e
    eval {set VIEW(succ,$e) [$getChildren $e]}
    if {![info exists VIEW(pred,$e)]} {set VIEW(pred,$e) {}}
    
    # examine the new element if not already done
    foreach n $VIEW(succ,$e) {
      if {[lsearch -exact $VIEW(nodes) $n]==-1} {
        lappend examSet $n
      }
      
      # store the inverse direction
      if {![info exists VIEW(pred,$n)]} {set VIEW(pred,$n) {}}
      lappend VIEW(pred,$n) $e
    }
  }
  
  # remove double nodes
  set VIEW(nodes) [lsort -uniq $VIEW(nodes)]
  set layoutNodes {}

  # compute the number of predecessors and successors of each node
  foreach e $VIEW(nodes) {
    set VIEW(pred,$e) [lsort -uniq $VIEW(pred,$e)]
    set VIEW(#pred,$e) [llength $VIEW(pred,$e)]
    set VIEW(succ,$e) [lsort -uniq $VIEW(succ,$e)]
    set VIEW(#succ,$e) [llength $VIEW(succ,$e)]
		
		# special case: standalone nodes
		if {$VIEW(#pred,$e)==0 && $VIEW(#succ,$e)==0} {
			set VIEW(level,$e) -1
		} else {
			lappend layoutNodes $e
		}
  }

  #debug "after reading: [array get VIEW]"

  # loop over all nodes to be examined
  set maxLevel 0
	set examSet $layoutNodes
  while {$examSet!={}} {
    # calc the minimum #pred of the nodes to be examined
    set minPred [llength $VIEW(nodes)]
    foreach e $examSet {
      if {$VIEW(#pred,$e)<$minPred} {
      	set minPred $VIEW(#pred,$e)
      }
    }

    # get all nodes with that minimum number
    set minSet {}
    foreach e $examSet {
      if {$VIEW(#pred,$e)==$minPred} {lappend minSet $e}
    }

    #debug "minSet is $minSet"

    # loop over all nodes in the min set
    set doneSet {}
    while {$minSet!={}} {
      # remove the node from the exam and the min set, include it in the done set
      set e [lindex $minSet 0]
      set p [lsearch -exact $examSet $e]
      set examSet [lreplace $examSet $p $p]
      set p [lsearch -exact $minSet $e]
      set minSet [lreplace $minSet $p $p]
      lappend doneSet $e

      if {[info exists VIEW(level,$e)]} {
        #debug "process $e (level $VIEW(level,$e))..."
      } else {
        #debug "process $e (no level yet)..."
      }
      
      # assign the node level 0, if it does not already have one
      if {$VIEW(opt,overwriteExistingMinLevel)||![info exists VIEW(level,$e)]} {
        set VIEW(level,$e) 0
	#puts "i level,$e <-- fix 0"
        if {![info exists VIEW(count,0)]} {set VIEW(count,0) 0}
        set VIEW(xlevel,$e) $VIEW(count,0)
        incr VIEW(count,0)
      }

      # apply level x+1 to any successor of $e and examine these successors next
      foreach n $VIEW(succ,$e) {
        if {[lsearch -exact $doneSet $n]==-1} {
#          lappend minSet $n
           set minSet [linsert $minSet 0 $n]
        }
        if {$VIEW(opt,overwriteExistingSuccLevel)||![info exists VIEW(level,$n)]} {
          if {$VIEW(opt,setSuccLevel1)} {
            set VIEW(level,$n) 1
            #puts "s level,$n <-- fix 1"
            if {![info exists VIEW(count,1)]} {set VIEW(count,1) 0}
            set VIEW(xlevel,$n) $VIEW(count,1)
            incr VIEW(count,1)
          } else {
            set VIEW(level,$n) [expr 1+$VIEW(level,$e)]
            #puts "s level,$n <-- $VIEW(level,$n)"
            if {![info exists VIEW(count,$VIEW(level,$n))]} {
              set VIEW(count,$VIEW(level,$n)) 0
            }
            set VIEW(xlevel,$n) $VIEW(count,$VIEW(level,$n))
            incr VIEW(count,$VIEW(level,$n))
          }
          set maxLevel [util::max $maxLevel $VIEW(level,$n)]
        }
      }
    }
  }
  
  #debug "after levelling: [array get VIEW level,*], maxLevel is $maxLevel"
  
  #debug "after layout for $VIEW(nodes): [array get VIEW xlevel,*]"

	# compact levels
	set diff_levels {}
  foreach e $layoutNodes {
		lappend diff_levels $VIEW(level,$e)
	}	
	set diff_levels [lsort -uniq $diff_levels]
  foreach e $layoutNodes {
		set VIEW(level,$e) [lsearch -exact $diff_levels $VIEW(level,$e)]
	}	

	# compact xlevels
	set maxx 0
	for {set yl 0} {$yl<[llength $diff_levels]} {incr yl} {
		set VIEW(diff_xlevels,$yl) {}
	  foreach e $layoutNodes {
			if {$VIEW(level,$e)==$yl} {
				lappend VIEW(diff_xlevels,$yl) $VIEW(xlevel,$e)
			}
		}	
		set VIEW(diff_xlevels,$yl) [lsort -uniq $VIEW(diff_xlevels,$yl)]
		if {[llength $VIEW(diff_xlevels,$yl)]>$maxx} {set maxx [llength $VIEW(diff_xlevels,$yl)]}
	  foreach e $layoutNodes {
			if {$VIEW(level,$e)==$yl} {
				set VIEW(xlevel,$e) [lsearch -exact $VIEW(diff_xlevels,$yl) $VIEW(xlevel,$e)]
			}
		}	
	}
  
  #debug "after compacting levels: [array get VIEW level,*]"
  
  #debug "after compacting xlevels: [array get VIEW xlevel,*]"

	if {$VIEW(opt,center)} {
		# center x for all levels
		for {set yl 0} {$yl<[llength $diff_levels]} {incr yl} {
			foreach e $layoutNodes {
				if {$VIEW(level,$e)==$yl} {
					#debug "maxx=$maxx; VIEW(diff_xlevels,$yl)=$VIEW(diff_xlevels,$yl);expr=[expr round(0.5*($maxx-[llength $VIEW(diff_xlevels,$yl)]))]"
					incr VIEW(xlevel,$e) [expr round(0.5*($maxx-[llength $VIEW(diff_xlevels,$yl)]))]
				}
			}
		}	
	}

  # calculate coordinates
	set redo_cmd ""
	set undo_cmd ""
	
	set y 0
	set changed 0
  foreach e $VIEW(nodes) {
		if {[util::value VIEW(level,$e)]!=-1} {
	    set nx [expr $VIEW(opt,dx)*$VIEW(xlevel,$e)]
	    set ny [expr $VIEW(opt,dy)*$VIEW(level,$e)]
		} else {
	    set nx [expr $VIEW(opt,dx)*$maxx]
	    set ny $y
	    incr y $VIEW(opt,dy)
		}
		if {$nx!=$plugin($e,x) || $ny!=$plugin($e,y)} {
			set changed 1
			append undo_cmd "set plugin($e,x) $plugin($e,x);
										   set plugin($e,y) $plugin($e,y);"
			append redo_cmd "set plugin($e,x) $nx;
											 set plugin($e,y) $ny;"
		}
  }

	if {$changed} {
		move::add $redo_cmd $undo_cmd [res::Str undo_layout]
		prjFlags "-" 1 "-"
	}
}
