proc BNF2SRGS {n str {silent 0}} {
	global plugin
	
	set tmp $str
  set sfn "[ini::Str files.GRMskelFileC]"
  if {[file exists $sfn]} {
		if {![catch {set fs [open $sfn]}]} {
		  # read the file and replace the id
      regsub -all "{CONCEPT_ID}" [read $fs] "$n" tmp
			close $fs
      regsub -all "{GRM_LANG}" $tmp $plugin(language) tmp

			# use the generated YETI parser to parse BNF and convert it to SRGS
			set scanner [GrmScanner #auto]
			set mp [GrmParser #auto -scanner $scanner]
	    $scanner start $str
	    $mp reset
			if {[catch {set tmpx [$mp parse]} msg]} {
				log "ERROR converting $n from BNF to SRGS: $msg"
				tk_messageBox -type ok -message "ERROR converting $n from BNF to SRGS

$msg"
				set tmpx $str
			} else {
				if {!$silent} {
					tk_messageBox -type ok -message "Grammar is OKAY!"
				}
			}
			set tmp [util::save_regsub_all "{GRM_ROOT_RULE}" $tmp $tmpx]
    } else {
			tk_messageBox -type ok -message "Grammar skeleton file '$sfn'
does not exist --> conversion is not possible!"
    }
  }
	return $tmp
}

# transfer values from plugin to current object (P Q C S)
proc valList2Curobj {tag} {
  global widget curobj plugin

  # nothing to do
  if {$tag==""} return
  set curobj(id) $tag

  #log "Transferring List --> Curobj for $tag: [array get curobj]"

  # copy values in respect to the entry type
  set prefix "[string tolower [string index $tag 0]]ix"
  foreach n [concat $plugin($prefix,vars) $plugin($prefix,text)] {
    set curobj($n) $plugin($tag,$n)
  }
  foreach n $plugin($prefix,text) {
    $widget($n) delete 0.0 end
    $widget($n) insert end $plugin($tag,$n)
  }
  
  # fill computed fields
  if {[isPrompt $tag]} {
  	set curobj(p_predecessors) [calcPredecessors $tag]
  } elseif {[isQuestion $tag]} {
  	set curobj(q_predecessors) [calcPredecessors $tag]
  } elseif {[isConcept $tag]} {
  	set curobj(c_usedby) [calcUsedBy $tag]
  }
}

# transfer values from current object to plugin (P Q C S)
# if checkOnly, 1 is returned as soon as the first change has been detected
proc valCurobj2List {{checkOnly 0}} {
  global widget curobj plugin

  if {[info exists curobj(id)]} {
    set id $curobj(id)
  } else {
    return
  }

  #log "Transferring Curobj --> List for $id: [array get curobj]"

	set undo_cmd ""
	set redo_cmd ""
	set changed 0

  # copy values in respect to the entry type
  set prefix "[string tolower [string index $id 0]]ix"
	set ix 0
  foreach n $plugin($prefix,vars) {
    if {$plugin($id,$n)!=$curobj($n)} {
			if {$checkOnly} {return 1}
			append undo_cmd "set plugin($id,$n) [list $plugin($id,$n)];"
			append redo_cmd "set plugin($id,$n) [list $curobj($n)];"
      set changed 1
    }
		incr ix
  }

  foreach n $plugin($prefix,text) {
    set tmp [string trim [$widget($n) get 0.0 end]]
    if {$plugin($id,$n)!=$tmp} {
			if {$checkOnly} {return 1}
			append undo_cmd "set plugin($id,$n) [list $plugin($id,$n)];"
			append redo_cmd "set plugin($id,$n) [list $tmp];"
      set changed 1
    }
  }

	if {$changed} {
		move::add $redo_cmd $undo_cmd [res::Str undo_change $id]
    set plugin(prj,changed) 1
	  updateList
	} else {
		return 0
	}
}

# transfer connection values from plugin to current object (O)
proc connection2Curobj {tag_from tag_to} {
  global widget curobj plugin

  # nothing to do
  if {$tag_from=="" || $tag_to==""} return
  set curobj(id) "$tag_from > $tag_to"
	set curobj(tag) "O,$tag_from,$tag_to"

  # copy value
	$widget(o_condition) delete 0.0 end
	if {[info exists plugin(O,$tag_from,$tag_to)]} {
		$widget(o_condition) insert end $plugin(O,$tag_from,$tag_to)
	}
}

# transfer values from current object to plugin (O)
# if checkOnly, 1 is returned as soon as the first change has been detected
proc curobj2Connection {{checkOnly 0}} {
  global widget curobj plugin

  if {[info exists curobj(tag)]} {
    set tag $curobj(tag)
  } else {
    return
  }

	set changed 0
	set undo_cmd ""
	set redo_cmd ""
	
  # copy value
  set tmp [string trim [$widget(o_condition) get 0.0 end]]
	if {![info exists plugin($tag)]} {set plugin($tag) ""}
  if {$plugin($tag)!=$tmp} {
		if {$checkOnly} {return 1}
		append undo_cmd "set plugin($tag) [list $plugin($tag)];"
		append redo_cmd "set plugin($tag) [list $tmp];"
    set changed 1
  }

  if {$changed} {
		move::add $redo_cmd $undo_cmd [res::Str undo_change [string range $tag 2 end]]
    set plugin(prj,changed) 1
	  updateList
  }
  
  return 0
}

proc deleteConnection {from to} {
	global plugin

	if {[info exists plugin(O,$from,$to)]} {
		set undo_cmd "set plugin(O,$from,$to) [list $plugin(O,$from,$to)];"
		set redo_cmd "array unset plugin O,$from,$to;"
	} else {
		set undo_cmd ""
		set redo_cmd ""
	}

	if {[isPrompt $from]} {
		set p [lsearch $plugin($from,p_follows) $to]
		if {$p>-1} {
			append undo_cmd "set plugin($from,p_follows) [list $plugin($from,p_follows)];"
			append redo_cmd "set plugin($from,p_follows) [list [lreplace $plugin($from,p_follows) $p $p]];"
		}
	}	
	if {[isQuestion $from]} {
		set p [lsearch $plugin($from,q_follows) $to]
		if {$p>-1} {
			append undo_cmd "set plugin($from,q_follows) [list $plugin($from,q_follows)];"
			append redo_cmd "set plugin($from,q_follows) [list [lreplace $plugin($from,q_follows) $p $p]];"
		}
	}	

	if {$redo_cmd!=""} {
		move::add $redo_cmd $undo_cmd [res::Str undo_del_conn $from $to]
	  set plugin(prj,changed) 1
		updateList
	}
}

proc deleteAllIncomingConnections {from} {
	global plugin

	set undo_cmd ""
	set redo_cmd ""

	# delete connection properties
	foreach n [array names O,*,$from] {
		append redo_cmd "array unset plugin $n;"
		append undo_cmd "set plugin($n) [list $plugin($n)];"
	}

	# remove $from from all follow ups of other objects
	foreach n [allEntries P] {
		set p [lsearch $plugin($n,p_follows) $from]
		if {$p>-1} {
			append undo_cmd "set plugin($n,p_follows) [list $plugin($n,p_follows)];"
			append redo_cmd "set plugin($n,p_follows) [list [lreplace $plugin($n,p_follows) $p $p]];"
		}
	}	
	foreach n [allEntries Q] {
		set p [lsearch $plugin($n,q_follows) $from]
		if {$p>-1} {
			append undo_cmd "set plugin($n,q_follows) [list $plugin($n,q_follows)];"
			append redo_cmd "set plugin($n,q_follows) [list [lreplace $plugin($n,q_follows) $p $p]];"
		}
	}	

	if {$redo_cmd!=""} {
		move::add $redo_cmd $undo_cmd [res::Str undo_del_in_conn $from]
	  set plugin(prj,changed) 1
		updateList
	}
}

proc deleteAllOutgoingConnections {from} {
	global plugin

	set undo_cmd ""
	set redo_cmd ""

	# delete connection properties
	foreach n [array names O,$from,*] {
		append redo_cmd "array unset plugin $n;"
		append undo_cmd "set plugin($n) [list $plugin($n)];"
	}

	# remove outgoing connections
	if {[isPrompt $from]} {
		append undo_cmd "set plugin($from,p_follows) [list $plugin($from,p_follows)];"
		append redo_cmd "set plugin($from,p_follows) [list ""];"
	}	else {
		append undo_cmd "set plugin($from,q_follows) [list $plugin($from,q_follows)];"
		append redo_cmd "set plugin($from,q_follows) [list ""];"
	}

	if {$redo_cmd!=""} {
		move::add $redo_cmd $undo_cmd [res::Str undo_del_out_conn $from]
	  set plugin(prj,changed) 1
		updateList
	}
}

proc deleteAllConnections {from} {
	global plugin

	set undo_cmd ""
	set redo_cmd ""

	# delete connection properties
	foreach n [array names O,$from,*] {
		append redo_cmd "array unset plugin $n;"
		append undo_cmd "set plugin($n) [list $plugin($n)];"
	}
	foreach n [array names O,*,$from] {
		append redo_cmd "array unset plugin $n;"
		append undo_cmd "set plugin($n) [list $plugin($n)];"
	}

	# remove $from from all follow ups of other objects
	foreach n [allEntries P] {
		set p [lsearch $plugin($n,p_follows) $from]
		if {$p>-1} {
			append undo_cmd "set plugin($n,p_follows) [list $plugin($n,p_follows)];"
			append redo_cmd "set plugin($n,p_follows) [list [lreplace $plugin($n,p_follows) $p $p]];"
		}
	}	
	foreach n [allEntries Q] {
		set p [lsearch $plugin($n,q_follows) $from]
		if {$p>-1} {
			append undo_cmd "set plugin($n,q_follows) [list $plugin($n,q_follows)];"
			append redo_cmd "set plugin($n,q_follows) [list [lreplace $plugin($n,q_follows) $p $p]];"
		}
	}	

	# remove outgoing connections
	if {[isPrompt $from]} {
		append undo_cmd "set plugin($from,p_follows) [list $plugin($from,p_follows)];"
		append redo_cmd "set plugin($from,p_follows) [list ""];"
	}	else {
		append undo_cmd "set plugin($from,q_follows) [list $plugin($from,q_follows)];"
		append redo_cmd "set plugin($from,q_follows) [list ""];"
	}
	
	if {$redo_cmd!=""} {
		move::add $redo_cmd $undo_cmd [res::Str undo_del_all_conn $from]
	  set plugin(prj,changed) 1
		updateList
	}
}

proc deleteObj {id} {
  global widget plugin options

  set plugin(prj,changed) 1

	# create undo and redo commands for restoring the object
	set redo_cmd ""
	set undo_cmd ""

	# remove it from connections	
	foreach n [array names plugin O,$id,*] {
		append redo_cmd "array unset plugin $n;"
		append undo_cmd "set plugin($n) [list $plugin($n)];"
	}
	foreach n [array names plugin O,*,$id] {
		append redo_cmd "array unset plugin $n;"
		append undo_cmd "set plugin($n) [list $plugin($n)];"
	}

	# remove it from sample calls
	#...

	append redo_cmd "array unset plugin $id,*;"
	foreach n [array name plugin $id,*] {
		append undo_cmd "set plugin($n) [list $plugin($n)];"
	}

  # remove the current object from all other objects link lists
  foreach n [allEntries P] {
    set p [lsearch -exact $plugin($n,p_follows) $id]
    if {$p!=-1} {
			append undo_cmd "set plugin($n,p_follows) [list $plugin($n,p_follows)];"
			append redo_cmd "set plugin($n,p_follows) [list [lreplace $plugin($n,p_follows) $p $p]];"
    }
  }
  foreach n [allEntries Q] {
    set p [lsearch -exact $plugin($n,q_follows) $id]
    if {$p!=-1} {
			append undo_cmd "set plugin($n,q_follows) [list $plugin($n,q_follows)];"
			append redo_cmd "set plugin($n,q_follows) [list [lreplace $plugin($n,q_follows) $p $p]];"
    }
    set p [lsearch -exact $plugin($n,q_concepts) $id]
    if {$p!=-1} {
			append undo_cmd "set plugin($n,q_concepts) [list $plugin($n,q_concepts)];"
			append redo_cmd "set plugin($n,q_concepts) [list [lreplace $plugin($n,q_concepts) $p $p]];"
    }
  }

	# delete the object
	move::add $redo_cmd $undo_cmd [res::Str undo_delete $id]

	#debug "$undo_cmd\n\n\n$redo_cmd"

	updateList
	menuSetState
}

proc copyObj {old_id} {
  global widget plugin options

  set plugin(prj,changed) 1

  # copy the contents old --> new
  if {[isPrompt $old_id]} {
    set new_id [nextId "P"]
    foreach n [concat $plugin(pix,vars) $plugin(pix,text)] {
      set plugin($new_id,$n) $plugin($old_id,$n)
    }
    set plugin($new_id,p_id) $new_id
    set plugin($new_id,id) $new_id
	} elseif {[isQuestion $old_id]} {
    set new_id [nextId "Q"]
    foreach n [concat $plugin(qix,vars) $plugin(qix,text)] {
      set plugin($new_id,$n) $plugin($old_id,$n)
    }
    set plugin($new_id,q_id) $new_id
    set plugin($new_id,id) $new_id
	} elseif {[isSampleCall $old_id]} {
    set new_id [nextId "S"]
    foreach n [concat $plugin(six,vars) $plugin(six,text)] {
      set plugin($new_id,$n) $plugin($old_id,$n)
    }
    set plugin($new_id,s_id) $new_id
    set plugin($new_id,id) $new_id
	} else {
    set new_id [nextId "C"]
    foreach n [concat $plugin(cix,vars) $plugin(cix,text)] {
      set plugin($new_id,$n) $plugin($old_id,$n)
    }
    set plugin($new_id,c_id) $new_id
    set plugin($new_id,id) $new_id
  }
  set plugin($new_id,x) 0
  set plugin($new_id,y) 0

	move::add "copyObj $old_id" "array unset plugin $new_id,*" [res::Str undo_copy $old_id $new_id] 0

	updateList
}

proc fillEmptyFields {id from} {
	global plugin statusMsg

	set undo_cmd ""
	set redo_cmd ""

	# fill different fields for different objects
	set type [objectType $id]
	if {$type!=[objectType $from]} {
		set statusMsg [res::Str prjobjdiffer]
		return 0
	}

	set clist $plugin([set type]ix,text)

	# fill the empty fields
	foreach n $clist {
		if {$plugin($id,$n)=="" && $plugin($from,$n)!=""} {
			append undo_cmd "set plugin($id,$n) [list ""];"
			append redo_cmd "set plugin($id,$n) [list $plugin($from,$n)];"
		}
	}

	if {$redo_cmd!=""} {
		move::add $redo_cmd $undo_cmd [res::Str undo_fillempty $id $from] 1
	  set plugin(prj,changed) 1
		updateList
	}

	return 1
}

proc compactNamespace {} {
  # remove name gaps in questions, prompts, and concepts
  global plugin statusMsg options rename

	proc _renameList {list} {
		global rename
		set ret {}
		foreach e $list {
			if {[info exists rename($e)]} {
				lappend ret $rename($e)
			} else {
				lappend ret $e
			}
		}
		return $ret
	}
 
 	proc _regsub_all {str} {
		# replace all object IDs in a string with the renamed ones
		global rename
		set ret $str
		# start with the small ones to avoid overwriting the others
		foreach n [lsort -increasing [array names rename]] {
			regsub -all -nocase $n $ret $rename($n) ret
		}
		return $ret
 	}
 
  # compute new ids for the new data
  array unset rename
  
  set allP [lsort [allEntries P]]
  set nextP 1
  foreach n $allP {
    set new "P[util::nullNum $nextP 3]"
		if {$n!=$new} {set rename($n) $new}
    incr nextP
  }
  
  set allQ [lsort [allEntries Q]]
  set nextQ 1
  foreach n $allQ {
    set new "Q[util::nullNum $nextQ 3]"
		if {$n!=$new} {set rename($n) $new}
    incr nextQ
  }
  
  set allC [lsort [allEntries C]]
  set nextC 1
  foreach n $allC {
    set new "C[util::nullNum $nextC 3]"
		if {$n!=$new} {set rename($n) $new}
    incr nextC
  }

  set allS [lsort [allEntries S]]
  set nextS 1
  foreach n $allS {
    set new "S[util::nullNum $nextS 3]"
		if {$n!=$new} {set rename($n) $new}
    incr nextS
  }

	# is there something to rename?
	if {[array names rename]=={}} {
	  set statusMsg [res::Str prjnocompact]
		return 0
	}

#  switch [tk_messageBox -type yesno \
#                        -message [res::Str prjnomoreundo]] {
#    yes {
#    }
#    no {
#			return 0
#    }
#  }
 
  #debug "[array get rename]"

	# global undo to the current point
	set undo_cmd "array unset plugin; array set plugin [list [array get plugin]]"
	set redo_cmd ""

	# transfer all data from the old to the new objects and delete the old ones
	append redo_cmd "array unset plugin P*,*;"
	append redo_cmd "array unset plugin Q*,*;"
	append redo_cmd "array unset plugin C*,*;"
	append redo_cmd "array unset plugin S*,*;"
	foreach n [concat $allP $allQ $allC $allS] {
    foreach nn [array names plugin $n,*] {
			set name [lindex [split $nn ,] 1]
			append redo_cmd "set plugin([_renameList $n],$name) [list [_regsub_all $plugin($nn)]];"
		}
	}

  # rename the prompts and questions in connections
	foreach n [array names plugin O,*] {
		set c1 [lindex [split $n ,] 1]
		set c2 [lindex [split $n ,] 2]
		set ren [_renameList [list $c1 $c2]]
		set tmpplugin(O,[lindex $ren 0],[lindex $ren 1]) [_regsub_all $plugin($n)]
	}
	append redo_cmd "array unset plugin O,*;"
	append redo_cmd "array set plugin [list [array get tmpplugin]];"
	array unset tmpplugin

	#puts "[join [split $redo_cmd \;] \n]"

	move::add $redo_cmd $undo_cmd [res::Str undo_compact [llength [array names rename]]] 1
	updateList
  prjFlags "-" 1
  set statusMsg [res::Str prjcompact]
  return 1
}

proc objectStandardAudioFilename {id n} {
	return [ini::Str common.audioPrefix]$id[ini::Str Files.$n].wav
}

proc correctStandardAudioFilename {id n} {
	global plugin

	set ret [string trim $plugin($id,$n)]
	if {[string index $ret 0]=="\["} {
		regsub {^\[A:[^\]]*\]} $ret "\[A:[objectStandardAudioFilename $id $n]\]" ret
	} else {
		set ret "\[A:[objectStandardAudioFilename $id $n]\]$ret"
	}
	return $ret
}

proc removeStandardAudioFilename {id n} {
	global plugin

	set ret [string trim $plugin($id,$n)]
	if {[string index $ret 0]=="\["} {
		regsub {^\[A:[^\]]*\]} $ret "" ret
	}
	return $ret
}

proc correctAllStandardAudioFilenames {} {
	global plugin

	set changed 0
	foreach id [allEntries P] {
		foreach n $plugin(pix,phr) {
			set tmp [correctStandardAudioFilename $id $n]
			if {$tmp!=$plugin($id,$n)} {
				set changed 1
				append undo_cmd "set plugin($id,$n) [list $plugin($id,$n)];"
				append redo_cmd "set plugin($id,$n) [list $tmp];"
			}
		}
	}
	foreach id [allEntries Q] {
		foreach n $plugin(qix,phr) {
			set tmp [correctStandardAudioFilename $id $n]
			if {$tmp!=$plugin($id,$n)} {
				set changed 1
				append undo_cmd "set plugin($id,$n) [list $plugin($id,$n)];"
				append redo_cmd "set plugin($id,$n) [list $tmp];"
			}
		}
	}

	if {$changed} {
		move::add $redo_cmd $undo_cmd [res::Str undo_correct_audio $id]
    set plugin(prj,changed) 1
	  updateList
	} else {
		return 0
	}
}

proc removeAllStandardAudioFilenames {} {
	global plugin

	set changed 0
	foreach id [allEntries P] {
		foreach n $plugin(pix,phr) {
			set tmp [removeStandardAudioFilename $id $n]
			if {$tmp!=$plugin($id,$n)} {
				set changed 1
				append undo_cmd "set plugin($id,$n) [list $plugin($id,$n)];"
				append redo_cmd "set plugin($id,$n) [list $tmp];"
			}
		}
	}
	foreach id [allEntries Q] {
		foreach n $plugin(qix,phr) {
			set tmp [removeStandardAudioFilename $id $n]
			if {$tmp!=$plugin($id,$n)} {
				set changed 1
				append undo_cmd "set plugin($id,$n) [list $plugin($id,$n)];"
				append redo_cmd "set plugin($id,$n) [list $tmp];"
			}
		}
	}

	if {$changed} {
		move::add $redo_cmd $undo_cmd [res::Str undo_remove_audio $id]
    set plugin(prj,changed) 1
	  updateList
	} else {
		return 0
	}
}

proc createPromptCommands {id} {
	global plugin

	set redo_cmd ""
	
  # preset empty values
  foreach n [concat $plugin(pix,vars) $plugin(pix,text)] {
  	append redo_cmd "set plugin($id,$n) [list {}];"
  }
	if {[ini::Exists common.audioPrefix]} {
		foreach n $plugin(pix,phr) {
			append redo_cmd "set plugin($id,$n) [list \[A:[objectStandardAudioFilename $id $n]\]];"
		}
	}
  append redo_cmd "set plugin($id,id) $id;"
  append redo_cmd "set plugin($id,p_id) $id;"
  append redo_cmd "set plugin($id,p_title) $id;"
  append redo_cmd "set plugin($id,p_prompt_bi) 1;"
  append redo_cmd "set plugin($id,p_endstate) 0;"
  append redo_cmd "set plugin($id,p_reachable) 0;"

  append redo_cmd "set plugin($id,x) 0;"
  append redo_cmd "set plugin($id,y) 0;"

	return [list $redo_cmd "array unset plugin $id,*;"]
}

proc createSampleCallCommands {id} {
	global plugin

	set redo_cmd ""
	
  # preset empty values
  foreach n [concat $plugin(six,vars) $plugin(six,text)] {
  	append redo_cmd "set plugin($id,$n) [list {}];"
  }
  append redo_cmd "set plugin($id,id) $id;"
  append redo_cmd "set plugin($id,s_id) $id;"
  append redo_cmd "set plugin($id,s_name) $id;"

	return [list $redo_cmd "array unset plugin $id,*;"]
}

proc createConceptCommands {id} {
	global plugin

	set redo_cmd ""
	
  # preset empty values
  foreach n [concat $plugin(cix,vars) $plugin(cix,text)] {
  	append redo_cmd "set plugin($id,$n) [list {}];"
  }

  # if grammarStrategy is use_skeleton, and a skeleton file for the grammar exists,
  # include it here
  set fn "[ini::Str files.GRMskelFileC]"
  if {[file exists $fn] && [ini::Str VoiceXML.grammarStrategy]=="use_skeleton"} {
		if {![catch {set f [open $fn]}]} {
		  # read the file and replace the id
      regsub -all "{CONCEPT_ID}" [read $f] $id tmp
      regsub -all "{GRM_LANG}" $tmp $plugin(language) tmp
      append redo_cmd "set plugin($id,c_memo) [list $tmp];"
      close $f
    }
  }

  append redo_cmd "set plugin($id,id) $id;"
  append redo_cmd "set plugin($id,c_id) $id;"
  append redo_cmd "set plugin($id,c_title) $id;"

  append redo_cmd "set plugin($id,x) 0;"
  append redo_cmd "set plugin($id,y) 0;"

	return [list $redo_cmd "array unset plugin $id,*;"]
}

proc createQuestionCommands {id} {
	global plugin

	set redo_cmd ""

  # preset empty values
  foreach n [concat $plugin(qix,vars) $plugin(qix,text)] {
  	append redo_cmd "set plugin($id,$n) [list {}];"
  }
	if {[ini::Exists common.audioPrefix]} {
		foreach n $plugin(qix,phr) {
			append redo_cmd "set plugin($id,$n) [list \[A:[objectStandardAudioFilename $id $n].wav\]];"
		}
	}
  append redo_cmd "set plugin($id,id) $id;"
  append redo_cmd "set plugin($id,q_id) $id;"
  append redo_cmd "set plugin($id,q_title) $id;"
  append redo_cmd "set plugin($id,q_endstate) 0;"
  append redo_cmd "set plugin($id,q_reachable) 0;"

	append redo_cmd "set plugin($id,q_init_bi) 1;"
	append redo_cmd "set plugin($id,q_nu1_bi) 1;"
	append redo_cmd "set plugin($id,q_nu2_bi) 1;"
	append redo_cmd "set plugin($id,q_nu3_bi) 1;"
	append redo_cmd "set plugin($id,q_nr1_bi) 1;"
	append redo_cmd "set plugin($id,q_nr2_bi) 1;"
	append redo_cmd "set plugin($id,q_nr3_bi) 1;"
	append redo_cmd "set plugin($id,q_tl1_bi) 1;"
	append redo_cmd "set plugin($id,q_tl2_bi) 1;"
	append redo_cmd "set plugin($id,q_tl3_bi) 1;"
	append redo_cmd "set plugin($id,q_ctxhelp_bi) 1;"

  append redo_cmd "set plugin($id,x) 0;"
  append redo_cmd "set plugin($id,y) 0;"

	return [list $redo_cmd "array unset plugin $id,*;"]
}
