set plugin(prj,loaded) ""
wozInit

proc prjFlags {{loaded "-"} {changed "-"} {analysis "-"}} {
  # set the flags, "-" are ignored
  global plugin
  
  # LOADED
  if {$loaded!="-"} {
    set plugin(prj,loaded) $loaded
    foreach v {changed analysis} {set plugin(prj,$v) 0}
  }
  if {$plugin(prj,loaded)==""} {return}

  # CHANGED
  if {$changed!="-"} {set plugin(prj,changed) $changed}

  # ANALYSIS
  if {$analysis!="-"} {set plugin(prj,analysis) $analysis}
}

proc prjOpen {fn} {
  # open a project, read the report and the parameters
  global plugin statusMsg options woz
 
  # close the current project of necessary
  if {![prjClose]} {return 0}
 
  # open the project file
  if {[catch {set f [open $fn r]} statusMsg]} {return 0}
	fconfigure $f -encoding utf-8
 
  # read the project file
  if {[catch {set plugin(prj) [read $f]} statusMsg]} {return 0}
  close $f
 
  # catch illegal project contents
  if {[catch {set prjVersion [lindex $plugin(prj) $options(ix,version)]} statusMsg]} {return 0}
  if {$prjVersion>[res::Str prjversion]} {
    set statusMsg [res::Str prjillegal $prjVersion [res::Str prjversion]]
    return 0
  }
 
  # get options and woz data
  catch {array set options [lindex $plugin(prj) $options(ix,options)]}
  catch {array set plugin  [lindex $plugin(prj) $options(ix,woz)]}
 
  # for version 1.0
  set allnames {}
  foreach n [array names plugin P*] {lappend allnames [lindex [split $n ","] 0]}
  foreach n [array names plugin Q*] {lappend allnames [lindex [split $n ","] 0]}
  foreach n [array names plugin C*] {lappend allnames [lindex [split $n ","] 0]}
 
  foreach n [lsort -uniq $allnames] {
    if {![info exists plugin($n,p_title)]} {set plugin($n,p_title) $n}
    if {![info exists plugin($n,q_title)]} {set plugin($n,q_title) $n}
    if {![info exists plugin($n,c_title)]} {set plugin($n,c_title) $n}
  }
 
  # reset the flags
  prjFlags $fn 0 0
 
	# if version is <1.1, create screen coordinates
	if {$prjVersion<"WOZ Project 1.1"} {
#  tk_messageBox -message "Create Coordinates" -type ok
	  set plugin(prj) [lreplace $plugin(prj) $options(ix,version) $options(ix,version) [res::Str prjversion]]
	  set x 0
	  set y 0
	  foreach n [concat [array names plugin P*,p_id] [array names plugin Q*,q_id]] {
			set name [lindex [split $n ","] 0]
			set plugin($name,x) $x
			set plugin($name,y) $y
			incr x 120
			if {[expr $x%600]==0} {
				set x 0
				incr y 80
			}
		}
		# reset the flags
		prjFlags $fn 1 0
	}
 
	# if version is <1.3, create barge-in flags
	if {$prjVersion<"WOZ Project 1.3"} {
	  foreach n [array names plugin P*,p_id] {
			set name [lindex [split $n ","] 0]
			set plugin($name,p_prompt_bi) 1
	  }
	  foreach n [array names plugin Q*,q_id] {
			set name [lindex [split $n ","] 0]
			set plugin($name,q_init_bi) 1
			set plugin($name,q_nu1_bi) 1
			set plugin($name,q_nu2_bi) 1
			set plugin($name,q_nu3_bi) 1
			set plugin($name,q_nr1_bi) 1
			set plugin($name,q_nr2_bi) 1
			set plugin($name,q_nr3_bi) 1
			set plugin($name,q_tl1_bi) 1
			set plugin($name,q_tl2_bi) 1
			set plugin($name,q_tl3_bi) 1
			set plugin($name,q_ctxhelp_bi) 1
		}
		# reset the flags
		prjFlags $fn 1 0
 	}
    
	# if version is <1.4, create endstate flags
	if {$prjVersion<"WOZ Project 1.4"} {
	  foreach n [array names plugin P*,p_id] {
			set name [lindex [split $n ","] 0]
			set plugin($name,p_endstate) 0
	  }
	  foreach n [array names plugin Q*,q_id] {
			set name [lindex [split $n ","] 0]
			set plugin($name,q_endstate) 0
		}
		# reset the flags
		prjFlags $fn 1 0
 	}
    
	# if version is <1.5, create reachable flags
	if {$prjVersion<"WOZ Project 1.5"} {
	  foreach n [array names plugin P*,p_id] {
			set name [lindex [split $n ","] 0]
			set plugin($name,p_reachable) 0
	  }
	  foreach n [array names plugin Q*,q_id] {
			set name [lindex [split $n ","] 0]
			set plugin($name,q_reachable) 0
		}
		# reset the flags
		prjFlags $fn 1 0
 	}
    
	# if version is <1.6, create VXML flags
	if {$prjVersion<"WOZ Project 1.6"} {
	  foreach n [array names plugin P*,p_id] {
			set name [lindex [split $n ","] 0]
			set plugin($name,p_vxml) 0
	  }
	  foreach n [array names plugin Q*,q_id] {
			set name [lindex [split $n ","] 0]
			set plugin($name,q_vxml) 0
		}
		# reset the flags
		prjFlags $fn 1 0
 	}
    
	# if version is <1.7, create language setting, default is en-US
	if {$prjVersion<"WOZ Project 1.7"} {
		set plugin(language) en-US
		# reset the flags
		prjFlags $fn 1 0
 	}
    
	res::Read $woz(RES)/languages/$plugin(language).dat
  set statusMsg [res::Str prjloaded "$fn ($plugin(name))"]
	move::init
  return 1
}

proc prjNew {} {
  # create a new project and open it
  global plugin statusMsg options woz

  # close the current project of necessary
  if {![prjClose]} {return 0}

  # construct a new structure
  set plugin(name) Untitled
	set plugin(language) en-US
  set plugin(prj) [list [res::Str prjversion] [array get options] [array get plugin]]

  # reset the flags
  prjFlags "NEWPROJECT" 0 0
	res::Read $woz(RES)/languages/en-US.ini
  set statusMsg [res::Str prjnewprj]
	move::init
  return 1
}

proc prjSave {fn} {
  # save a project to file
  global plugin statusMsg options

  # nothing loaded --> nothing to save
  if {$plugin(prj,loaded)==""} {set statusMsg [res::Str prjnoproject]; return 1}

  if {$fn=="" || $fn=="NEWPROJECT"} {
    set ft [list [list [ini::Str files.wozType] [ini::Str files.wozExt]]]
    set fn [tk_getSaveFile \
	      	-filetypes $ft \
	  	-defaultextension  [ini::Str files.wozExt] \
		-initialdir [ini::Str files.wozDir]]
    if {$fn=={}} {set statusMsg [res::Str prjnotsaved]; return 0}
  }

	# if there exists a project file, copy it to .bak
	if {[file exists $fn]} {
		if {[catch {file copy -force $fn [file rootname $fn].bak} statusMsg]} {return 0}
	}

  # open the project file
  if {[catch {set f [open $fn w]} statusMsg]} {return 0}
 	fconfigure $f -encoding utf-8

  # replace project contents with the current ones
  set plugin(prj) [list [res::Str prjversion] [array get options] \
			[concat [array get plugin name] [array get plugin language] [array get plugin P*] [array get plugin Q*] [array get plugin C*] [array get plugin S*] [array get plugin O*]]]

  # write the project file
  if {[catch {puts $f $plugin(prj)} statusMsg]} {return 0}
  close $f

  prjFlags $fn 0
  set statusMsg [res::Str prjsaved $fn]
  return 1
}

proc prjClose {} {
  # close the current project
  global plugin statusMsg curobj

  # nothing loaded --> nothing to close
  if {$plugin(prj,loaded)==""} {set statusMsg [res::Str prjnoproject]; return 1}

  # Changes --> ask
  if {$plugin(prj,changed)} {
    switch [tk_messageBox -type yesnocancel \
                          -message [res::Str prjsaveex $plugin(prj,loaded)]] {
      yes {
        if {![prjSave $plugin(prj,loaded)]} {
          return 0
        }
      }
      no {
      }
      cancel {
        set statusMsg [res::Str prjnotsaved]
        return 0
      }
    }
  }

  # close all
  #catch {
  #  array unset plugin P*
  #  array unset plugin Q*
  #  array unset plugin C*
  #  array unset plugin name
  #}
  array unset plugin
  array unset curobj
  prjFlags "" 0 0
  wozInit

	move::init
  return 1
}

proc prjMerge {fn} {
  # open a project, read the report and the parameters, merge into existing one
  global plugin statusMsg options

  if {$plugin(prj,changed)} {
    switch [tk_messageBox -type yesno \
                          -message [res::Str prjnomoreundo]] {
      yes {
				move::init
      }
      no {
				return 0
      }
    }
  }
 
  # open the project file
  if {[catch {set f [open $fn r]} statusMsg]} {return 0}
	fconfigure $f -encoding utf-8
 
  # read the project file
  if {[catch {set tmp [read $f]} statusMsg]} {return 0}
  close $f
 
  # catch illegal project contents
  if {[catch {set prjVersion [lindex $tmp $options(ix,version)]} statusMsg]} {return 0}
  if {$prjVersion!=[res::Str prjversion]} {
    set statusMsg [res::Str prjillegal $prjVersion [res::Str prjversion]]
    return 0
  }
 
  # read woz data
  array unset newdata
  catch {array set newdata [lindex $tmp $options(ix,woz)]}
 
  # compute new ids for the new data
  array unset rename
  set nextP [string trimleft [string range [nextId P] 1 end] 0]
  set nextC [string trimleft [string range [nextId C] 1 end] 0]
  set nextQ [string trimleft [string range [nextId Q] 1 end] 0]
  
  #tk_messageBox -message "new IDs start with: P $nextP Q $nextQ C $nextC" -type ok

  # error handling for overflows
  if {$nextP=="" || $nextQ=="" || $nextC==""} {
    set statusMsg [res::Str prjnamespace]
    return 0
  }

  set allP {}
  foreach n [lsort [array names newdata P*,p_id]] {
    lappend allP [lindex [split $n ","] 0]
    set rename([lindex [split $n ","] 0]) "P[util::nullNum $nextP 3]"
    incr nextP
    if {$nextP>999} {
      set statusMsg [res::Str prjnamespace]
      return 0
    }
  }
  
  set allQ {}
  foreach n [lsort [array names newdata Q*,q_id]] {
    lappend allQ [lindex [split $n ","] 0]
    set rename([lindex [split $n ","] 0]) "Q[util::nullNum $nextQ 3]"
    incr nextQ
    if {$nextQ>999} {
      set statusMsg [res::Str prjnamespace]
      return 0
    }
  }
  
  set allC {}
  foreach n [lsort [array names newdata C*,c_id]] {
    lappend allC [lindex [split $n ","] 0]
    set rename([lindex [split $n ","] 0]) "C[util::nullNum $nextC 3]"
    incr nextC
    if {$nextC>999} {
      set statusMsg [res::Str prjnamespace]
      return 0
    }
  }

  #tk_messageBox -message "[array get rename]" -type ok
  
  # rename the prompts and questions consistently in the follows and concepts lists of questions
  foreach n $allQ {
    set f {}
    foreach e $newdata($n,q_follows) {
      lappend f $rename($e)
    }
    set newdata($n,q_follows) $f

    set f {}
    foreach e $newdata($n,q_concepts) {
      lappend f $rename($e)
    }
    set newdata($n,q_concepts) $f
  }

  # rename the prompts and questions consistently in the follows lists of prompts
  foreach n $allP {
    set f {}
    foreach e $newdata($n,p_follows) {
      lappend f $rename($e)
    }
    set newdata($n,p_follows) $f
  }
  
  # rename the object indices
  foreach n [concat $allP $allQ $allC] {
    set new $rename($n)

    # rename the id's
    set newdata($n,id) $new
    if {[isPrompt $n]} {
      set newdata($n,p_id) $new
    } elseif {[isQuestion $n]} {
      set newdata($n,q_id) $new
    } else {
      set newdata($n,c_id) $new
    }

    foreach nn [array names newdata $n,*] {
      set plugin($new,[lindex [split $nn ,] 1]) $newdata($nn)
    }
  }

  # set the flags
  prjFlags "-" 1
  set statusMsg [res::Str prjmerged "$fn ($newdata(name))"]

  # clean up
  array unset newdata
  array unset rename

  return 1
}

prjFlags "" 0 0
