#==========================================================================================
#
# 	Written by Richard Breuer (richard.breuer@gmx.de)
#
#==========================================================================================
# Use Tix notebooks if available, otherwise simulate them
#==========================================================================================
# Jan 12, 2000	First Version [RB]
#==========================================================================================

namespace eval note {

# do we have Tix in a correct version?
variable tix
set tix [expr [string compare [package version Tix] "4.1.8.0"]!=-1]

variable note

proc New {note} {
  variable tix
  if {$tix} {
    tixNoteBook $note
  } else {
    # simulate a notebook with a radio button and several subframes
    frame $note
    frame $note.1 -relief groove -borderwidth 2
    frame $note.2
    pack $note.1 -fill x
    pack $note.2 -fill both -expand 1
  }
  return $note
}

proc Card {note card label {underline 0}} {
  variable tix
  if {$tix} {
    $note add $card -label $label -underline $underline
  } else {
    # create the card
    frame $note.2.$card -relief groove -borderwidth 2

    # create a button for the card
    radiobutton $note.1.$card \
      -text "$label" \
      -command "note::Raise $note $card" \
      -variable note::note($note) \
      -value $card
    pack $note.1.$card -side left

    # raise the new card
    $note.1.$card invoke

    # return the card frame
    return $note.2.$card
  }
}

proc Raise {note card} {
  catch {eval pack forget [pack slaves $note.2]}
  pack $note.2.$card -expand 1 -fill both
}

}


# ------------------------------------------------------------------------------------- TEST
if {0} {
catch {destroy .x}
note::New .x
pack .x -expand 1 -fill both

set w [note::Card .x help Hilfe]
button $w.b -text Help
pack $w.b

set w [note::Card .x info Info]
button $w.b -text Info
pack $w.b
}