
if {![info exists vTcl(sourcing)]} {

    package require Tk
    switch $tcl_platform(platform) {
	windows {
            option add *Button.padY 0
	}
	default {
            option add *Scrollbar.width 10
            option add *Scrollbar.highlightThickness 0
            option add *Scrollbar.elementBorderWidth 2
            option add *Scrollbar.borderWidth 2
	}
    }
    
}

#############################################################################
# Visual Tcl v1.60 Project
#


#################################
# VTCL LIBRARY PROCEDURES
#

if {![info exists vTcl(sourcing)]} {
#############################################################################
## Library Procedure:  Window

proc ::Window {args} {
    ## This procedure may be used free of restrictions.
    ##    Exception added by Christian Gavin on 08/08/02.
    ## Other packages and widget toolkits have different licensing requirements.
    ##    Please read their license agreements for details.

    global vTcl
    foreach {cmd name newname} [lrange $args 0 2] {}
    set rest    [lrange $args 3 end]
    if {$name == "" || $cmd == ""} { return }
    if {$newname == ""} { set newname $name }
    if {$name == "."} { wm withdraw $name; return }
    set exists [winfo exists $newname]
    switch $cmd {
        show {
            if {$exists} {
                wm deiconify $newname
            } elseif {[info procs vTclWindow$name] != ""} {
                eval "vTclWindow$name $newname $rest"
            }
            if {[winfo exists $newname] && [wm state $newname] == "normal"} {
                vTcl:FireEvent $newname <<Show>>
            }
        }
        hide    {
            if {$exists} {
                wm withdraw $newname
                vTcl:FireEvent $newname <<Hide>>
                return}
        }
        iconify { if $exists {wm iconify $newname; return} }
        destroy { if $exists {destroy $newname; return} }
    }
}
#############################################################################
## Library Procedure:  vTcl:DefineAlias

proc ::vTcl:DefineAlias {target alias widgetProc top_or_alias cmdalias} {
    ## This procedure may be used free of restrictions.
    ##    Exception added by Christian Gavin on 08/08/02.
    ## Other packages and widget toolkits have different licensing requirements.
    ##    Please read their license agreements for details.

    global widget
    set widget($alias) $target
    set widget(rev,$target) $alias
    if {$cmdalias} {
        interp alias {} $alias {} $widgetProc $target
    }
    if {$top_or_alias != ""} {
        set widget($top_or_alias,$alias) $target
        if {$cmdalias} {
            interp alias {} $top_or_alias.$alias {} $widgetProc $target
        }
    }
}
#############################################################################
## Library Procedure:  vTcl:DoCmdOption

proc ::vTcl:DoCmdOption {target cmd} {
    ## This procedure may be used free of restrictions.
    ##    Exception added by Christian Gavin on 08/08/02.
    ## Other packages and widget toolkits have different licensing requirements.
    ##    Please read their license agreements for details.

    ## menus are considered toplevel windows
    set parent $target
    while {[winfo class $parent] == "Menu"} {
        set parent [winfo parent $parent]
    }

    regsub -all {\%widget} $cmd $target cmd
    regsub -all {\%top} $cmd [winfo toplevel $parent] cmd

    uplevel #0 [list eval $cmd]
}
#############################################################################
## Library Procedure:  vTcl:FireEvent

proc ::vTcl:FireEvent {target event {params {}}} {
    ## This procedure may be used free of restrictions.
    ##    Exception added by Christian Gavin on 08/08/02.
    ## Other packages and widget toolkits have different licensing requirements.
    ##    Please read their license agreements for details.

    ## The window may have disappeared
    if {![winfo exists $target]} return
    ## Process each binding tag, looking for the event
    foreach bindtag [bindtags $target] {
        set tag_events [bind $bindtag]
        set stop_processing 0
        foreach tag_event $tag_events {
            if {$tag_event == $event} {
                set bind_code [bind $bindtag $tag_event]
                foreach rep "\{%W $target\} $params" {
                    regsub -all [lindex $rep 0] $bind_code [lindex $rep 1] bind_code
                }
                set result [catch {uplevel #0 $bind_code} errortext]
                if {$result == 3} {
                    ## break exception, stop processing
                    set stop_processing 1
                } elseif {$result != 0} {
                    bgerror $errortext
                }
                break
            }
        }
        if {$stop_processing} {break}
    }
}
#############################################################################
## Library Procedure:  vTcl:Toplevel:WidgetProc

proc ::vTcl:Toplevel:WidgetProc {w args} {
    ## This procedure may be used free of restrictions.
    ##    Exception added by Christian Gavin on 08/08/02.
    ## Other packages and widget toolkits have different licensing requirements.
    ##    Please read their license agreements for details.

    if {[llength $args] == 0} {
        ## If no arguments, returns the path the alias points to
        return $w
    }
    set command [lindex $args 0]
    set args [lrange $args 1 end]
    switch -- [string tolower $command] {
        "setvar" {
            foreach {varname value} $args {}
            if {$value == ""} {
                return [set ::${w}::${varname}]
            } else {
                return [set ::${w}::${varname} $value]
            }
        }
        "hide" - "show" {
            Window [string tolower $command] $w
        }
        "showmodal" {
            ## modal dialog ends when window is destroyed
            Window show $w; raise $w
            grab $w; tkwait window $w; grab release $w
        }
        "startmodal" {
            ## ends when endmodal called
            Window show $w; raise $w
            set ::${w}::_modal 1
            grab $w; tkwait variable ::${w}::_modal; grab release $w
        }
        "endmodal" {
            ## ends modal dialog started with startmodal, argument is var name
            set ::${w}::_modal 0
            Window hide $w
        }
        default {
            uplevel $w $command $args
        }
    }
}
#############################################################################
## Library Procedure:  vTcl:WidgetProc

proc ::vTcl:WidgetProc {w args} {
    ## This procedure may be used free of restrictions.
    ##    Exception added by Christian Gavin on 08/08/02.
    ## Other packages and widget toolkits have different licensing requirements.
    ##    Please read their license agreements for details.

    if {[llength $args] == 0} {
        ## If no arguments, returns the path the alias points to
        return $w
    }

    set command [lindex $args 0]
    set args [lrange $args 1 end]
    uplevel $w $command $args
}
#############################################################################
## Library Procedure:  vTcl:toplevel

proc ::vTcl:toplevel {args} {
    ## This procedure may be used free of restrictions.
    ##    Exception added by Christian Gavin on 08/08/02.
    ## Other packages and widget toolkits have different licensing requirements.
    ##    Please read their license agreements for details.

    uplevel #0 eval toplevel $args
    set target [lindex $args 0]
    namespace eval ::$target {set _modal 0}
}
}


if {[info exists vTcl(sourcing)]} {

proc vTcl:project:info {} {
    set base .top17
    namespace eval ::widgets::$base {
        set set,origin 0
        set set,size 1
        set runvisible 1
    }
    namespace eval ::widgets::$base.cpd19 {
        array set save {-background 1 -borderwidth 1 -height 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -width 1}
    }
    set site_3_0 $base.cpd19
    namespace eval ::widgets::$site_3_0.01 {
        array set save {-anchor 1 -background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -menu 1 -padx 1 -pady 1 -text 1}
    }
    namespace eval ::widgets::$site_3_0.01.02 {
        array set save {-activeborderwidth 1 -background 1 -borderwidth 1 -cursor 1 -font 1 -foreground 1 -tearoff 1}
        namespace eval subOptions {
            array set save {-accelerator 1 -command 1 -label 1 -menu 1}
        }
    }
    set site_5_0 $site_3_0.01.02
    namespace eval ::widgets::$site_5_0.men17 {
        array set save {-activeborderwidth 1 -background 1 -borderwidth 1 -font 1 -foreground 1 -tearoff 1}
        namespace eval subOptions {
            array set save {-command 1 -label 1}
        }
    }
    set site_5_0 $site_3_0.01.02
    namespace eval ::widgets::$site_5_0.men43 {
        array set save {-activeborderwidth 1 -borderwidth 1 -disabledforeground 1 -selectcolor 1 -tearoff 1}
        namespace eval subOptions {
            array set save {-command 1 -label 1}
        }
    }
    namespace eval ::widgets::$site_3_0.men17 {
        array set save {-anchor 1 -background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -menu 1 -padx 1 -pady 1 -text 1}
    }
    namespace eval ::widgets::$site_3_0.men17.01 {
        array set save {-activeborderwidth 1 -background 1 -borderwidth 1 -cursor 1 -font 1 -foreground 1 -tearoff 1}
        namespace eval subOptions {
            array set save {-accelerator 1 -command 1 -label 1}
        }
    }
    namespace eval ::widgets::$site_3_0.men71 {
        array set save {-background 1 -menu 1 -padx 1 -pady 1 -text 1}
    }
    namespace eval ::widgets::$site_3_0.men71.m {
        array set save {-activeborderwidth 1 -borderwidth 1 -font 1 -tearoff 1}
        namespace eval subOptions {
            array set save {-accelerator 1 -command 1 -label 1 -variable 1}
        }
    }
    namespace eval ::widgets::$site_3_0.men20 {
        array set save {-anchor 1 -background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -menu 1 -padx 1 -pady 1 -text 1}
    }
    namespace eval ::widgets::$site_3_0.men20.01 {
        array set save {-activeborderwidth 1 -background 1 -borderwidth 1 -cursor 1 -font 1 -foreground 1 -tearoff 1}
        namespace eval subOptions {
            array set save {-command 1 -label 1 -variable 1}
        }
    }
    namespace eval ::widgets::$site_3_0.05 {
        array set save {-anchor 1 -background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -menu 1 -padx 1 -pady 1 -text 1}
    }
    namespace eval ::widgets::$site_3_0.05.06 {
        array set save {-activeborderwidth 1 -background 1 -borderwidth 1 -cursor 1 -font 1 -foreground 1 -tearoff 1}
        namespace eval subOptions {
            array set save {-accelerator 1 -command 1 -label 1}
        }
    }
    namespace eval ::widgets::$base.fra25 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_3_0 $base.fra25
    namespace eval ::widgets::$site_3_0.but36 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.but26 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.but28 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.but29 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.but30 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.but31 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.fra17 {
        array set save {-background 1 -borderwidth 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    namespace eval ::widgets::$site_3_0.but32 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.but33 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.but34 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.but35 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.but18 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.but19 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.fra20 {
        array set save {-background 1 -borderwidth 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    namespace eval ::widgets::$site_3_0.but21 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.but22 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.but23 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.but24 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.but25 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.but27 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1}
    }
    namespace eval ::widgets::$site_3_0.but74 {
        array set save {-borderwidth 1 -command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.cpd71 {
        array set save {-background 1 -textvariable 1 -width 1}
    }
    namespace eval ::widgets::$base.fra20 {
        array set save {-background 1 -borderwidth 1 -height 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -width 1}
    }
    set site_3_0 $base.fra20
    namespace eval ::widgets::$site_3_0.cpd71 {
        array set save {-background 1 -height 1 -width 1}
    }
    set site_4_0 $site_3_0.cpd71
    namespace eval ::widgets::$site_4_0.01 {
        array set save {-background 1}
    }
    namespace eval ::widgets::$site_4_0.02 {
        array set save {-background 1}
    }
    set site_5_0 $site_4_0.02
    namespace eval ::widgets::$site_5_0.cpd72 {
        array set save {-borderwidth 1 -relief 1}
    }
    set site_6_0 $site_5_0.cpd72
    namespace eval ::widgets::$site_6_0.01 {
        array set save {-background 1 -font 1 -xscrollcommand 1 -yscrollcommand 1}
    }
    namespace eval ::widgets::$site_6_0.03 {
        array set save {-borderwidth 1 -command 1 -elementborderwidth 1 -width 1}
    }
    namespace eval ::widgets::$site_4_0.03 {
        array set save {-background 1 -borderwidth 1 -relief 1}
    }
    namespace eval ::widgets::$site_3_0.cpd73 {
        array set save {-height 1 -width 1}
    }
    set site_4_0 $site_3_0.cpd73
    namespace eval ::widgets::$site_4_0.01 {
        array set save {-background 1 -xscrollcommand 1 -yscrollcommand 1}
    }
    namespace eval ::widgets::$site_4_0.02 {
        array set save {-command 1 -orient 1}
    }
    namespace eval ::widgets::$site_4_0.03 {
        array set save {-command 1}
    }
    namespace eval ::widgets::$base.fra21 {
        array set save {-background 1 -borderwidth 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_3_0 $base.fra21
    namespace eval ::widgets::$site_3_0.ent22 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -state 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.fra17 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.fra17
    namespace eval ::widgets::$site_4_0.che21 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -highlightthickness 1 -padx 1 -pady 1 -selectcolor 1 -state 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_4_0.ent24 {
        array set save {-background 1 -borderwidth 1 -font 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -insertwidth 1 -selectbackground 1 -selectforeground 1 -show 1 -state 1 -textvariable 1 -width 1}
    }
    namespace eval ::widgets::$site_4_0.lab25 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -textvariable 1}
    }
    set base .top30
    namespace eval ::widgets::$base {
        set set,origin 0
        set set,size 1
        set runvisible 0
    }
    namespace eval ::widgets::$base.fra31 {
        array set save {-background 1 -borderwidth 1 -height 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -width 1}
    }
    set site_3_0 $base.fra31
    namespace eval ::widgets::$site_3_0.cpd33 {
        array set save {-background 1 -borderwidth 1 -highlightbackground 1 -highlightcolor 1 -relief 1}
    }
    set site_4_0 $site_3_0.cpd33
    namespace eval ::widgets::$site_4_0.01 {
        array set save {-borderwidth 1 -command 1 -cursor 1 -highlightbackground 1 -highlightcolor 1 -orient 1 -troughcolor 1 -width 1}
    }
    namespace eval ::widgets::$site_4_0.02 {
        array set save {-borderwidth 1 -command 1 -cursor 1 -highlightbackground 1 -highlightcolor 1 -troughcolor 1 -width 1}
    }
    namespace eval ::widgets::$site_4_0.03 {
        array set save {-background 1 -font 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1 -xscrollcommand 1 -yscrollcommand 1}
    }
    namespace eval ::widgets::$base.fra32 {
        array set save {-background 1 -borderwidth 1 -height 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -width 1}
    }
    set site_3_0 $base.fra32
    namespace eval ::widgets::$site_3_0.but34 {
        array set save {-background 1 -command 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1 -text 1 -textvariable 1}
    }
    set base .top35
    namespace eval ::widgets::$base {
        set set,origin 0
        set set,size 1
        set runvisible 0
    }
    namespace eval ::widgets::$base.fra36 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_3_0 $base.fra36
    namespace eval ::widgets::$site_3_0.fra39 {
        array set save {-background 1 -borderwidth 1 -height 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -width 1}
    }
    set site_4_0 $site_3_0.fra39
    namespace eval ::widgets::$site_4_0.lab41 {
        array set save {-background 1 -borderwidth 1 -font 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.fra42 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_5_0 $site_4_0.fra42
    namespace eval ::widgets::$site_5_0.lab44 {
        array set save {-anchor 1 -background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.lab45 {
        array set save {-anchor 1 -background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.lab46 {
        array set save {-anchor 1 -background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.lab47 {
        array set save {-anchor 1 -background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.ent48 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -textvariable 1 -width 1}
    }
    namespace eval ::widgets::$site_5_0.ent49 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -textvariable 1 -width 1}
    }
    namespace eval ::widgets::$site_5_0.ent50 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -textvariable 1 -width 1}
    }
    namespace eval ::widgets::$site_5_0.ent51 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -textvariable 1 -width 1}
    }
    namespace eval ::widgets::$base.fra37 {
        array set save {-background 1 -borderwidth 1 -height 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -width 1}
    }
    set site_3_0 $base.fra37
    namespace eval ::widgets::$site_3_0.but38 {
        array set save {-background 1 -command 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1 -text 1 -textvariable 1}
    }
    set base .top71
    namespace eval ::widgets::$base {
        set set,origin 0
        set set,size 1
        set runvisible 0
    }
    namespace eval ::widgets::$base.cpd72 {
        array set save {-background 1 -borderwidth 1 -height 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -width 1}
    }
    set site_3_0 $base.cpd72
    namespace eval ::widgets::$site_3_0.fra28 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.fra28
    namespace eval ::widgets::$site_4_0.lab29 {
        array set save {-anchor 1 -background 1 -borderwidth 1 -font 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.ent27 {
        array set save {-background 1 -font 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.lab30 {
        array set save {-background 1 -borderwidth 1 -font 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.fra19 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.fra19
    namespace eval ::widgets::$site_4_0.01 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_5_0 $site_4_0.01
    namespace eval ::widgets::$site_5_0.03 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.but20 {
        array set save {-background 1 -command 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.cpd74 {
        array set save {-background 1 -command 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1 -state 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.04 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_5_0 $site_4_0.04
    namespace eval ::widgets::$site_5_0.fra81 {
        array set save {-borderwidth 1 -height 1 -relief 1 -width 1}
    }
    set site_6_0 $site_5_0.fra81
    namespace eval ::widgets::$site_6_0.cpd82 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -variable 1}
    }
    namespace eval ::widgets::$site_6_0.cpd83 {
        array set save {-background 1 -borderwidth 1 -text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_6_0.cpd75 {
        array set save {-background 1 -borderwidth 1 -text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_5_0.fra22 {
        array set save {-height 1 -width 1}
    }
    set site_6_0 $site_5_0.fra22
    namespace eval ::widgets::$site_6_0.ent23 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -state 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.cpd77 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -state 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.fra22 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.fra22
    namespace eval ::widgets::$site_4_0.fra56 {
        array set save {-height 1 -width 1}
    }
    set site_5_0 $site_4_0.fra56
    namespace eval ::widgets::$site_5_0.cpd85 {
        array set save {-background 1 -borderwidth 1 -text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_5_0.lab58 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.but60 {
        array set save {-borderwidth 1 -highlightthickness 1 -padx 1 -pady 1}
    }
    namespace eval ::widgets::$site_4_0.tex24 {
        array set save {-background 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1}
    }
    namespace eval ::widgets::$site_3_0.fra25 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.fra25
    namespace eval ::widgets::$site_4_0.fra57 {
        array set save {-height 1 -width 1}
    }
    set site_5_0 $site_4_0.fra57
    namespace eval ::widgets::$site_5_0.lab59 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.cpd74 {
        array set save {-background 1 -borderwidth 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_4_0.tex27 {
        array set save {-background 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1}
    }
    namespace eval ::widgets::$site_3_0.cpd74 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.cpd75 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    set base .top72
    namespace eval ::widgets::$base {
        set set,origin 0
        set set,size 1
        set runvisible 0
    }
    namespace eval ::widgets::$base.cpd73 {
        array set save {-background 1 -borderwidth 1 -height 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -width 1}
    }
    set site_3_0 $base.cpd73
    namespace eval ::widgets::$site_3_0.fra28 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.fra28
    namespace eval ::widgets::$site_4_0.lab29 {
        array set save {-anchor 1 -background 1 -borderwidth 1 -font 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.cpd74 {
        array set save {-background 1 -borderwidth 1 -font 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.fra22 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.fra22
    namespace eval ::widgets::$site_4_0.fra56 {
        array set save {-height 1 -width 1}
    }
    set site_5_0 $site_4_0.fra56
    namespace eval ::widgets::$site_5_0.lab58 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.tex24 {
        array set save {-background 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1}
    }
    namespace eval ::widgets::$site_3_0.cpd74 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.cpd75 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    set base .top73
    namespace eval ::widgets::$base {
        set set,origin 1
        set set,size 1
        set runvisible 0
    }
    namespace eval ::widgets::$base.cpd74 {
        array set save {-background 1 -borderwidth 1 -height 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -width 1}
    }
    set site_3_0 $base.cpd74
    namespace eval ::widgets::$site_3_0.01 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.01
    namespace eval ::widgets::$site_4_0.02 {
        array set save {-anchor 1 -background 1 -borderwidth 1 -font 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.ent26 {
        array set save {-background 1 -font 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.03 {
        array set save {-background 1 -borderwidth 1 -font 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.fra59 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.fra59
    namespace eval ::widgets::$site_4_0.cpd75 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.cpd76 {
        array set save {-background 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1}
    }
    namespace eval ::widgets::$site_4_0.01 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.fra71 {
        array set save {-height 1 -width 1}
    }
    set site_5_0 $site_4_0.fra71
    namespace eval ::widgets::$site_5_0.cpd71 {
        array set save {-height 1 -width 1}
    }
    set site_6_0 $site_5_0.cpd71
    namespace eval ::widgets::$site_6_0.03 {
        array set save {-background 1 -closeenough 1 -height 1 -width 1 -yscrollcommand 1}
    }
    namespace eval ::widgets::$site_6_0.02 {
        array set save {-borderwidth 1 -command 1 -width 1}
    }
    namespace eval ::widgets::$site_3_0.fra71 {
        array set save {-borderwidth 1 -height 1 -relief 1 -width 1}
    }
    set site_4_0 $site_3_0.fra71
    namespace eval ::widgets::$site_4_0.cpd76 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.fra78 {
        array set save {-height 1 -width 1}
    }
    set site_5_0 $site_4_0.fra78
    namespace eval ::widgets::$site_5_0.cpd79 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.cpd80 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.cpd81 {
        array set save {-height 1 -width 1}
    }
    set site_5_0 $site_4_0.cpd81
    namespace eval ::widgets::$site_5_0.cpd79 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.cpd80 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.cpd71 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.cpd72 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.cpd72 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    set base .top74
    namespace eval ::widgets::$base {
        set set,origin 0
        set set,size 1
        set runvisible 0
    }
    namespace eval ::widgets::$base.fra75 {
        array set save {-borderwidth 1 -height 1 -relief 1 -width 1}
    }
    set site_3_0 $base.fra75
    namespace eval ::widgets::$site_3_0.lab72 {
        array set save {-background 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.che80 {
        array set save {-text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_3_0.che81 {
        array set save {-anchor 1 -text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_3_0.che82 {
        array set save {-text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_3_0.cpd73 {
        array set save {-text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$base.fra79 {
        array set save {-borderwidth 1 -height 1 -relief 1 -width 1}
    }
    set site_3_0 $base.fra79
    namespace eval ::widgets::$site_3_0.lab71 {
        array set save {-background 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.rad83 {
        array set save {-text 1 -textvariable 1 -value 1 -variable 1}
    }
    namespace eval ::widgets::$site_3_0.rad84 {
        array set save {-text 1 -textvariable 1 -value 1 -variable 1}
    }
    namespace eval ::widgets::$site_3_0.rad85 {
        array set save {-text 1 -textvariable 1 -value 1 -variable 1}
    }
    namespace eval ::widgets::$base.fra72 {
        array set save {-borderwidth 1 -height 1 -relief 1 -width 1}
    }
    set site_3_0 $base.fra72
    namespace eval ::widgets::$site_3_0.cpd74 {
        array set save {-background 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.fra75 {
        array set save {-borderwidth 1 -height 1 -relief 1 -width 1}
    }
    set site_4_0 $site_3_0.fra75
    namespace eval ::widgets::$site_4_0.ent76 {
        array set save {-background 1 -disabledbackground 1 -disabledforeground 1 -state 1 -textvariable 1 -width 1}
    }
    namespace eval ::widgets::$site_4_0.but77 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$base.fra76 {
        array set save {-borderwidth 1 -height 1 -relief 1 -width 1}
    }
    set site_3_0 $base.fra76
    namespace eval ::widgets::$site_3_0.but77 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.but78 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    set base .top75
    namespace eval ::widgets::$base {
        set set,origin 0
        set set,size 1
        set runvisible 0
    }
    namespace eval ::widgets::$base.fra76 {
        array set save {-borderwidth 1 -height 1 -relief 1 -width 1}
    }
    set site_3_0 $base.fra76
    namespace eval ::widgets::$site_3_0.lab80 {
        array set save {-textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.lab81 {
        array set save {-text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.lab82 {
        array set save {-text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.lab83 {
        array set save {-text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.ent85 {
        array set save {-background 1 -textvariable 1 -width 1}
    }
    namespace eval ::widgets::$site_3_0.che86 {
        array set save {-variable 1}
    }
    namespace eval ::widgets::$site_3_0.ent87 {
        array set save {-background 1 -textvariable 1 -width 1}
    }
    namespace eval ::widgets::$site_3_0.ent88 {
        array set save {-background 1 -textvariable 1 -width 1}
    }
    namespace eval ::widgets::$site_3_0.ent89 {
        array set save {-background 1 -disabledbackground 1 -state 1 -textvariable 1 -width 1}
    }
    namespace eval ::widgets::$site_3_0.but90 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$base.fra77 {
        array set save {-borderwidth 1 -height 1 -relief 1 -width 1}
    }
    set site_3_0 $base.fra77
    namespace eval ::widgets::$site_3_0.but78 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.but79 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    set base .top76
    namespace eval ::widgets::$base {
        set set,origin 0
        set set,size 1
        set runvisible 0
    }
    namespace eval ::widgets::$base.cpd77 {
        array set save {-background 1 -borderwidth 1 -height 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -width 1}
    }
    set site_3_0 $base.cpd77
    namespace eval ::widgets::$site_3_0.01 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.01
    namespace eval ::widgets::$site_4_0.02 {
        array set save {-anchor 1 -background 1 -borderwidth 1 -font 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.ent26 {
        array set save {-background 1 -font 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.03 {
        array set save {-background 1 -borderwidth 1 -font 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.fra47 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.fra47
    namespace eval ::widgets::$site_4_0.fra48 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_5_0 $site_4_0.fra48
    namespace eval ::widgets::$site_5_0.lab50 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.lab52 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.cpd83 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.fra49 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_5_0 $site_4_0.fra49
    namespace eval ::widgets::$site_5_0.ent51 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.che54 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -variable 1}
    }
    namespace eval ::widgets::$site_5_0.cpd82 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -state 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.fra59 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.fra59
    namespace eval ::widgets::$site_4_0.fra75 {
        array set save {-height 1 -width 1}
    }
    set site_5_0 $site_4_0.fra75
    namespace eval ::widgets::$site_5_0.but77 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.cpd76 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.02 {
        array set save {-background 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1}
    }
    namespace eval ::widgets::$site_3_0.cpd79 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.cpd80 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    set base .top81
    namespace eval ::widgets::$base {
        set set,origin 0
        set set,size 1
        set runvisible 0
    }
    namespace eval ::widgets::$base.cpd82 {
        array set save {-background 1 -borderwidth 1 -height 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -width 1}
    }
    set site_3_0 $base.cpd82
    namespace eval ::widgets::$site_3_0.01 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.01
    namespace eval ::widgets::$site_4_0.02 {
        array set save {-anchor 1 -background 1 -borderwidth 1 -font 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.ent25 {
        array set save {-background 1 -font 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.03 {
        array set save {-background 1 -borderwidth 1 -font 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.fra25 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.fra25
    namespace eval ::widgets::$site_4_0.fra35 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_5_0 $site_4_0.fra35
    namespace eval ::widgets::$site_5_0.but37 {
        array set save {-background 1 -command 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.but38 {
        array set save {-background 1 -command 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.cpd73 {
        array set save {-background 1 -command 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -pady 1 -state 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_4_0.fra36 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_5_0 $site_4_0.fra36
    namespace eval ::widgets::$site_5_0.cpd78 {
        array set save {-height 1 -width 1}
    }
    set site_6_0 $site_5_0.cpd78
    namespace eval ::widgets::$site_6_0.cpd79 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -state 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_6_0.cpd80 {
        array set save {-background 1 -borderwidth 1 -text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_6_0.cpd73 {
        array set save {-background 1 -borderwidth 1 -text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_5_0.fra19 {
        array set save {-height 1 -width 1}
    }
    set site_6_0 $site_5_0.fra19
    namespace eval ::widgets::$site_6_0.ent20 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -state 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.cpd72 {
        array set save {-background 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -state 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.04 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.04
    namespace eval ::widgets::$site_4_0.fra18 {
        array set save {-height 1 -width 1}
    }
    set site_5_0 $site_4_0.fra18
    namespace eval ::widgets::$site_5_0.cpd74 {
        array set save {-background 1 -borderwidth 1 -text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_5_0.lab19 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.but20 {
        array set save {-borderwidth 1 -highlightthickness 1 -padx 1 -pady 1}
    }
    namespace eval ::widgets::$site_4_0.06 {
        array set save {-background 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1}
    }
    namespace eval ::widgets::$site_3_0.fra20 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.fra20
    namespace eval ::widgets::$site_4_0.fra21 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_5_0 $site_4_0.fra21
    namespace eval ::widgets::$site_5_0.fra24 {
        array set save {-height 1 -width 1}
    }
    set site_6_0 $site_5_0.fra24
    namespace eval ::widgets::$site_6_0.cpd75 {
        array set save {-background 1 -borderwidth 1 -text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_6_0.lab33 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_6_0.but42 {
        array set save {-borderwidth 1 -highlightthickness 1 -padx 1 -pady 1}
    }
    namespace eval ::widgets::$site_5_0.02 {
        array set save {-background 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1}
    }
    namespace eval ::widgets::$site_4_0.fra22 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_5_0 $site_4_0.fra22
    namespace eval ::widgets::$site_5_0.fra25 {
        array set save {-height 1 -width 1}
    }
    set site_6_0 $site_5_0.fra25
    namespace eval ::widgets::$site_6_0.cpd76 {
        array set save {-background 1 -borderwidth 1 -text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_6_0.lab34 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_6_0.but45 {
        array set save {-borderwidth 1 -highlightthickness 1 -padx 1 -pady 1}
    }
    namespace eval ::widgets::$site_5_0.02 {
        array set save {-background 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1}
    }
    namespace eval ::widgets::$site_4_0.fra23 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_5_0 $site_4_0.fra23
    namespace eval ::widgets::$site_5_0.fra26 {
        array set save {-height 1 -width 1}
    }
    set site_6_0 $site_5_0.fra26
    namespace eval ::widgets::$site_6_0.cpd77 {
        array set save {-background 1 -borderwidth 1 -text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_6_0.lab35 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_6_0.but46 {
        array set save {-borderwidth 1 -highlightthickness 1 -padx 1 -pady 1}
    }
    namespace eval ::widgets::$site_5_0.02 {
        array set save {-background 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1}
    }
    namespace eval ::widgets::$site_3_0.fra24 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.fra24
    namespace eval ::widgets::$site_4_0.01 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_5_0 $site_4_0.01
    namespace eval ::widgets::$site_5_0.fra27 {
        array set save {-height 1 -width 1}
    }
    set site_6_0 $site_5_0.fra27
    namespace eval ::widgets::$site_6_0.cpd78 {
        array set save {-background 1 -borderwidth 1 -text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_6_0.lab36 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_6_0.but47 {
        array set save {-borderwidth 1 -highlightthickness 1 -padx 1 -pady 1}
    }
    namespace eval ::widgets::$site_5_0.03 {
        array set save {-background 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1}
    }
    namespace eval ::widgets::$site_4_0.04 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_5_0 $site_4_0.04
    namespace eval ::widgets::$site_5_0.fra28 {
        array set save {-height 1 -width 1}
    }
    set site_6_0 $site_5_0.fra28
    namespace eval ::widgets::$site_6_0.cpd79 {
        array set save {-background 1 -borderwidth 1 -text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_6_0.lab37 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_6_0.but48 {
        array set save {-borderwidth 1 -highlightthickness 1 -padx 1 -pady 1}
    }
    namespace eval ::widgets::$site_5_0.06 {
        array set save {-background 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1}
    }
    namespace eval ::widgets::$site_4_0.07 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_5_0 $site_4_0.07
    namespace eval ::widgets::$site_5_0.fra29 {
        array set save {-height 1 -width 1}
    }
    set site_6_0 $site_5_0.fra29
    namespace eval ::widgets::$site_6_0.cpd80 {
        array set save {-background 1 -borderwidth 1 -text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_6_0.lab38 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_6_0.but49 {
        array set save {-borderwidth 1 -highlightthickness 1 -padx 1 -pady 1}
    }
    namespace eval ::widgets::$site_5_0.09 {
        array set save {-background 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1}
    }
    namespace eval ::widgets::$site_3_0.fra29 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.fra29
    namespace eval ::widgets::$site_4_0.01 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_5_0 $site_4_0.01
    namespace eval ::widgets::$site_5_0.fra30 {
        array set save {-height 1 -width 1}
    }
    set site_6_0 $site_5_0.fra30
    namespace eval ::widgets::$site_6_0.cpd81 {
        array set save {-background 1 -borderwidth 1 -text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_6_0.lab39 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_6_0.but50 {
        array set save {-borderwidth 1 -highlightthickness 1 -padx 1 -pady 1}
    }
    namespace eval ::widgets::$site_5_0.03 {
        array set save {-background 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1}
    }
    namespace eval ::widgets::$site_4_0.04 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_5_0 $site_4_0.04
    namespace eval ::widgets::$site_5_0.fra31 {
        array set save {-height 1 -width 1}
    }
    set site_6_0 $site_5_0.fra31
    namespace eval ::widgets::$site_6_0.cpd82 {
        array set save {-background 1 -borderwidth 1 -text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_6_0.lab40 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_6_0.but51 {
        array set save {-borderwidth 1 -highlightthickness 1 -padx 1 -pady 1}
    }
    namespace eval ::widgets::$site_5_0.06 {
        array set save {-background 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1}
    }
    namespace eval ::widgets::$site_4_0.07 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_5_0 $site_4_0.07
    namespace eval ::widgets::$site_5_0.fra32 {
        array set save {-height 1 -width 1}
    }
    set site_6_0 $site_5_0.fra32
    namespace eval ::widgets::$site_6_0.cpd83 {
        array set save {-background 1 -borderwidth 1 -text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_6_0.but52 {
        array set save {-borderwidth 1 -highlightthickness 1 -padx 1 -pady 1}
    }
    namespace eval ::widgets::$site_6_0.lab41 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.09 {
        array set save {-background 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1}
    }
    namespace eval ::widgets::$site_3_0.fra17 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.fra17
    namespace eval ::widgets::$site_4_0.01 {
        array set save {-height 1 -width 1}
    }
    set site_5_0 $site_4_0.01
    namespace eval ::widgets::$site_5_0.cpd84 {
        array set save {-background 1 -borderwidth 1 -text 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_5_0.02 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.03 {
        array set save {-borderwidth 1 -highlightthickness 1 -padx 1 -pady 1}
    }
    namespace eval ::widgets::$site_4_0.04 {
        array set save {-background 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1}
    }
    namespace eval ::widgets::$site_3_0.fra18 {
        array set save {-background 1 -height 1 -highlightbackground 1 -highlightcolor 1 -width 1}
    }
    set site_4_0 $site_3_0.fra18
    namespace eval ::widgets::$site_4_0.01 {
        array set save {-height 1 -width 1}
    }
    set site_5_0 $site_4_0.01
    namespace eval ::widgets::$site_5_0.02 {
        array set save {-background 1 -borderwidth 1 -foreground 1 -highlightbackground 1 -highlightcolor 1 -relief 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_5_0.cpd73 {
        array set save {-background 1 -borderwidth 1 -textvariable 1 -variable 1}
    }
    namespace eval ::widgets::$site_4_0.03 {
        array set save {-background 1 -foreground 1 -height 1 -highlightbackground 1 -highlightcolor 1 -selectbackground 1 -selectforeground 1 -width 1 -wrap 1}
    }
    namespace eval ::widgets::$site_3_0.cpd84 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets::$site_3_0.cpd85 {
        array set save {-command 1 -pady 1 -text 1 -textvariable 1}
    }
    namespace eval ::widgets_bindings {
        set tagslist {_TopLevel <Key-Escape>}
    }
    namespace eval ::vTcl::modules::main {
        set procs {
            init
            main
            vTclWindow.
            vTclWindow.top17
            vTclWindow.top18
            vTclWindow.top19
            vTclWindow.top20
            vTclWindow.top30
            vTclWindow.top35
            vTclWindow.top63
        }
        set compounds {
        }
        set projectType single
    }
}
}

#################################
# USER DEFINED PROCEDURES
#
#############################################################################
## Procedure:  main

proc ::main {argc argv} {

}

#############################################################################
## Initialization Procedure:  init

proc ::init {argc argv} {

}

init $argc $argv

#################################
# VTCL GENERATED GUI PROCEDURES
#

proc vTclWindow. {base} {
    if {$base == ""} {
        set base .
    }
    ###################
    # CREATING WIDGETS
    ###################
    wm focusmodel $top passive
    wm geometry $top 200x200+66+66; update
    wm maxsize $top 1028 753
    wm minsize $top 104 1
    wm overrideredirect $top 0
    wm resizable $top 1 1
    wm withdraw $top
    wm title $top "vtcl"
    bindtags $top "$top Vtcl all"
    vTcl:FireEvent $top <<Create>>
    wm protocol $top WM_DELETE_WINDOW "vTcl:FireEvent $top <<DeleteWindow>>"

    ###################
    # SETTING GEOMETRY
    ###################

    vTcl:FireEvent $base <<Ready>>
}

proc vTclWindow.top17 {base} {
    if {$base == ""} {
        set base .top17
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    set top $base
    ###################
    # CREATING WIDGETS
    ###################
    vTcl:toplevel $top -class Toplevel \
        -background #dcdcdc -highlightbackground #dcdcdc \
        -highlightcolor #000000 
    wm focusmodel $top passive
    wm geometry $top 841x621; update
    #wm maxsize $top 1028 753
    wm minsize $top 115 1
    wm overrideredirect $top 0
    wm resizable $top 1 1
    wm deiconify $top
    wm title $top "Wizard of Oz"
    vTcl:DefineAlias "$top" "main" vTcl:Toplevel:WidgetProc "" 0
    bindtags $top "$top Toplevel all _TopLevel"
    bind $top <Control-Key-y> {
        menuCommand edit,redo
    }
    bind $top <Control-Key-z> {
        menuCommand edit,undo
    }
    bind $top <Key-F5> {
        menuCommand view,refresh
    }
    vTcl:FireEvent $top <<Create>>
    wm protocol $top WM_DELETE_WINDOW "vTcl:FireEvent $top <<DeleteWindow>>"

    frame $top.cpd19 \
        -borderwidth 1 -relief sunken -background #dcdcdc -height 30 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -width 30 
    set site_3_0 $top.cpd19
    menubutton $site_3_0.01 \
        -anchor w -background #dcdcdc -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 \
        -menu "$site_3_0.01.02" -padx 4 -pady 3 -text Datei 
    vTcl:DefineAlias "$site_3_0.01" "mm_file" vTcl:WidgetProc "main" 0
    menu $site_3_0.01.02 \
        -activeborderwidth 1 -background #dcdcdc -borderwidth 1 -cursor {} \
        -font {Tahoma 8} -foreground #000000 -tearoff 0 
    vTcl:DefineAlias "$site_3_0.01.02" "m_file" vTcl:WidgetProc "" 0
    $site_3_0.01.02 add command \
        -accelerator Ctrl+O -command {menuCommand file,open} -label ffnen... 
    $site_3_0.01.02 add command \
        -accelerator Ctrl+N -command {menuCommand file,new} -label Neu 
    $site_3_0.01.02 add command \
        -accelerator {} -command {menuCommand file,close} -label Schlieen 
    $site_3_0.01.02 add command \
        -accelerator Ctrl+S -command {menuCommand file,save} -label Sichern 
    $site_3_0.01.02 add command \
        -accelerator {} -command {menuCommand file,saveas} \
        -label {Sichern als...} 
    $site_3_0.01.02 add separator \
        
    $site_3_0.01.02 add command \
        -accelerator Ctrl+M -command {menuCommand file,merge} \
        -label Hinzuladen 
    $site_3_0.01.02 add cascade \
        -menu "$site_3_0.01.02.men17" -accelerator {} -command {} \
        -label Exportieren 
    set site_5_0 $site_3_0.01.02
    menu $site_5_0.men17 \
        -activeborderwidth 1 -background #dcdcdc -borderwidth 1 \
        -font {Tahoma 8} -foreground #000000 -tearoff 0 
    vTcl:DefineAlias "$site_5_0.men17" "m_file_export" vTcl:WidgetProc "" 0
    $site_5_0.men17 add command \
        -command {menuCommand file,export,phr} -label {Als Phrasenliste....} 
    $site_5_0.men17 add command \
        -command {menuCommand file,export,rtf} -label {Als RTF-Doku...} 
    $site_5_0.men17 add command \
        -command {menuCommand file,export,xml} -label {Als XML-Datei...} 
    $site_5_0.men17 add command \
        -command {menuCommand file,export,html} -label {Als HTMLDokument} 
    $site_5_0.men17 add command \
        -command {menuCommand file,export,vdx} -label {Als vdx-Datei} 
    $site_5_0.men17 add command \
        -command {menuCommand file,export,stw} -label {Als StudioWorks-Datei} 
    $site_5_0.men17 add command \
        -command {menuCommand file,export,ps} -label {Als Postscript-Grafik} 
    $site_5_0.men17 add command \
        -command {menuCommand file,export,gif} -label {Als GIF Bild} 
    $site_5_0.men17 add command \
        -command {menuCommand file,export,gml} -label {Als GML File} 
    $site_5_0.men17 add command \
        -command {menuCommand file,export,scall} -label Beispielcalls 
    $site_5_0.men17 add command \
        -command {menuCommand file,export,audio} -label Audio 
    $site_5_0.men17 add separator \
        
    $site_5_0.men17 add command \
        -command {menuCommand run,session} -label {Als Programm...} 
    $site_5_0.men17 add command \
        -command {menuCommand file,export,xhmi} -label {Als xHMI-Applikation} 
    $site_5_0.men17 add command \
        -command {menuCommand file,export,vxml} -label {VXML Prog} 
    $site_5_0.men17 add command \
        -command {menuCommand file,export,osdm} -label {Als OSDM} 
    $site_3_0.01.02 add cascade \
        -menu "$site_3_0.01.02.men43" -accelerator {} -command {} \
        -label Importieren 
    set site_5_0 $site_3_0.01.02
    menu $site_5_0.men43 \
        -activeborderwidth 1 -borderwidth 1 -disabledforeground #a3a3a3 \
        -selectcolor #b03060 -tearoff 0 
    vTcl:DefineAlias "$site_5_0.men43" "m_file_import" vTcl:WidgetProc "" 0
    $site_5_0.men43 add command \
        -command {menuCommand file,import,txt} -label {Von Textdatei...} 
    $site_5_0.men43 add command \
        -command {menuCommand file,import,scall} -label {Importiere scall} 
    $site_3_0.01.02 add separator \
        
    $site_3_0.01.02 add command \
        -accelerator {} -command {menuCommand file,exit} -label Beenden 
    menubutton $site_3_0.men17 \
        -anchor w -background #dcdcdc -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 \
        -menu "$site_3_0.men17.01" -padx 4 -pady 3 -text Editieren 
    vTcl:DefineAlias "$site_3_0.men17" "mm_edit" vTcl:WidgetProc "main" 0
    menu $site_3_0.men17.01 \
        -activeborderwidth 1 -background #dcdcdc -borderwidth 1 -cursor {} \
        -font {Tahoma 8} -foreground #000000 -tearoff 0 
    vTcl:DefineAlias "$site_3_0.men17.01" "m_edit" vTcl:WidgetProc "" 0
    $site_3_0.men17.01 add command \
        -accelerator Alt+P -command {menuCommand edit,prompt} \
        -label {Neuer Prompt} 
    $site_3_0.men17.01 add command \
        -accelerator Alt+Q -command {menuCommand edit,question} \
        -label {Neue Frage} 
    $site_3_0.men17.01 add command \
        -accelerator Alt+C -command {menuCommand edit,concept} \
        -label {Neues Konzept} 
    $site_3_0.men17.01 add command \
        -accelerator {} -command {menuCommand edit,scall} \
        -label {Neuer Beispielcall} 
    $site_3_0.men17.01 add command \
        -accelerator {} -command {menuCommand edit,rename} -label Umbenennen 
    $site_3_0.men17.01 add separator \
        
    $site_3_0.men17.01 add command \
        -accelerator Ctrl+Z -command {menuCommand edit,undo} -label Untuen 
    $site_3_0.men17.01 add command \
        -accelerator Ctrl+Y -command {menuCommand edit,redo} -label Retuen 
    $site_3_0.men17.01 add command \
        -accelerator {} -command {menuCommand edit,undo_all} \
        -label {Alles untuen} 
    $site_3_0.men17.01 add command \
        -accelerator {} -command {menuCommand edit,redo_all} \
        -label {Alles retuen} 
    $site_3_0.men17.01 add separator \
        
    $site_3_0.men17.01 add command \
        -accelerator {} -command {menuCommand edit,group} \
        -label Gruppenverbindung 
    $site_3_0.men17.01 add command \
        -accelerator {} -command {menuCommand edit,ungroup} \
        -label {Verbindungen lschen} 
    $site_3_0.men17.01 add command \
        -accelerator {} -command {menuCommand edit,compact} \
        -label Komprimieren 
    $site_3_0.men17.01 add command \
        -accelerator {} -command {menuCommand edit,sync} -label Sync 
    $site_3_0.men17.01 add command \
        -accelerator {} -command {menuCommand edit,remaudio} \
        -label {Alle Audio lsch} 
    $site_3_0.men17.01 add command \
        -accelerator {} -command {menuCommand edit,newaudio} \
        -label {Alle Audio neu} 
    $site_3_0.men17.01 add command \
        -accelerator {} -command {menuCommand edit,autocalls} \
        -label {Auto Calls} 
    $site_3_0.men17.01 add separator \
        
    $site_3_0.men17.01 add command \
        -accelerator {} -command {menuCommand edit,opt} -label Optionen 
    menubutton $site_3_0.men71 \
        -background #dcdcdc -menu "$site_3_0.men71.m" -padx 5 -pady 4 \
        -text Anzeige 
    vTcl:DefineAlias "$site_3_0.men71" "mm_view" vTcl:WidgetProc "main" 0
    menu $site_3_0.men71.m \
        -activeborderwidth 1 -borderwidth 1 -font {Tahoma 8} -tearoff 0 
    vTcl:DefineAlias "$site_3_0.men71.m" "m_view" vTcl:WidgetProc "" 0
    $site_3_0.men71.m add command \
        -accelerator F5 -command {menuCommand view,refresh} \
        -label Auffrischen 
    $site_3_0.men71.m add command \
        -accelerator {} -command {menuCommand view,layout} \
        -label {Layout berechnen} 
    $site_3_0.men71.m add checkbutton \
        -variable var(showconn) -accelerator {} \
        -command {menuCommand view,refresh} -label {Zeige Verbindungen} 
    menubutton $site_3_0.men20 \
        -anchor w -background #dcdcdc -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 \
        -menu "$site_3_0.men20.01" -padx 4 -pady 3 -text Ausfhren 
    vTcl:DefineAlias "$site_3_0.men20" "mm_run" vTcl:WidgetProc "main" 0
    menu $site_3_0.men20.01 \
        -activeborderwidth 1 -background #dcdcdc -borderwidth 1 -cursor {} \
        -font {Tahoma 8} -foreground #000000 -tearoff 0 
    vTcl:DefineAlias "$site_3_0.men20.01" "m_run" vTcl:WidgetProc "" 0
    $site_3_0.men20.01 add command \
        -command {menuCommand run,session} -label {Sitzung speichern} 
    $site_3_0.men20.01 add command \
        -command {menuCommand run,run} -label {Sitzung ausfhren} 
    $site_3_0.men20.01 add separator \
        
    $site_3_0.men20.01 add command \
        -command {menuCommand run,analyze} -label Analyse 
    $site_3_0.men20.01 add command \
        -command {menuCommand run,sc_analyze} -label {SC Analyse} 
    $site_3_0.men20.01 add command \
        -command {menuCommand edit,updlog} -label {Dalog-Log updaten} 
    $site_3_0.men20.01 add separator \
        
    $site_3_0.men20.01 add checkbutton \
        -variable var(tts) -command {} -label Textsynthese 
    menubutton $site_3_0.05 \
        -anchor w -background #dcdcdc -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 \
        -menu "$site_3_0.05.06" -padx 4 -pady 3 -text Hilfe 
    vTcl:DefineAlias "$site_3_0.05" "mm_help" vTcl:WidgetProc "main" 0
    menu $site_3_0.05.06 \
        -activeborderwidth 1 -background #dcdcdc -borderwidth 1 -cursor {} \
        -font {Tahoma 8} -foreground #000000 -tearoff 0 
    vTcl:DefineAlias "$site_3_0.05.06" "m_help" vTcl:WidgetProc "" 0
    $site_3_0.05.06 add command \
        -accelerator F1 -command {menuCommand help,docs} -label Doku 
    $site_3_0.05.06 add separator \
        
    $site_3_0.05.06 add command \
        -accelerator {} -command {menuCommand help,about} -label ber 
    $site_3_0.05.06 add command \
        -accelerator {} -command {menuCommand help,log} -label {Zeige Log} 
    pack $site_3_0.01 \
        -in $site_3_0 -anchor center -expand 0 -fill none -side left 
    pack $site_3_0.men17 \
        -in $site_3_0 -anchor center -expand 0 -fill none -side left 
    pack $site_3_0.men71 \
        -in $site_3_0 -anchor center -expand 0 -fill none -side left 
    pack $site_3_0.men20 \
        -in $site_3_0 -anchor center -expand 0 -fill none -side left 
    pack $site_3_0.05 \
        -in $site_3_0 -anchor center -expand 0 -fill none -side left 
    frame $top.fra25 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_3_0 $top.fra25
    button $site_3_0.but36 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but36" "tool00" vTcl:WidgetProc "main" 0
    button $site_3_0.but26 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but26" "tool01" vTcl:WidgetProc "main" 0
    button $site_3_0.but28 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but28" "tool02" vTcl:WidgetProc "main" 0
    button $site_3_0.but29 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but29" "tool03" vTcl:WidgetProc "main" 0
    button $site_3_0.but30 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but30" "tool04" vTcl:WidgetProc "main" 0
    button $site_3_0.but31 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but31" "tool05" vTcl:WidgetProc "main" 0
    frame $site_3_0.fra17 \
        -borderwidth 2 -background #dcdcdc -height 1 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -width 10 
    button $site_3_0.but32 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but32" "tool10" vTcl:WidgetProc "main" 0
    button $site_3_0.but33 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but33" "tool11" vTcl:WidgetProc "main" 0
    button $site_3_0.but34 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but34" "tool12" vTcl:WidgetProc "main" 0
    button $site_3_0.but35 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but35" "tool13" vTcl:WidgetProc "main" 0
    button $site_3_0.but18 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but18" "tool14" vTcl:WidgetProc "main" 0
    button $site_3_0.but19 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but19" "tool15" vTcl:WidgetProc "main" 0
    frame $site_3_0.fra20 \
        -borderwidth 2 -background #dcdcdc -height 1 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -width 10 
    button $site_3_0.but21 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but21" "tool20" vTcl:WidgetProc "main" 0
    button $site_3_0.but22 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but22" "tool21" vTcl:WidgetProc "main" 0
    button $site_3_0.but23 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but23" "tool22" vTcl:WidgetProc "main" 0
    button $site_3_0.but24 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but24" "tool23" vTcl:WidgetProc "main" 0
    button $site_3_0.but25 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but25" "tool24" vTcl:WidgetProc "main" 0
    button $site_3_0.but27 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 
    vTcl:DefineAlias "$site_3_0.but27" "tool25" vTcl:WidgetProc "main" 0
    button $site_3_0.but74 \
        -borderwidth 1 \
        -command {set tmplang [dia::chooseSingle [ini::Str voiceXML.supportedLanguages]]
if {$tmplang!=""} {
    set plugin(language) $tmplang
		global woz
    res::Read $woz(RES)/languages/$tmplang.dat
}} \
        -pady 0 -text lang -textvariable plugin(language) 
    vTcl:DefineAlias "$site_3_0.but74" "language" vTcl:WidgetProc "main" 0
    entry $site_3_0.cpd71 \
        -background white -textvariable plugin(name) -width 50 
    vTcl:DefineAlias "$site_3_0.cpd71" "Entry1" vTcl:WidgetProc "main" 0
    bind $site_3_0.cpd71 <KeyRelease> {
        global plugin
prjFlags "-" 1
wm title $widget(main) [res::Str win_main $plugin(name) [file tail $plugin(prj,loaded)]]
    }
    pack $site_3_0.but36 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.but26 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.but28 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.but29 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.but30 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.but31 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.fra17 \
        -in $site_3_0 -anchor center -expand 0 -fill none -side left 
    pack $site_3_0.but32 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.but33 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.but34 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.but35 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.but18 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.but19 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.fra20 \
        -in $site_3_0 -anchor center -expand 0 -fill none -side left 
    pack $site_3_0.but21 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.but22 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.but23 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.but24 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.but25 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.but27 \
        -in $site_3_0 -anchor w -expand 0 -fill none -side left 
    pack $site_3_0.but74 \
        -in $site_3_0 -anchor center -expand 0 -fill none -side right 
    pack $site_3_0.cpd71 \
        -in $site_3_0 -anchor center -expand 0 -fill none -side right 
    frame $top.fra20 \
        -borderwidth 1 -relief sunken -background #dcdcdc -height 75 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -width 125 
    set site_3_0 $top.fra20
    frame $site_3_0.cpd71 \
        -background #000000 -height 100 -width 200 
    set site_4_0 $site_3_0.cpd71
    frame $site_4_0.01 \
        -background #cccccc 
    vTcl:DefineAlias "$site_4_0.01" "mainframe" vTcl:WidgetProc "main" 0
    frame $site_4_0.02 \
        -background #9900991B99FE 
    set site_5_0 $site_4_0.02
    frame $site_5_0.cpd72 \
        -borderwidth 1 -relief raised 
    set site_6_0 $site_5_0.cpd72
    listbox $site_6_0.01 \
        -background #cccccc \
        -font -Adobe-Helvetica-Medium-R-Normal-*-*-120-*-*-*-*-*-* \
        -xscrollcommand {} -yscrollcommand "$site_6_0.03 set" 
    vTcl:DefineAlias "$site_6_0.01" "textlist" vTcl:WidgetProc "main" 0
    bind $site_6_0.01 <Button-3> {
        if {[winfo exists %W]} {
	tk::ListboxBeginSelect %W [%W index @%x,%y]
    }
createListContextMenu %x %y
    }
    bind $site_6_0.01 <Double-Button-1> {
        editPropertiesFromList
    }
    scrollbar $site_6_0.03 \
        -borderwidth 1 -command "$site_6_0.01 yview" -elementborderwidth 0 \
        -width 12 
    pack $site_6_0.01 \
        -in $site_6_0 -anchor center -expand 1 -fill both -side left 
    pack $site_6_0.03 \
        -in $site_6_0 -anchor center -expand 0 -fill y -side left 
    pack $site_5_0.cpd72 \
        -in $site_5_0 -anchor center -expand 1 -fill both -side top 
    frame $site_4_0.03 \
        -borderwidth 2 -relief raised -background #ff0000 
    bind $site_4_0.03 <B1-Motion> {
        set root [ split %W . ]
    set nb [ llength $root ]
    incr nb -1
    set root [ lreplace $root $nb $nb ]
    set root [ join $root . ]
    set width [ winfo width $root ].0
    
    set val [ expr (%X - [winfo rootx $root]) /$width ]

    if { $val >= 0 && $val <= 1.0 } {
    
        place $root.01 -relwidth $val
        place $root.03 -relx $val
        place $root.02 -relwidth [ expr 1.0 - $val ]
    }
    }
    place $site_4_0.01 \
        -in $site_4_0 -x 0 -y 0 -width -1 -relwidth 0.7521 -relheight 1 \
        -anchor nw -bordermode ignore 
    place $site_4_0.02 \
        -in $site_4_0 -x 0 -relx 1 -y 0 -width -1 -relwidth 0.2479 \
        -relheight 1 -anchor ne -bordermode ignore 
    place $site_4_0.03 \
        -in $site_4_0 -x 0 -relx 0.7521 -y 0 -rely 0.9 -width 10 -height 10 \
        -anchor s -bordermode ignore 
    frame $site_3_0.cpd73 \
        -height 30 -width 30 
    vTcl:DefineAlias "$site_3_0.cpd73" "Frame8" vTcl:WidgetProc "main" 0
    set site_4_0 $site_3_0.cpd73
    listbox $site_4_0.01 \
        -background white -xscrollcommand "$site_4_0.02 set" \
        -yscrollcommand "$site_4_0.03 set" 
    vTcl:DefineAlias "$site_4_0.01" "Listbox1" vTcl:WidgetProc "main" 0
    scrollbar $site_4_0.02 \
        -command "$site_4_0.01 xview" -orient horizontal 
    vTcl:DefineAlias "$site_4_0.02" "Scrollbar3" vTcl:WidgetProc "main" 0
    scrollbar $site_4_0.03 \
        -command "$site_4_0.01 yview" 
    vTcl:DefineAlias "$site_4_0.03" "Scrollbar4" vTcl:WidgetProc "main" 0
    grid $site_4_0.01 \
        -in $site_4_0 -column 0 -row 0 -columnspan 1 -rowspan 1 -sticky nesw 
    grid $site_4_0.02 \
        -in $site_4_0 -column 0 -row 1 -columnspan 1 -rowspan 1 -sticky ew 
    grid $site_4_0.03 \
        -in $site_4_0 -column 1 -row 0 -columnspan 1 -rowspan 1 -sticky ns 
    pack $site_3_0.cpd71 \
        -in $site_3_0 -anchor center -expand 1 -fill both -side top 
    frame $top.fra21 \
        -borderwidth 1 -background #dcdcdc -height 75 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -width 125 
    set site_3_0 $top.fra21
    entry $site_3_0.ent22 \
        -background #c0c0c0 -foreground #000000 -highlightbackground #ffffff \
        -highlightcolor #000000 -selectbackground #547098 \
        -selectforeground #ffffff -state disabled -textvariable statusMsg 
    vTcl:DefineAlias "$site_3_0.ent22" "status,bar" vTcl:WidgetProc "main" 0
    frame $site_3_0.fra17 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.fra17
    checkbutton $site_4_0.che21 \
        -background #dcdcdc -borderwidth 0 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 \
        -highlightthickness 0 -padx 0 -pady 0 -selectcolor #c0c0c0 \
        -state disabled -textvariable label(changed) \
        -variable plugin(prj,changed) 
    vTcl:DefineAlias "$site_4_0.che21" "status,changed" vTcl:WidgetProc "main" 0
    entry $site_4_0.ent24 \
        -background #c0c0c0 -borderwidth 1 -font {{Courier} 8} \
        -foreground #000000 -highlightbackground #ffffff \
        -highlightcolor #000000 -insertwidth 1 -selectbackground #547098 \
        -selectforeground #ffffff -show * -state disabled \
        -textvariable plugin(prj,loaded) -width 1 
    vTcl:DefineAlias "$site_4_0.ent24" "status,loaded" vTcl:WidgetProc "main" 0
    label $site_4_0.lab25 \
        -background #dcdcdc -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 \
        -textvariable label(loaded) 
    pack $site_4_0.che21 \
        -in $site_4_0 -anchor center -expand 0 -fill none -side left 
    pack $site_4_0.ent24 \
        -in $site_4_0 -anchor center -expand 0 -fill none -side left 
    pack $site_4_0.lab25 \
        -in $site_4_0 -anchor center -expand 0 -fill none -side left 
    pack $site_3_0.ent22 \
        -in $site_3_0 -anchor center -expand 1 -fill x -side left 
    pack $site_3_0.fra17 \
        -in $site_3_0 -anchor center -expand 0 -fill none -side left 
    ###################
    # SETTING GEOMETRY
    ###################
    pack $top.cpd19 \
        -in $top -anchor center -expand 0 -fill x -side top 
    pack $top.fra25 \
        -in $top -anchor center -expand 0 -fill x -side top 
    pack $top.fra20 \
        -in $top -anchor center -expand 1 -fill both -side top 
    pack $top.fra21 \
        -in $top -anchor center -expand 0 -fill x -side top 

    vTcl:FireEvent $base <<Ready>>
}

proc vTclWindow.top30 {base} {
    if {$base == ""} {
        set base .top30
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    set top $base
    ###################
    # CREATING WIDGETS
    ###################
    vTcl:toplevel $top -class Toplevel \
        -background #dcdcdc -highlightbackground #dcdcdc \
        -highlightcolor #000000 
    wm withdraw $top
    wm focusmodel $top passive
    wm geometry $top 349x303; update
    wm maxsize $top 1009 738
    wm minsize $top 115 1
    wm overrideredirect $top 0
    wm resizable $top 1 1
    wm title $top "Log"
    vTcl:DefineAlias "$top" "winlog" vTcl:Toplevel:WidgetProc "" 0
    bindtags $top "$top Toplevel all _TopLevel"
    bind $top <Key-Escape> {
        $widget(b,logclose) invoke
    }
    bind $top <Key-Return> {
        $widget(b,logclose) invoke
    }
    vTcl:FireEvent $top <<Create>>
    wm protocol $top WM_DELETE_WINDOW "vTcl:FireEvent $top <<DeleteWindow>>"

    frame $top.fra31 \
        -borderwidth 2 -relief groove -background #dcdcdc -height 75 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -width 125 
    set site_3_0 $top.fra31
    frame $site_3_0.cpd33 \
        -borderwidth 1 -relief raised -background #dcdcdc \
        -highlightbackground #dcdcdc -highlightcolor #000000 
    set site_4_0 $site_3_0.cpd33
    scrollbar $site_4_0.01 \
        -borderwidth 1 -command "$site_4_0.03 xview" -cursor left_ptr \
        -highlightbackground #dcdcdc -highlightcolor #000000 \
        -orient horizontal -troughcolor #dcdcdc -width 12 
    scrollbar $site_4_0.02 \
        -borderwidth 1 -command "$site_4_0.03 yview" -cursor left_ptr \
        -highlightbackground #dcdcdc -highlightcolor #000000 \
        -troughcolor #dcdcdc -width 12 
    text $site_4_0.03 \
        -background #ffffff \
        -font -Adobe-Helvetica-Medium-R-Normal-*-*-120-*-*-*-*-*-* \
        -foreground #000000 -height 1 -highlightbackground #ffffff \
        -highlightcolor #000000 -selectbackground #547098 \
        -selectforeground #ffffff -width 8 -wrap none \
        -xscrollcommand "$site_4_0.01 set" -yscrollcommand "$site_4_0.02 set" 
    vTcl:DefineAlias "$site_4_0.03" "log" vTcl:WidgetProc "winlog" 0
    grid $site_4_0.01 \
        -in $site_4_0 -column 0 -row 1 -columnspan 1 -rowspan 1 -sticky ew 
    grid $site_4_0.02 \
        -in $site_4_0 -column 1 -row 0 -columnspan 1 -rowspan 1 -sticky ns 
    grid $site_4_0.03 \
        -in $site_4_0 -column 0 -row 0 -columnspan 1 -rowspan 1 -sticky nesw 
    pack $site_3_0.cpd33 \
        -in $site_3_0 -anchor center -expand 1 -fill both -side top 
    grid columnconf $site_3_0.cpd33 0 -weight 1
    grid rowconf $site_3_0.cpd33 0 -weight 1
    frame $top.fra32 \
        -borderwidth 2 -relief groove -background #dcdcdc -height 75 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -width 125 
    set site_3_0 $top.fra32
    button $site_3_0.but34 \
        -background #dcdcdc \
        -command {::vTcl:Toplevel:WidgetProc $widget(winlog) endmodal} \
        -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 -text Schlieen \
        -textvariable label(close) 
    vTcl:DefineAlias "$site_3_0.but34" "b,logclose" vTcl:WidgetProc "winlog" 0
    pack $site_3_0.but34 \
        -in $site_3_0 -anchor center -expand 1 -fill none -padx 10 -pady 10 \
        -side left 
    ###################
    # SETTING GEOMETRY
    ###################
    pack $top.fra31 \
        -in $top -anchor center -expand 1 -fill both -side top 
    pack $top.fra32 \
        -in $top -anchor center -expand 0 -fill x -side top 

    vTcl:FireEvent $base <<Ready>>
}

proc vTclWindow.top35 {base} {
    if {$base == ""} {
        set base .top35
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    set top $base
    ###################
    # CREATING WIDGETS
    ###################
    vTcl:toplevel $top -class Toplevel \
        -background #dcdcdc -highlightbackground #dcdcdc \
        -highlightcolor #000000 
    wm withdraw $top
    wm focusmodel $top passive
    wm geometry $top 490x203; update
    wm maxsize $top 1009 738
    wm minsize $top 115 1
    wm overrideredirect $top 0
    wm resizable $top 1 1
    wm title $top "Optionen"
    vTcl:DefineAlias "$top" "winopt" vTcl:Toplevel:WidgetProc "" 0
    bindtags $top "$top Toplevel all _TopLevel"
    bind $top <Key-Escape> {
        $widget(b,setclose) invoke
    }
    bind $top <Key-Return> {
        $widget(b,setclose) invoke
    }
    vTcl:FireEvent $top <<Create>>
    wm protocol $top WM_DELETE_WINDOW "vTcl:FireEvent $top <<DeleteWindow>>"

    frame $top.fra36 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_3_0 $top.fra36
    frame $site_3_0.fra39 \
        -borderwidth 2 -relief groove -background #dcdcdc -height 75 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.fra39
    label $site_4_0.lab41 \
        -background #dec8de -borderwidth 1 -font {Helvetica 24 bold} \
        -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -relief raised -text Ausnahme-Einstellungen \
        -textvariable label(opt,exc,header) 
    frame $site_4_0.fra42 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_5_0 $site_4_0.fra42
    label $site_5_0.lab44 \
        -anchor w -background #dcdcdc -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 \
        -text {Wahrscheinlichkeit fr NOTHING_UNDERSTOOD in %} \
        -textvariable label(opt,exc,nu) 
    label $site_5_0.lab45 \
        -anchor w -background #dcdcdc -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 \
        -text {Wahrscheinlichkeit fr NOTHING_RECORDED in %} \
        -textvariable label(opt,exc,nr) 
    label $site_5_0.lab46 \
        -anchor w -background #dcdcdc -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 \
        -text {Wahrscheinlichkeit fr TOO_LONG in %} \
        -textvariable label(opt,exc,tl) 
    label $site_5_0.lab47 \
        -anchor w -background #dcdcdc -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 \
        -text {Wahrscheinlichkeit fr Erkennungsfehler in %} \
        -textvariable label(opt,exc,re) 
    entry $site_5_0.ent48 \
        -background #ffffff -foreground #000000 -highlightbackground #ffffff \
        -highlightcolor #000000 -selectbackground #547098 \
        -selectforeground #ffffff -textvariable options(exc_nu) -width 10 
    entry $site_5_0.ent49 \
        -background #ffffff -foreground #000000 -highlightbackground #ffffff \
        -highlightcolor #000000 -selectbackground #547098 \
        -selectforeground #ffffff -textvariable options(exc_tl) -width 10 
    entry $site_5_0.ent50 \
        -background #ffffff -foreground #000000 -highlightbackground #ffffff \
        -highlightcolor #000000 -selectbackground #547098 \
        -selectforeground #ffffff -textvariable options(exc_re) -width 10 
    entry $site_5_0.ent51 \
        -background #ffffff -foreground #000000 -highlightbackground #ffffff \
        -highlightcolor #000000 -selectbackground #547098 \
        -selectforeground #ffffff -textvariable options(exc_nr) -width 10 
    grid $site_5_0.lab44 \
        -in $site_5_0 -column 0 -row 0 -columnspan 1 -rowspan 1 -sticky ew 
    grid $site_5_0.lab45 \
        -in $site_5_0 -column 0 -row 1 -columnspan 1 -rowspan 1 -sticky ew 
    grid $site_5_0.lab46 \
        -in $site_5_0 -column 0 -row 2 -columnspan 1 -rowspan 1 -sticky ew 
    grid $site_5_0.lab47 \
        -in $site_5_0 -column 0 -row 3 -columnspan 1 -rowspan 1 -sticky ew 
    grid $site_5_0.ent48 \
        -in $site_5_0 -column 1 -row 0 -columnspan 1 -rowspan 1 
    grid $site_5_0.ent49 \
        -in $site_5_0 -column 1 -row 2 -columnspan 1 -rowspan 1 
    grid $site_5_0.ent50 \
        -in $site_5_0 -column 1 -row 3 -columnspan 1 -rowspan 1 
    grid $site_5_0.ent51 \
        -in $site_5_0 -column 1 -row 1 -columnspan 1 -rowspan 1 
    pack $site_4_0.lab41 \
        -in $site_4_0 -anchor center -expand 0 -fill x -side top 
    pack $site_4_0.fra42 \
        -in $site_4_0 -anchor center -expand 1 -fill x -side top 
    pack $site_3_0.fra39 \
        -in $site_3_0 -anchor center -expand 1 -fill both -side top 
    frame $top.fra37 \
        -borderwidth 2 -relief groove -background #dcdcdc -height 75 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -width 125 
    set site_3_0 $top.fra37
    button $site_3_0.but38 \
        -background #dcdcdc \
        -command {::vTcl:Toplevel:WidgetProc $widget(winopt) endmodal} \
        -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 -text Schlieen \
        -textvariable label(close) 
    vTcl:DefineAlias "$site_3_0.but38" "b,setclose" vTcl:WidgetProc "winopt" 0
    pack $site_3_0.but38 \
        -in $site_3_0 -anchor center -expand 0 -fill none -padx 10 -pady 10 \
        -side top 
    ###################
    # SETTING GEOMETRY
    ###################
    pack $top.fra36 \
        -in $top -anchor center -expand 1 -fill both -side top 
    pack $top.fra37 \
        -in $top -anchor center -expand 0 -fill x -side top 

    vTcl:FireEvent $base <<Ready>>
}

proc vTclWindow.top71 {base} {
    if {$base == ""} {
        set base .top71
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    set top $base
    ###################
    # CREATING WIDGETS
    ###################
    vTcl:toplevel $top -class Toplevel \
        -background #d9d9d9 -highlightbackground #d9d9d9 \
        -highlightcolor black 
    wm withdraw $top
    wm focusmodel $top passive
    wm geometry $top 609x422; update
    wm maxsize $top 1028 753
    wm minsize $top 104 1
    wm overrideredirect $top 0
    wm resizable $top 1 1
    wm title $top "Prompteigenschaften"
    vTcl:DefineAlias "$top" "property,p" vTcl:Toplevel:WidgetProc "" 0
    bindtags $top "$top Toplevel all _TopLevel"
    bind $top <Key-Escape> {
        $widget(property,p,cancel) invoke
    }
    vTcl:FireEvent $top <<Create>>
    wm protocol $top WM_DELETE_WINDOW "vTcl:FireEvent $top <<DeleteWindow>>"

    frame $top.cpd72 \
        -borderwidth 1 -relief sunken -background #dcdcdc -height 75 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -width 125 
    vTcl:DefineAlias "$top.cpd72" "pwindow" vTcl:WidgetProc "property,p" 0
    set site_3_0 $top.cpd72
    frame $site_3_0.fra28 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.fra28
    label $site_4_0.lab29 \
        -anchor w -background #dec8de -borderwidth 1 \
        -font {helvetica 24 bold} -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text Prompt -textvariable label(opt,p_header) 
    entry $site_4_0.ent27 \
        -background #c69fc6 -font {helvetica 12 bold} -foreground #000000 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff \
        -textvariable curobj(p_title) 
    label $site_4_0.lab30 \
        -background #dec8de -borderwidth 1 -font {helvetica 24 bold} \
        -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -relief raised -text {P 001} \
        -textvariable curobj(p_id) 
    pack $site_4_0.lab29 \
        -in $site_4_0 -anchor center -expand 1 -fill x -side left 
    pack $site_4_0.ent27 \
        -in $site_4_0 -anchor center -expand 0 -fill y -side left 
    pack $site_4_0.lab30 \
        -in $site_4_0 -anchor center -expand 0 -fill x -side left 
    frame $site_3_0.fra19 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.fra19
    frame $site_4_0.01 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_5_0 $site_4_0.01
    label $site_5_0.03 \
        -background #ddff78 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text Startprompt? -textvariable label(opt,p_start) 
    vTcl:DefineAlias "$site_5_0.03" "label,p_start" vTcl:WidgetProc "property,p" 0
    button $site_5_0.but20 \
        -background #ddff78 \
        -command {set curobj(p_follows) [dia::choosePQmulti $curobj(p_follows)]} \
        -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 -text Folgeaktionen \
        -textvariable label(opt,p_follows) 
    vTcl:DefineAlias "$site_5_0.but20" "label,p_follows" vTcl:WidgetProc "property,p" 0
    button $site_5_0.cpd74 \
        -background #ddff78 \
        -command {set curobj(p_follows) [dia::choosePQmulti $curobj(p_follows)]} \
        -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 -state disabled -text Folgeaktionen \
        -textvariable label(opt,p_predecessors) 
    vTcl:DefineAlias "$site_5_0.cpd74" "label,p_predecessors" vTcl:WidgetProc "property,p" 0
    pack $site_5_0.03 \
        -in $site_5_0 -anchor center -expand 0 -fill x -ipadx 10 -ipady 2 \
        -side top 
    pack $site_5_0.but20 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    pack $site_5_0.cpd74 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    frame $site_4_0.04 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_5_0 $site_4_0.04
    frame $site_5_0.fra81 \
        -borderwidth 2 -relief groove -height 75 -width 125 
    set site_6_0 $site_5_0.fra81
    checkbutton $site_6_0.cpd82 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -variable curobj(p_start) 
    checkbutton $site_6_0.cpd83 \
        -background #ddff78 -borderwidth 0 -text B-In \
        -textvariable label(endstate) -variable curobj(p_endstate) 
    checkbutton $site_6_0.cpd75 \
        -background #ddff78 -borderwidth 0 -text B-In \
        -textvariable label(reachable) -variable curobj(p_reachable) 
    pack $site_6_0.cpd82 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side left 
    pack $site_6_0.cpd83 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side right 
    pack $site_6_0.cpd75 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side right 
    frame $site_5_0.fra22 \
        -height 75 -width 125 
    set site_6_0 $site_5_0.fra22
    entry $site_6_0.ent23 \
        -background #dcdcdc -foreground #000000 -highlightbackground #ffffff \
        -highlightcolor #000000 -selectbackground #547098 \
        -selectforeground #ffffff -state disabled \
        -textvariable curobj(p_follows) 
    pack $site_6_0.ent23 \
        -in $site_6_0 -anchor center -expand 1 -fill x -pady 1 -side left 
    entry $site_5_0.cpd77 \
        -background #dcdcdc -foreground #000000 -highlightbackground #ffffff \
        -highlightcolor #000000 -selectbackground #547098 \
        -selectforeground #ffffff -state disabled \
        -textvariable curobj(p_predecessors) 
    pack $site_5_0.fra81 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    pack $site_5_0.fra22 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    pack $site_5_0.cpd77 \
        -in $site_5_0 -anchor center -expand 0 -fill x -pady 1 -side top 
    pack $site_4_0.01 \
        -in $site_4_0 -anchor center -expand 0 -fill none -side left 
    pack $site_4_0.04 \
        -in $site_4_0 -anchor center -expand 1 -fill x -side left 
    frame $site_3_0.fra22 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.fra22
    frame $site_4_0.fra56 \
        -height 75 -width 125 
    set site_5_0 $site_4_0.fra56
    checkbutton $site_5_0.cpd85 \
        -background #ddff78 -borderwidth 0 -text B-In \
        -textvariable label(barge_in) -variable curobj(p_prompt_bi) 
    label $site_5_0.lab58 \
        -background #ddff78 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text Prompt-Text -textvariable label(opt,p_prompt) 
    vTcl:DefineAlias "$site_5_0.lab58" "label,p_prompt" vTcl:WidgetProc "property,p" 0
    button $site_5_0.but60 \
        -borderwidth 1 -highlightthickness 0 -padx 0 -pady 0 
    vTcl:DefineAlias "$site_5_0.but60" "ls,p_prompt" vTcl:WidgetProc "property,p" 0
    pack $site_5_0.cpd85 \
        -in $site_5_0 -anchor center -expand 0 -fill none -side left 
    pack $site_5_0.lab58 \
        -in $site_5_0 -anchor center -expand 1 -fill x -side left 
    pack $site_5_0.but60 \
        -in $site_5_0 -anchor center -expand 0 -fill none -side right 
    text $site_4_0.tex24 \
        -background #ffffff -foreground #000000 -height 1 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff -width 1 \
        -wrap word 
    vTcl:DefineAlias "$site_4_0.tex24" "p_prompt" vTcl:WidgetProc "property,p" 0
    pack $site_4_0.fra56 \
        -in $site_4_0 -anchor center -expand 0 -fill x -side top 
    pack $site_4_0.tex24 \
        -in $site_4_0 -anchor center -expand 1 -fill both -side top 
    frame $site_3_0.fra25 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.fra25
    frame $site_4_0.fra57 \
        -height 75 -width 125 
    set site_5_0 $site_4_0.fra57
    label $site_5_0.lab59 \
        -background #ddff78 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text Kommentar -textvariable label(opt,p_memo) 
    vTcl:DefineAlias "$site_5_0.lab59" "label,p_memo" vTcl:WidgetProc "property,p" 0
    checkbutton $site_5_0.cpd74 \
        -background #ddff78 -borderwidth 0 -textvariable label(vxml) \
        -variable curobj(p_vxml) 
    pack $site_5_0.lab59 \
        -in $site_5_0 -anchor center -expand 1 -fill x -side left 
    pack $site_5_0.cpd74 \
        -in $site_5_0 -anchor center -expand 0 -fill none -side left 
    text $site_4_0.tex27 \
        -background #ffffff -foreground #000000 -height 1 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff -width 1 \
        -wrap word 
    vTcl:DefineAlias "$site_4_0.tex27" "p_memo" vTcl:WidgetProc "property,p" 0
    pack $site_4_0.fra57 \
        -in $site_4_0 -anchor center -expand 0 -fill x -side top 
    pack $site_4_0.tex27 \
        -in $site_4_0 -anchor center -expand 1 -fill both -side top 
    button $site_3_0.cpd74 \
        \
        -command {valCurobj2List
::vTcl:Toplevel:WidgetProc $widget(property,p) endmodal} \
        -pady 0 -text Anwenden -textvariable label(apply) 
    vTcl:DefineAlias "$site_3_0.cpd74" "Button1" vTcl:WidgetProc "property,p" 0
    button $site_3_0.cpd75 \
        \
        -command {if {![valCurobj2List 1] ||
     [tk_messageBox -message [res::Str leaveprop] -type yesno]=="yes"} {
    ::vTcl:Toplevel:WidgetProc $widget(property,p) endmodal
}} \
        -pady 0 -text Abbrechen -textvariable label(cancel) 
    vTcl:DefineAlias "$site_3_0.cpd75" "property,p,cancel" vTcl:WidgetProc "property,p" 0
    pack $site_3_0.fra28 \
        -in $site_3_0 -anchor center -expand 0 -fill x -side top 
    pack $site_3_0.fra19 \
        -in $site_3_0 -anchor center -expand 0 -fill x -side top 
    pack $site_3_0.fra22 \
        -in $site_3_0 -anchor center -expand 1 -fill both -side top 
    pack $site_3_0.fra25 \
        -in $site_3_0 -anchor center -expand 1 -fill both -side top 
    pack $site_3_0.cpd74 \
        -in $site_3_0 -anchor center -expand 1 -fill none -side left 
    pack $site_3_0.cpd75 \
        -in $site_3_0 -anchor center -expand 1 -fill none -side left 
    ###################
    # SETTING GEOMETRY
    ###################
    pack $top.cpd72 \
        -in $top -anchor center -expand 1 -fill both -side top 

    vTcl:FireEvent $base <<Ready>>
}

proc vTclWindow.top72 {base} {
    if {$base == ""} {
        set base .top72
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    set top $base
    ###################
    # CREATING WIDGETS
    ###################
    vTcl:toplevel $top -class Toplevel \
        -background #d9d9d9 -highlightbackground #d9d9d9 \
        -highlightcolor black 
    wm withdraw $top
    wm focusmodel $top passive
    wm geometry $top 442x165; update
    wm maxsize $top 1028 753
    wm minsize $top 104 1
    wm overrideredirect $top 0
    wm resizable $top 1 1
    wm title $top "Verbindungseigenschaften"
    vTcl:DefineAlias "$top" "property,o" vTcl:Toplevel:WidgetProc "" 0
    bindtags $top "$top Toplevel all _TopLevel"
    bind $top <Key-Escape> {
        $widget(property,o,cancel) invoke
    }
    vTcl:FireEvent $top <<Create>>
    wm protocol $top WM_DELETE_WINDOW "vTcl:FireEvent $top <<DeleteWindow>>"

    frame $top.cpd73 \
        -borderwidth 1 -relief sunken -background #dcdcdc -height 75 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -width 125 
    vTcl:DefineAlias "$top.cpd73" "pwindow1" vTcl:WidgetProc "property,o" 0
    set site_3_0 $top.cpd73
    frame $site_3_0.fra28 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.fra28
    label $site_4_0.lab29 \
        -anchor w -background #dec8de -borderwidth 1 \
        -font {helvetica 24 bold} -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text Prompt -textvariable label(opt,o_header) 
    label $site_4_0.cpd74 \
        -background #dec8de -borderwidth 1 -font {helvetica 24 bold} \
        -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -relief raised -text {P 001} \
        -textvariable curobj(id) 
    pack $site_4_0.lab29 \
        -in $site_4_0 -anchor center -expand 1 -fill x -side left 
    pack $site_4_0.cpd74 \
        -in $site_4_0 -anchor center -expand 0 -fill none -side left 
    frame $site_3_0.fra22 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.fra22
    frame $site_4_0.fra56 \
        -height 75 -width 125 
    set site_5_0 $site_4_0.fra56
    label $site_5_0.lab58 \
        -background #9999c28fffff -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text Prompt-Text -textvariable label(opt,o_condition) 
    vTcl:DefineAlias "$site_5_0.lab58" "label,o_condition" vTcl:WidgetProc "property,o" 0
    pack $site_5_0.lab58 \
        -in $site_5_0 -anchor center -expand 1 -fill x -side left 
    text $site_4_0.tex24 \
        -background #ffffff -foreground #000000 -height 1 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff -width 1 \
        -wrap word 
    vTcl:DefineAlias "$site_4_0.tex24" "o_condition" vTcl:WidgetProc "property,o" 0
    pack $site_4_0.fra56 \
        -in $site_4_0 -anchor center -expand 0 -fill x -side top 
    pack $site_4_0.tex24 \
        -in $site_4_0 -anchor center -expand 1 -fill both -side top 
    button $site_3_0.cpd74 \
        \
        -command {curobj2Connection
::vTcl:Toplevel:WidgetProc $widget(property,o) endmodal} \
        -pady 0 -text Anwenden -textvariable label(apply) 
    vTcl:DefineAlias "$site_3_0.cpd74" "Button1" vTcl:WidgetProc "property,o" 0
    button $site_3_0.cpd75 \
        \
        -command {if {![curobj2Connection 1] ||
     [tk_messageBox -message [res::Str leaveprop] -type yesno]=="yes"} {
    ::vTcl:Toplevel:WidgetProc $widget(property,o) endmodal
}} \
        -pady 0 -text Abbrechen -textvariable label(cancel) 
    vTcl:DefineAlias "$site_3_0.cpd75" "property,o,cancel" vTcl:WidgetProc "property,o" 0
    pack $site_3_0.fra28 \
        -in $site_3_0 -anchor center -expand 0 -fill x -side top 
    pack $site_3_0.fra22 \
        -in $site_3_0 -anchor center -expand 1 -fill both -side top 
    pack $site_3_0.cpd74 \
        -in $site_3_0 -anchor center -expand 1 -fill none -side left 
    pack $site_3_0.cpd75 \
        -in $site_3_0 -anchor center -expand 1 -fill none -side left 
    ###################
    # SETTING GEOMETRY
    ###################
    pack $top.cpd73 \
        -in $top -anchor center -expand 1 -fill both -side top 

    vTcl:FireEvent $base <<Ready>>
}

proc vTclWindow.top73 {base} {
    if {$base == ""} {
        set base .top73
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    set top $base
    ###################
    # CREATING WIDGETS
    ###################
    vTcl:toplevel $top -class Toplevel \
        -background #d9d9d9 -highlightbackground #d9d9d9 \
        -highlightcolor black 
    wm withdraw $top
    wm focusmodel $top passive
    wm geometry $top 750x400+94+87; update
    wm maxsize $top 1028 753
    wm minsize $top 104 1
    wm overrideredirect $top 0
    wm resizable $top 1 1
    wm title $top "Beispielcall Eigenschaften"
    vTcl:DefineAlias "$top" "property,s" vTcl:Toplevel:WidgetProc "" 0
    bindtags $top "$top Toplevel all _TopLevel"
    bind $top <Control-Key-y> {
        global curobj widget
move::redo 1 "s_call"
createSampleCallWidgets $curobj(s_id) $widget(s_call)
    }
    bind $top <Control-Key-z> {
        global curobj widget
move::undo 1 "s_call"
createSampleCallWidgets $curobj(s_id) $widget(s_call)
    }
    bind $top <Key-Escape> {
        $widget(property,s,cancel) invoke
    }
    vTcl:FireEvent $top <<Create>>
    wm protocol $top WM_DELETE_WINDOW "vTcl:FireEvent $top <<DeleteWindow>>"

    frame $top.cpd74 \
        -borderwidth 1 -relief sunken -background #dcdcdc -height 75 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -width 125 
    vTcl:DefineAlias "$top.cpd74" "cwindow1" vTcl:WidgetProc "property,s" 0
    set site_3_0 $top.cpd74
    frame $site_3_0.01 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.01
    label $site_4_0.02 \
        -anchor w -background #dec8de -borderwidth 1 \
        -font {helvetica 24 bold} -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text Konzept -textvariable label(opt,s_header) 
    entry $site_4_0.ent26 \
        -background #c69fc6 -font {helvetica 12 bold} -foreground #000000 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff \
        -textvariable curobj(s_name) 
    label $site_4_0.03 \
        -background #dec8de -borderwidth 1 -font {helvetica 24 bold} \
        -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -relief raised -text {C 001} \
        -textvariable curobj(s_id) 
    pack $site_4_0.02 \
        -in $site_4_0 -anchor center -expand 1 -fill x -side left 
    pack $site_4_0.ent26 \
        -in $site_4_0 -anchor center -expand 0 -fill y -side left 
    pack $site_4_0.03 \
        -in $site_4_0 -anchor center -expand 0 -fill x -side left 
    frame $site_3_0.fra59 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.fra59
    label $site_4_0.cpd75 \
        -background #ffffcb -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text Kommentar -textvariable label(opt,s_memo) 
    vTcl:DefineAlias "$site_4_0.cpd75" "label,s_memo" vTcl:WidgetProc "property,s" 0
    text $site_4_0.cpd76 \
        -background #ffffff -foreground #000000 -height 3 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff -width 1 \
        -wrap word 
    vTcl:DefineAlias "$site_4_0.cpd76" "s_memo" vTcl:WidgetProc "property,s" 0
    label $site_4_0.01 \
        -background #ffffcb -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text Kommentar -textvariable label(opt,s_call) 
    vTcl:DefineAlias "$site_4_0.01" "label,s_call" vTcl:WidgetProc "property,s" 0
    frame $site_4_0.fra71 \
        -height 75 -width 125 
    set site_5_0 $site_4_0.fra71
    frame $site_5_0.cpd71 \
        -height 30 -width 30 
    set site_6_0 $site_5_0.cpd71
    canvas $site_6_0.03 \
        -background #cccccc -closeenough 1.0 -height 0 -width 0 \
        -yscrollcommand "$site_6_0.02 set" 
    vTcl:DefineAlias "$site_6_0.03" "s_call" vTcl:WidgetProc "property,s" 0
    scrollbar $site_6_0.02 \
        -borderwidth 1 -command "$site_6_0.03 yview" -width 12 
    vTcl:DefineAlias "$site_6_0.02" "s_call_scroll" vTcl:WidgetProc "property,s" 0
    pack $site_6_0.03 \
        -in $site_6_0 -anchor center -expand 1 -fill both -side left 
    pack $site_6_0.02 \
        -in $site_6_0 -anchor center -expand 0 -fill y -side left 
    pack $site_5_0.cpd71 \
        -in $site_5_0 -anchor center -expand 1 -fill both -side top 
    pack $site_4_0.cpd75 \
        -in $site_4_0 -anchor center -expand 0 -fill x -side top 
    pack $site_4_0.cpd76 \
        -in $site_4_0 -anchor center -expand 0 -fill x -side top 
    pack $site_4_0.01 \
        -in $site_4_0 -anchor center -expand 0 -fill x -side top 
    pack $site_4_0.fra71 \
        -in $site_4_0 -anchor center -expand 1 -fill both -side top 
    frame $site_3_0.fra71 \
        -borderwidth 2 -relief groove -height 75 -width 125 
    vTcl:DefineAlias "$site_3_0.fra71" "Frame1" vTcl:WidgetProc "property,s" 0
    set site_4_0 $site_3_0.fra71
    button $site_4_0.cpd76 \
        \
        -command {saveTextContents
valCurobj2List
::vTcl:Toplevel:WidgetProc $widget(property,s) endmodal} \
        -pady 0 -text Anwenden -textvariable label(apply) 
    vTcl:DefineAlias "$site_4_0.cpd76" "Button1" vTcl:WidgetProc "property,s" 0
    frame $site_4_0.fra78 \
        -height 75 -width 125 
    vTcl:DefineAlias "$site_4_0.fra78" "Frame2" vTcl:WidgetProc "property,s" 0
    set site_5_0 $site_4_0.fra78
    button $site_5_0.cpd79 \
        -command appendSampleCallPrompt -pady 0 -text Anwenden \
        -textvariable label(append_p) 
    vTcl:DefineAlias "$site_5_0.cpd79" "s_append_p" vTcl:WidgetProc "property,s" 0
    button $site_5_0.cpd80 \
        -command appendSampleCallQuestion -pady 0 -text Anwenden \
        -textvariable label(append_q) 
    vTcl:DefineAlias "$site_5_0.cpd80" "s_append_q" vTcl:WidgetProc "property,s" 0
    pack $site_5_0.cpd79 \
        -in $site_5_0 -anchor center -expand 0 -fill none -side left 
    pack $site_5_0.cpd80 \
        -in $site_5_0 -anchor center -expand 0 -fill none -side left 
    frame $site_4_0.cpd81 \
        -height 75 -width 125 
    vTcl:DefineAlias "$site_4_0.cpd81" "Frame3" vTcl:WidgetProc "property,s" 0
    set site_5_0 $site_4_0.cpd81
    button $site_5_0.cpd79 \
        \
        -command {global curobj widget
move::undo 1 "s_call"
createSampleCallWidgets $curobj(s_id) $widget(s_call)} \
        -pady 0 -text Anwenden -textvariable Undo 
    vTcl:DefineAlias "$site_5_0.cpd79" "s_undo" vTcl:WidgetProc "property,s" 0
    button $site_5_0.cpd80 \
        \
        -command {global curobj widget
move::redo 1 "s_call"
createSampleCallWidgets $curobj(s_id) $widget(s_call)} \
        -pady 0 -text Anwenden -textvariable Redo 
    vTcl:DefineAlias "$site_5_0.cpd80" "s_redo" vTcl:WidgetProc "property,s" 0
    button $site_5_0.cpd71 \
        \
        -command {global curobj widget
move::undo "all" "s_call"
createSampleCallWidgets $curobj(s_id) $widget(s_call)} \
        -pady 0 -text Anwenden -textvariable {Undo all} 
    vTcl:DefineAlias "$site_5_0.cpd71" "s_undo_all" vTcl:WidgetProc "property,s" 0
    button $site_5_0.cpd72 \
        \
        -command {global curobj widget
move::redo "all" "s_call"
createSampleCallWidgets $curobj(s_id) $widget(s_call)} \
        -pady 0 -text Anwenden -textvariable {Redo all} 
    vTcl:DefineAlias "$site_5_0.cpd72" "s_redo_all" vTcl:WidgetProc "property,s" 0
    pack $site_5_0.cpd79 \
        -in $site_5_0 -anchor center -expand 0 -fill none -side left 
    pack $site_5_0.cpd80 \
        -in $site_5_0 -anchor center -expand 0 -fill none -side left 
    pack $site_5_0.cpd71 \
        -in $site_5_0 -anchor center -expand 0 -fill none -side left 
    pack $site_5_0.cpd72 \
        -in $site_5_0 -anchor center -expand 0 -fill none -side left 
    button $site_4_0.cpd72 \
        \
        -command {saveTextContents
if {![valCurobj2List 1] ||
     [tk_messageBox -message [res::Str leaveprop] -type yesno]=="yes"} {
    ::vTcl:Toplevel:WidgetProc $widget(property,s) endmodal
}} \
        -pady 0 -text Abbrechen -textvariable label(cancel) 
    vTcl:DefineAlias "$site_4_0.cpd72" "property,s,cancel" vTcl:WidgetProc "property,s" 0
    pack $site_4_0.cpd76 \
        -in $site_4_0 -anchor center -expand 1 -fill none -side left 
    pack $site_4_0.fra78 \
        -in $site_4_0 -anchor center -expand 1 -fill none -side left 
    pack $site_4_0.cpd81 \
        -in $site_4_0 -anchor center -expand 1 -fill none -side left 
    pack $site_4_0.cpd72 \
        -in $site_4_0 -anchor center -expand 1 -fill none -side left 
    pack $site_3_0.01 \
        -in $site_3_0 -anchor center -expand 0 -fill x -side top 
    pack $site_3_0.fra59 \
        -in $site_3_0 -anchor center -expand 1 -fill both -side top 
    pack $site_3_0.fra71 \
        -in $site_3_0 -anchor center -expand 0 -fill x -side top 
    ###################
    # SETTING GEOMETRY
    ###################
    pack $top.cpd74 \
        -in $top -anchor center -expand 1 -fill both -side top 

    vTcl:FireEvent $base <<Ready>>
}

proc vTclWindow.top74 {base} {
    if {$base == ""} {
        set base .top74
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    set top $base
    ###################
    # CREATING WIDGETS
    ###################
    vTcl:toplevel $top -class Toplevel
    wm withdraw $top
    wm focusmodel $top passive
    wm geometry $top 395x285; update
    wm maxsize $top 1028 753
    wm minsize $top 104 1
    wm overrideredirect $top 0
    wm resizable $top 1 1
    wm title $top "Sync Sample Calls"
    vTcl:DefineAlias "$top" "sync_sc" vTcl:Toplevel:WidgetProc "" 0
    bindtags $top "$top Toplevel all _TopLevel"
    bind $top <Key-Escape> {
        $widget(sc_cancel) invoke
    }
    bind $top <Key-Return> {
        $widget(sc_apply) invoke
    }
    vTcl:FireEvent $top <<Create>>
    wm protocol $top WM_DELETE_WINDOW "vTcl:FireEvent $top <<DeleteWindow>>"

    frame $top.fra75 \
        -borderwidth 2 -relief groove -height 75 -width 125 
    set site_3_0 $top.fra75
    label $site_3_0.lab72 \
        -background #ffffcb -relief sunken -text label \
        -textvariable label(sc_create_header) 
    checkbutton $site_3_0.che80 \
        -text {crete missing obj} -textvariable label(sc_create_missing_obj) \
        -variable options(sc_create_missing_obj) 
    checkbutton $site_3_0.che81 \
        -anchor n -text {crete missing obj} \
        -textvariable label(sc_create_new_obj) \
        -variable options(sc_create_new_obj) 
    checkbutton $site_3_0.che82 \
        -text {new connections} \
        -textvariable label(sc_create_new_connections) \
        -variable options(sc_create_new_connections) 
    checkbutton $site_3_0.cpd73 \
        -text {new connections} -textvariable label(sc_create_new_concepts) \
        -variable options(sc_create_new_concepts) 
    pack $site_3_0.lab72 \
        -in $site_3_0 -anchor center -expand 0 -fill x -side top 
    pack $site_3_0.che80 \
        -in $site_3_0 -anchor nw -expand 0 -fill none -side top 
    pack $site_3_0.che81 \
        -in $site_3_0 -anchor nw -expand 0 -fill none -side top 
    pack $site_3_0.che82 \
        -in $site_3_0 -anchor nw -expand 0 -fill none -side top 
    pack $site_3_0.cpd73 \
        -in $site_3_0 -anchor nw -expand 0 -fill none -side top 
    frame $top.fra79 \
        -borderwidth 2 -relief groove -height 75 -width 125 
    set site_3_0 $top.fra79
    label $site_3_0.lab71 \
        -background #ffffcb -relief sunken -text label \
        -textvariable label(sc_update_header) 
    radiobutton $site_3_0.rad83 \
        -text {update sample calls} -textvariable label(sc_update_calls) \
        -value update_calls -variable options(sc_update_wordings) 
    radiobutton $site_3_0.rad84 \
        -text update_design -textvariable label(sc_update_design) \
        -value update_design -variable options(sc_update_wordings) 
    radiobutton $site_3_0.rad85 \
        -text {do nothing} -textvariable label(sc_do_nothing) \
        -value do_nothing -variable options(sc_update_wordings) 
    pack $site_3_0.lab71 \
        -in $site_3_0 -anchor center -expand 0 -fill x -side top 
    pack $site_3_0.rad83 \
        -in $site_3_0 -anchor nw -expand 0 -fill none -side top 
    pack $site_3_0.rad84 \
        -in $site_3_0 -anchor nw -expand 0 -fill none -side top 
    pack $site_3_0.rad85 \
        -in $site_3_0 -anchor nw -expand 0 -fill none -side top 
    frame $top.fra72 \
        -borderwidth 2 -relief groove -height 75 -width 125 
    vTcl:DefineAlias "$top.fra72" "Frame1" vTcl:WidgetProc "sync_sc" 0
    set site_3_0 $top.fra72
    label $site_3_0.cpd74 \
        -background #ffffcb -relief sunken -text label \
        -textvariable label(sc_sync_header) 
    frame $site_3_0.fra75 \
        -borderwidth 2 -relief groove -height 75 -width 125 
    vTcl:DefineAlias "$site_3_0.fra75" "Frame2" vTcl:WidgetProc "sync_sc" 0
    set site_4_0 $site_3_0.fra75
    entry $site_4_0.ent76 \
        -background white -disabledbackground #ffffff \
        -disabledforeground #000000 -state disabled -textvariable var(sync) \
        -width 0 
    vTcl:DefineAlias "$site_4_0.ent76" "sc_calls" vTcl:WidgetProc "sync_sc" 0
    button $site_4_0.but77 \
        -command {set var(sync) [dia::chooseSmulti $var(sync)]} -pady 0 \
        -text Sync -textvariable label(sc_sync) 
    vTcl:DefineAlias "$site_4_0.but77" "Button1" vTcl:WidgetProc "sync_sc" 0
    pack $site_4_0.ent76 \
        -in $site_4_0 -anchor center -expand 1 -fill x -side left 
    pack $site_4_0.but77 \
        -in $site_4_0 -anchor center -expand 0 -fill none -side left 
    pack $site_3_0.cpd74 \
        -in $site_3_0 -anchor center -expand 0 -fill x -side top 
    pack $site_3_0.fra75 \
        -in $site_3_0 -anchor center -expand 0 -fill x -side top 
    frame $top.fra76 \
        -borderwidth 2 -relief groove -height 75 -width 125 
    set site_3_0 $top.fra76
    button $site_3_0.but77 \
        \
        -command {global var widget
::vTcl:Toplevel:WidgetProc $widget(sync_sc) endmodal
if {$var(sync)!=""} {syncSampleCalls $var(sync) $var(undoDomain)}} \
        -pady 0 -text Anwenden -textvariable label(apply) 
    vTcl:DefineAlias "$site_3_0.but77" "sc_apply" vTcl:WidgetProc "sync_sc" 0
    button $site_3_0.but78 \
        -command {::vTcl:Toplevel:WidgetProc $widget(sync_sc) endmodal} \
        -pady 0 -text Abbrechen -textvariable label(cancel) 
    vTcl:DefineAlias "$site_3_0.but78" "sc_cancel" vTcl:WidgetProc "sync_sc" 0
    pack $site_3_0.but77 \
        -in $site_3_0 -anchor center -expand 1 -fill none -side left 
    pack $site_3_0.but78 \
        -in $site_3_0 -anchor center -expand 1 -fill none -side left 
    ###################
    # SETTING GEOMETRY
    ###################
    pack $top.fra75 \
        -in $top -anchor center -expand 1 -fill both -side top 
    pack $top.fra79 \
        -in $top -anchor center -expand 1 -fill both -side top 
    pack $top.fra72 \
        -in $top -anchor center -expand 0 -fill x -side top 
    pack $top.fra76 \
        -in $top -anchor center -expand 0 -fill x -side top 

    vTcl:FireEvent $base <<Ready>>
}

proc vTclWindow.top75 {base} {
    if {$base == ""} {
        set base .top75
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    set top $base
    ###################
    # CREATING WIDGETS
    ###################
    vTcl:toplevel $top -class Toplevel
    wm withdraw $top
    wm focusmodel $top passive
    wm geometry $top 364x158; update
    wm maxsize $top 1028 753
    wm minsize $top 104 1
    wm overrideredirect $top 0
    wm resizable $top 1 1
    wm title $top "Generate Sample Calls"
    vTcl:DefineAlias "$top" "generate_sc" vTcl:Toplevel:WidgetProc "" 0
    bindtags $top "$top Toplevel all _TopLevel"
    vTcl:FireEvent $top <<Create>>
    wm protocol $top WM_DELETE_WINDOW "vTcl:FireEvent $top <<DeleteWindow>>"

    frame $top.fra76 \
        -borderwidth 2 -relief groove -height 75 -width 125 
    set site_3_0 $top.fra76
    label $site_3_0.lab80 \
        -textvariable label(sc_max_occurrences) 
    label $site_3_0.lab81 \
        -text label -textvariable label(sc_ends_at_end) 
    label $site_3_0.lab82 \
        -text label -textvariable label(sc_min_length) 
    label $site_3_0.lab83 \
        -text label -textvariable label(sc_max_length) 
    entry $site_3_0.ent85 \
        -background white -textvariable options(sc_max_occurrences) -width 3 
    checkbutton $site_3_0.che86 \
        -variable options(sc_ends_at_end) 
    entry $site_3_0.ent87 \
        -background white -textvariable options(sc_min_length) -width 3 
    entry $site_3_0.ent88 \
        -background white -textvariable options(sc_max_length) -width 3 
    entry $site_3_0.ent89 \
        -background white -disabledbackground #ffffff -state disabled \
        -textvariable options(sc_mandatory_states) -width 30 
    button $site_3_0.but90 \
        \
        -command {set options(sc_mandatory_states) [dia::choosePQmulti $options(sc_mandatory_states)]} \
        -pady 0 -text label -textvariable label(sc_mandatory_states) 
    grid $site_3_0.lab80 \
        -in $site_3_0 -column 0 -row 0 -columnspan 1 -rowspan 1 -sticky w 
    grid $site_3_0.lab81 \
        -in $site_3_0 -column 0 -row 1 -columnspan 1 -rowspan 1 -sticky w 
    grid $site_3_0.lab82 \
        -in $site_3_0 -column 0 -row 2 -columnspan 1 -rowspan 1 -sticky w 
    grid $site_3_0.lab83 \
        -in $site_3_0 -column 0 -row 3 -columnspan 1 -rowspan 1 -sticky w 
    grid $site_3_0.ent85 \
        -in $site_3_0 -column 1 -row 0 -columnspan 1 -rowspan 1 -sticky w 
    grid $site_3_0.che86 \
        -in $site_3_0 -column 1 -row 1 -columnspan 1 -rowspan 1 -sticky w 
    grid $site_3_0.ent87 \
        -in $site_3_0 -column 1 -row 2 -columnspan 1 -rowspan 1 -sticky w 
    grid $site_3_0.ent88 \
        -in $site_3_0 -column 1 -row 3 -columnspan 1 -rowspan 1 -sticky w 
    grid $site_3_0.ent89 \
        -in $site_3_0 -column 1 -row 4 -columnspan 1 -rowspan 1 -sticky w 
    grid $site_3_0.but90 \
        -in $site_3_0 -column 0 -row 4 -columnspan 1 -rowspan 1 -sticky w 
    frame $top.fra77 \
        -borderwidth 2 -relief groove -height 75 -width 125 
    set site_3_0 $top.fra77
    button $site_3_0.but78 \
        \
        -command {set calls [generateSampleCalls $options(sc_max_occurrences) $options(sc_ends_at_end) $options(sc_min_length) $options(sc_max_length) $options(sc_mandatory_states)]
#set calls [generateSampleCalls]
set n_calls [llength $calls]
if {$n_calls==0} {
    tk_messageBox -message "Nothing to do" -type ok
} else {
    completeSampleCalls $calls
}
::vTcl:Toplevel:WidgetProc $widget(generate_sc) endmodal} \
        -pady 0 -text Anwenden -textvariable label(apply) 
    button $site_3_0.but79 \
        -command {::vTcl:Toplevel:WidgetProc $widget(generate_sc) endmodal} \
        -pady 0 -text Abbrechen -textvariable label(cancel) 
    pack $site_3_0.but78 \
        -in $site_3_0 -anchor center -expand 1 -fill none -side left 
    pack $site_3_0.but79 \
        -in $site_3_0 -anchor center -expand 1 -fill none -side left 
    ###################
    # SETTING GEOMETRY
    ###################
    pack $top.fra76 \
        -in $top -anchor center -expand 1 -fill both -side top 
    pack $top.fra77 \
        -in $top -anchor center -expand 0 -fill x -side top 

    vTcl:FireEvent $base <<Ready>>
}

proc vTclWindow.top76 {base} {
    if {$base == ""} {
        set base .top76
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    set top $base
    ###################
    # CREATING WIDGETS
    ###################
    vTcl:toplevel $top -class Toplevel \
        -background #d9d9d9 -highlightbackground #d9d9d9 \
        -highlightcolor black 
    wm withdraw $top
    wm focusmodel $top passive
    wm geometry $top 609x293; update
    wm maxsize $top 1028 753
    wm minsize $top 104 1
    wm overrideredirect $top 0
    wm resizable $top 1 1
    wm title $top "Grammatikeigenschaften"
    vTcl:DefineAlias "$top" "property,c" vTcl:Toplevel:WidgetProc "" 0
    bindtags $top "$top Toplevel all _TopLevel"
    bind $top <Key-Escape> {
        $widget(property,c,cancel) invoke
    }
    vTcl:FireEvent $top <<Create>>
    wm protocol $top WM_DELETE_WINDOW "vTcl:FireEvent $top <<DeleteWindow>>"

    frame $top.cpd77 \
        -borderwidth 1 -relief sunken -background #dcdcdc -height 75 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -width 125 
    vTcl:DefineAlias "$top.cpd77" "cwindow" vTcl:WidgetProc "property,c" 0
    set site_3_0 $top.cpd77
    frame $site_3_0.01 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.01
    label $site_4_0.02 \
        -anchor w -background #dec8de -borderwidth 1 \
        -font {helvetica 24 bold} -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text Konzept -textvariable label(opt,c_header) 
    entry $site_4_0.ent26 \
        -background #c69fc6 -font {helvetica 12 bold} -foreground #000000 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff \
        -textvariable curobj(c_title) 
    label $site_4_0.03 \
        -background #dec8de -borderwidth 1 -font {helvetica 24 bold} \
        -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -relief raised -text {C 001} \
        -textvariable curobj(c_id) 
    pack $site_4_0.02 \
        -in $site_4_0 -anchor center -expand 1 -fill x -side left 
    pack $site_4_0.ent26 \
        -in $site_4_0 -anchor center -expand 0 -fill y -side left 
    pack $site_4_0.03 \
        -in $site_4_0 -anchor center -expand 0 -fill x -side left 
    frame $site_3_0.fra47 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.fra47
    frame $site_4_0.fra48 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_5_0 $site_4_0.fra48
    label $site_5_0.lab50 \
        -background #78ffd2 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text {Name (ohne <>)} -textvariable label(opt,c_name) 
    vTcl:DefineAlias "$site_5_0.lab50" "label,c_name" vTcl:WidgetProc "property,c" 0
    label $site_5_0.lab52 \
        -background #78ffd2 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text {Immer gltig?} -textvariable label(opt,c_always) 
    vTcl:DefineAlias "$site_5_0.lab52" "label,c_always" vTcl:WidgetProc "property,c" 0
    label $site_5_0.cpd83 \
        -background #78ffd2 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text Folgeaktionen -textvariable label(opt,c_usedby) 
    vTcl:DefineAlias "$site_5_0.cpd83" "label,c_usedby" vTcl:WidgetProc "property,c" 0
    pack $site_5_0.lab50 \
        -in $site_5_0 -anchor center -expand 0 -fill x -ipadx 10 -ipady 2 \
        -side top 
    pack $site_5_0.lab52 \
        -in $site_5_0 -anchor center -expand 0 -fill x -ipadx 10 -ipady 2 \
        -side top 
    pack $site_5_0.cpd83 \
        -in $site_5_0 -anchor center -expand 0 -fill x -ipadx 10 -ipady 2 \
        -side top 
    frame $site_4_0.fra49 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_5_0 $site_4_0.fra49
    entry $site_5_0.ent51 \
        -background #ffffff -foreground #000000 -highlightbackground #ffffff \
        -highlightcolor #000000 -selectbackground #547098 \
        -selectforeground #ffffff -textvariable curobj(c_name) 
    checkbutton $site_5_0.che54 \
        -background #dcdcdc -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -variable curobj(c_always) 
    entry $site_5_0.cpd82 \
        -background #dcdcdc -foreground #000000 -highlightbackground #ffffff \
        -highlightcolor #000000 -selectbackground #547098 \
        -selectforeground #ffffff -state disabled \
        -textvariable curobj(c_usedby) 
    pack $site_5_0.ent51 \
        -in $site_5_0 -anchor center -expand 1 -fill x -pady 2 -side top 
    pack $site_5_0.che54 \
        -in $site_5_0 -anchor w -expand 0 -fill none -side top 
    pack $site_5_0.cpd82 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    pack $site_4_0.fra48 \
        -in $site_4_0 -anchor center -expand 0 -fill none -side left 
    pack $site_4_0.fra49 \
        -in $site_4_0 -anchor center -expand 1 -fill x -side left 
    frame $site_3_0.fra59 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.fra59
    frame $site_4_0.fra75 \
        -height 75 -width 125 
    set site_5_0 $site_4_0.fra75
    button $site_5_0.but77 \
        \
        -command {BNF2SRGS $curobj(id) [string trim [$widget(c_memo) get 0.0 end]]} \
        -pady 0 -text button -textvariable label(opt,c_check) 
    label $site_5_0.cpd76 \
        -background #78ffd2 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text Kommentar -textvariable label(opt,c_memo) 
    vTcl:DefineAlias "$site_5_0.cpd76" "label,c_memo" vTcl:WidgetProc "property,c" 0
    pack $site_5_0.but77 \
        -in $site_5_0 -anchor center -expand 0 -fill none -side right 
    pack $site_5_0.cpd76 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    text $site_4_0.02 \
        -background #ffffff -foreground #000000 -height 1 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff -width 1 \
        -wrap word 
    vTcl:DefineAlias "$site_4_0.02" "c_memo" vTcl:WidgetProc "property,c" 0
    pack $site_4_0.fra75 \
        -in $site_4_0 -anchor center -expand 0 -fill x -side top 
    pack $site_4_0.02 \
        -in $site_4_0 -anchor center -expand 1 -fill both -side top 
    button $site_3_0.cpd79 \
        \
        -command {valCurobj2List
::vTcl:Toplevel:WidgetProc $widget(property,c) endmodal} \
        -pady 0 -text Anwenden -textvariable label(apply) 
    vTcl:DefineAlias "$site_3_0.cpd79" "Button1" vTcl:WidgetProc "property,c" 0
    button $site_3_0.cpd80 \
        \
        -command {if {![valCurobj2List 1] ||
     [tk_messageBox -message [res::Str leaveprop] -type yesno]=="yes"} {
    ::vTcl:Toplevel:WidgetProc $widget(property,c) endmodal
}} \
        -pady 0 -text Abbrechen -textvariable label(cancel) 
    vTcl:DefineAlias "$site_3_0.cpd80" "property,c,cancel" vTcl:WidgetProc "property,c" 0
    pack $site_3_0.01 \
        -in $site_3_0 -anchor center -expand 0 -fill x -side top 
    pack $site_3_0.fra47 \
        -in $site_3_0 -anchor center -expand 0 -fill x -side top 
    pack $site_3_0.fra59 \
        -in $site_3_0 -anchor center -expand 1 -fill both -side top 
    pack $site_3_0.cpd79 \
        -in $site_3_0 -anchor center -expand 1 -fill none -side left 
    pack $site_3_0.cpd80 \
        -in $site_3_0 -anchor center -expand 1 -fill none -side left 
    ###################
    # SETTING GEOMETRY
    ###################
    pack $top.cpd77 \
        -in $top -anchor center -expand 1 -fill both -side top 

    vTcl:FireEvent $base <<Ready>>
}

proc vTclWindow.top81 {base} {
    if {$base == ""} {
        set base .top81
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    set top $base
    ###################
    # CREATING WIDGETS
    ###################
    vTcl:toplevel $top -class Toplevel \
        -background #d9d9d9 -highlightbackground #d9d9d9 \
        -highlightcolor black 
    wm withdraw $top
    wm focusmodel $top passive
    wm geometry $top 744x528; update
    wm maxsize $top 1028 753
    wm minsize $top 104 1
    wm overrideredirect $top 0
    wm resizable $top 1 1
    wm title $top "Frageeigenschaften"
    vTcl:DefineAlias "$top" "property,q" vTcl:Toplevel:WidgetProc "" 0
    bindtags $top "$top Toplevel all _TopLevel"
    bind $top <Key-Escape> {
        $widget(property,q,cancel) invoke
    }
    vTcl:FireEvent $top <<Create>>
    wm protocol $top WM_DELETE_WINDOW "vTcl:FireEvent $top <<DeleteWindow>>"

    frame $top.cpd82 \
        -borderwidth 1 -relief sunken -background #dcdcdc -height 75 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -width 125 
    vTcl:DefineAlias "$top.cpd82" "qwindow" vTcl:WidgetProc "property,q" 0
    set site_3_0 $top.cpd82
    frame $site_3_0.01 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.01
    label $site_4_0.02 \
        -anchor w -background #dec8de -borderwidth 1 \
        -font {helvetica 24 bold} -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text Frage -textvariable label(opt,q_header) 
    entry $site_4_0.ent25 \
        -background #c69fc6 -font {helvetica 12 bold} -foreground #000000 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff \
        -textvariable curobj(q_title) 
    label $site_4_0.03 \
        -background #dec8de -borderwidth 1 -font {helvetica 24 bold} \
        -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -relief raised -text {Q 001} \
        -textvariable curobj(q_id) 
    pack $site_4_0.02 \
        -in $site_4_0 -anchor center -expand 1 -fill x -side left 
    pack $site_4_0.ent25 \
        -in $site_4_0 -anchor center -expand 0 -fill y -side left 
    pack $site_4_0.03 \
        -in $site_4_0 -anchor center -expand 0 -fill x -side left 
    frame $site_3_0.fra25 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.fra25
    frame $site_4_0.fra35 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_5_0 $site_4_0.fra35
    button $site_5_0.but37 \
        -background #ffa850 \
        -command {set curobj(q_concepts) [dia::chooseCmulti $curobj(q_concepts)]} \
        -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 -text {Verstandene Konzepte} \
        -textvariable label(opt,q_concepts) 
    vTcl:DefineAlias "$site_5_0.but37" "label,q_concepts" vTcl:WidgetProc "property,q" 0
    button $site_5_0.but38 \
        -background #ffa850 \
        -command {set curobj(q_follows) [dia::choosePQmulti $curobj(q_follows)]} \
        -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 -text Folgeaktionen \
        -textvariable label(opt,q_follows) 
    vTcl:DefineAlias "$site_5_0.but38" "label,q_follows" vTcl:WidgetProc "property,q" 0
    button $site_5_0.cpd73 \
        -background #ffa850 \
        -command {set curobj(q_follows) [dia::choosePQmulti $curobj(q_follows)]} \
        -foreground #000000 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -pady 0 -state disabled -text Folgeaktionen \
        -textvariable label(opt,q_predecessors) 
    vTcl:DefineAlias "$site_5_0.cpd73" "label,q_predecessors" vTcl:WidgetProc "property,q" 0
    pack $site_5_0.but37 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    pack $site_5_0.but38 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    pack $site_5_0.cpd73 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    frame $site_4_0.fra36 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_5_0 $site_4_0.fra36
    frame $site_5_0.cpd78 \
        -height 75 -width 125 
    set site_6_0 $site_5_0.cpd78
    entry $site_6_0.cpd79 \
        -background #dcdcdc -foreground #000000 -highlightbackground #ffffff \
        -highlightcolor #000000 -selectbackground #547098 \
        -selectforeground #ffffff -state disabled \
        -textvariable curobj(q_concepts) 
    checkbutton $site_6_0.cpd80 \
        -background #ffa850 -borderwidth 0 -text B-In \
        -textvariable label(endstate) -variable curobj(q_endstate) 
    checkbutton $site_6_0.cpd73 \
        -background #ffa850 -borderwidth 0 -text B-In \
        -textvariable label(reachable) -variable curobj(q_reachable) 
    pack $site_6_0.cpd79 \
        -in $site_6_0 -anchor center -expand 1 -fill x -side left 
    pack $site_6_0.cpd80 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side right 
    pack $site_6_0.cpd73 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side top 
    frame $site_5_0.fra19 \
        -height 75 -width 125 
    set site_6_0 $site_5_0.fra19
    entry $site_6_0.ent20 \
        -background #dcdcdc -foreground #000000 -highlightbackground #ffffff \
        -highlightcolor #000000 -selectbackground #547098 \
        -selectforeground #ffffff -state disabled \
        -textvariable curobj(q_follows) 
    pack $site_6_0.ent20 \
        -in $site_6_0 -anchor center -expand 1 -fill x -pady 1 -side left 
    entry $site_5_0.cpd72 \
        -background #dcdcdc -foreground #000000 -highlightbackground #ffffff \
        -highlightcolor #000000 -selectbackground #547098 \
        -selectforeground #ffffff -state disabled \
        -textvariable curobj(q_predecessors) 
    pack $site_5_0.cpd78 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    pack $site_5_0.fra19 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    pack $site_5_0.cpd72 \
        -in $site_5_0 -anchor center -expand 0 -fill x -pady 1 -side top 
    pack $site_4_0.fra35 \
        -in $site_4_0 -anchor center -expand 0 -fill none -side left 
    pack $site_4_0.fra36 \
        -in $site_4_0 -anchor center -expand 1 -fill x -side left 
    frame $site_3_0.04 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.04
    frame $site_4_0.fra18 \
        -height 75 -width 125 
    set site_5_0 $site_4_0.fra18
    checkbutton $site_5_0.cpd74 \
        -background #ffa850 -borderwidth 0 -text B-In \
        -textvariable label(barge_in) -variable curobj(q_init_bi) 
    label $site_5_0.lab19 \
        -background #ffa850 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text {Initiales Prompt} -textvariable label(opt,q_init) 
    vTcl:DefineAlias "$site_5_0.lab19" "label,q_init" vTcl:WidgetProc "property,q" 0
    button $site_5_0.but20 \
        -borderwidth 1 -highlightthickness 0 -padx 0 -pady 0 
    vTcl:DefineAlias "$site_5_0.but20" "ls,q_init" vTcl:WidgetProc "property,q" 0
    pack $site_5_0.cpd74 \
        -in $site_5_0 -anchor center -expand 0 -fill none -side left 
    pack $site_5_0.lab19 \
        -in $site_5_0 -anchor center -expand 1 -fill x -side left 
    pack $site_5_0.but20 \
        -in $site_5_0 -anchor center -expand 0 -fill none -side right 
    text $site_4_0.06 \
        -background #ffffff -foreground #000000 -height 1 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff -width 1 \
        -wrap word 
    vTcl:DefineAlias "$site_4_0.06" "q_init" vTcl:WidgetProc "property,q" 0
    pack $site_4_0.fra18 \
        -in $site_4_0 -anchor center -expand 0 -fill x -side top 
    pack $site_4_0.06 \
        -in $site_4_0 -anchor center -expand 1 -fill both -side top 
    frame $site_3_0.fra20 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.fra20
    frame $site_4_0.fra21 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_5_0 $site_4_0.fra21
    frame $site_5_0.fra24 \
        -height 75 -width 125 
    set site_6_0 $site_5_0.fra24
    checkbutton $site_6_0.cpd75 \
        -background #ffa850 -borderwidth 0 -text B-In \
        -textvariable label(barge_in) -variable curobj(q_nu1_bi) 
    label $site_6_0.lab33 \
        -background #ffa850 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text {NV 1} -textvariable label(opt,q_nu1) 
    vTcl:DefineAlias "$site_6_0.lab33" "label,q_nu1" vTcl:WidgetProc "property,q" 0
    button $site_6_0.but42 \
        -borderwidth 1 -highlightthickness 0 -padx 0 -pady 0 
    vTcl:DefineAlias "$site_6_0.but42" "ls,q_nu1" vTcl:WidgetProc "property,q" 0
    pack $site_6_0.cpd75 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side left 
    pack $site_6_0.lab33 \
        -in $site_6_0 -anchor center -expand 1 -fill x -side left 
    pack $site_6_0.but42 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side right 
    text $site_5_0.02 \
        -background #ffffff -foreground #000000 -height 1 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff -width 1 \
        -wrap word 
    vTcl:DefineAlias "$site_5_0.02" "q_nu1" vTcl:WidgetProc "property,q" 0
    pack $site_5_0.fra24 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    pack $site_5_0.02 \
        -in $site_5_0 -anchor center -expand 1 -fill both -side top 
    frame $site_4_0.fra22 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_5_0 $site_4_0.fra22
    frame $site_5_0.fra25 \
        -height 75 -width 125 
    set site_6_0 $site_5_0.fra25
    checkbutton $site_6_0.cpd76 \
        -background #ffa850 -borderwidth 0 -text B-In \
        -textvariable label(barge_in) -variable curobj(q_nu2_bi) 
    label $site_6_0.lab34 \
        -background #ffa850 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text {NV 2} -textvariable label(opt,q_nu2) 
    vTcl:DefineAlias "$site_6_0.lab34" "label,q_nu2" vTcl:WidgetProc "property,q" 0
    button $site_6_0.but45 \
        -borderwidth 1 -highlightthickness 0 -padx 0 -pady 0 
    vTcl:DefineAlias "$site_6_0.but45" "ls,q_nu2" vTcl:WidgetProc "property,q" 0
    pack $site_6_0.cpd76 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side left 
    pack $site_6_0.lab34 \
        -in $site_6_0 -anchor center -expand 1 -fill x -side left 
    pack $site_6_0.but45 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side right 
    text $site_5_0.02 \
        -background #ffffff -foreground #000000 -height 1 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff -width 1 \
        -wrap word 
    vTcl:DefineAlias "$site_5_0.02" "q_nu2" vTcl:WidgetProc "property,q" 0
    pack $site_5_0.fra25 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    pack $site_5_0.02 \
        -in $site_5_0 -anchor center -expand 1 -fill both -side top 
    frame $site_4_0.fra23 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_5_0 $site_4_0.fra23
    frame $site_5_0.fra26 \
        -height 75 -width 125 
    set site_6_0 $site_5_0.fra26
    checkbutton $site_6_0.cpd77 \
        -background #ffa850 -borderwidth 0 -text B-In \
        -textvariable label(barge_in) -variable curobj(q_nu3_bi) 
    label $site_6_0.lab35 \
        -background #ffa850 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text {NV 3} -textvariable label(opt,q_nu3) 
    vTcl:DefineAlias "$site_6_0.lab35" "label,q_nu3" vTcl:WidgetProc "property,q" 0
    button $site_6_0.but46 \
        -borderwidth 1 -highlightthickness 0 -padx 0 -pady 0 
    vTcl:DefineAlias "$site_6_0.but46" "ls,q_nu3" vTcl:WidgetProc "property,q" 0
    pack $site_6_0.cpd77 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side left 
    pack $site_6_0.lab35 \
        -in $site_6_0 -anchor center -expand 1 -fill x -side left 
    pack $site_6_0.but46 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side right 
    text $site_5_0.02 \
        -background #ffffff -foreground #000000 -height 1 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff -width 1 \
        -wrap word 
    vTcl:DefineAlias "$site_5_0.02" "q_nu3" vTcl:WidgetProc "property,q" 0
    pack $site_5_0.fra26 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    pack $site_5_0.02 \
        -in $site_5_0 -anchor center -expand 1 -fill both -side top 
    pack $site_4_0.fra21 \
        -in $site_4_0 -anchor center -expand 1 -fill both -side left 
    pack $site_4_0.fra22 \
        -in $site_4_0 -anchor center -expand 1 -fill both -side left 
    pack $site_4_0.fra23 \
        -in $site_4_0 -anchor center -expand 1 -fill both -side left 
    frame $site_3_0.fra24 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.fra24
    frame $site_4_0.01 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_5_0 $site_4_0.01
    frame $site_5_0.fra27 \
        -height 75 -width 125 
    set site_6_0 $site_5_0.fra27
    checkbutton $site_6_0.cpd78 \
        -background #ffa850 -borderwidth 0 -text B-In \
        -textvariable label(barge_in) -variable curobj(q_nr1_bi) 
    label $site_6_0.lab36 \
        -background #ffa850 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text {NA 1} -textvariable label(opt,q_nr1) 
    vTcl:DefineAlias "$site_6_0.lab36" "label,q_nr1" vTcl:WidgetProc "property,q" 0
    button $site_6_0.but47 \
        -borderwidth 1 -highlightthickness 0 -padx 0 -pady 0 
    vTcl:DefineAlias "$site_6_0.but47" "ls,q_nr1" vTcl:WidgetProc "property,q" 0
    pack $site_6_0.cpd78 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side left 
    pack $site_6_0.lab36 \
        -in $site_6_0 -anchor center -expand 1 -fill x -side left 
    pack $site_6_0.but47 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side right 
    text $site_5_0.03 \
        -background #ffffff -foreground #000000 -height 1 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff -width 1 \
        -wrap word 
    vTcl:DefineAlias "$site_5_0.03" "q_nr1" vTcl:WidgetProc "property,q" 0
    pack $site_5_0.fra27 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    pack $site_5_0.03 \
        -in $site_5_0 -anchor center -expand 1 -fill both -side top 
    frame $site_4_0.04 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_5_0 $site_4_0.04
    frame $site_5_0.fra28 \
        -height 75 -width 125 
    set site_6_0 $site_5_0.fra28
    checkbutton $site_6_0.cpd79 \
        -background #ffa850 -borderwidth 0 -text B-In \
        -textvariable label(barge_in) -variable curobj(q_nr2_bi) 
    label $site_6_0.lab37 \
        -background #ffa850 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text {NA 2} -textvariable label(opt,q_nr2) 
    vTcl:DefineAlias "$site_6_0.lab37" "label,q_nr2" vTcl:WidgetProc "property,q" 0
    button $site_6_0.but48 \
        -borderwidth 1 -highlightthickness 0 -padx 0 -pady 0 
    vTcl:DefineAlias "$site_6_0.but48" "ls,q_nr2" vTcl:WidgetProc "property,q" 0
    pack $site_6_0.cpd79 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side left 
    pack $site_6_0.lab37 \
        -in $site_6_0 -anchor center -expand 1 -fill x -side left 
    pack $site_6_0.but48 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side right 
    text $site_5_0.06 \
        -background #ffffff -foreground #000000 -height 1 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff -width 1 \
        -wrap word 
    vTcl:DefineAlias "$site_5_0.06" "q_nr2" vTcl:WidgetProc "property,q" 0
    pack $site_5_0.fra28 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    pack $site_5_0.06 \
        -in $site_5_0 -anchor center -expand 1 -fill both -side top 
    frame $site_4_0.07 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_5_0 $site_4_0.07
    frame $site_5_0.fra29 \
        -height 75 -width 125 
    set site_6_0 $site_5_0.fra29
    checkbutton $site_6_0.cpd80 \
        -background #ffa850 -borderwidth 0 -text B-In \
        -textvariable label(barge_in) -variable curobj(q_nr3_bi) 
    label $site_6_0.lab38 \
        -background #ffa850 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text {NA 3} -textvariable label(opt,q_nr3) 
    vTcl:DefineAlias "$site_6_0.lab38" "label,q_nr3" vTcl:WidgetProc "property,q" 0
    button $site_6_0.but49 \
        -borderwidth 1 -highlightthickness 0 -padx 0 -pady 0 
    vTcl:DefineAlias "$site_6_0.but49" "ls,q_nr3" vTcl:WidgetProc "property,q" 0
    pack $site_6_0.cpd80 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side left 
    pack $site_6_0.lab38 \
        -in $site_6_0 -anchor center -expand 1 -fill x -side left 
    pack $site_6_0.but49 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side right 
    text $site_5_0.09 \
        -background #ffffff -foreground #000000 -height 1 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff -width 1 \
        -wrap word 
    vTcl:DefineAlias "$site_5_0.09" "q_nr3" vTcl:WidgetProc "property,q" 0
    pack $site_5_0.fra29 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    pack $site_5_0.09 \
        -in $site_5_0 -anchor center -expand 1 -fill both -side top 
    pack $site_4_0.01 \
        -in $site_4_0 -anchor center -expand 1 -fill both -side left 
    pack $site_4_0.04 \
        -in $site_4_0 -anchor center -expand 1 -fill both -side left 
    pack $site_4_0.07 \
        -in $site_4_0 -anchor center -expand 1 -fill both -side left 
    frame $site_3_0.fra29 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.fra29
    frame $site_4_0.01 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_5_0 $site_4_0.01
    frame $site_5_0.fra30 \
        -height 75 -width 125 
    set site_6_0 $site_5_0.fra30
    checkbutton $site_6_0.cpd81 \
        -background #ffa850 -borderwidth 0 -text B-In \
        -textvariable label(barge_in) -variable curobj(q_tl1_bi) 
    label $site_6_0.lab39 \
        -background #ffa850 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text {ZL 1} -textvariable label(opt,q_tl1) 
    vTcl:DefineAlias "$site_6_0.lab39" "label,q_tl1" vTcl:WidgetProc "property,q" 0
    button $site_6_0.but50 \
        -borderwidth 1 -highlightthickness 0 -padx 0 -pady 0 
    vTcl:DefineAlias "$site_6_0.but50" "ls,q_tl1" vTcl:WidgetProc "property,q" 0
    pack $site_6_0.cpd81 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side left 
    pack $site_6_0.lab39 \
        -in $site_6_0 -anchor center -expand 1 -fill x -side left 
    pack $site_6_0.but50 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side right 
    text $site_5_0.03 \
        -background #ffffff -foreground #000000 -height 1 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff -width 1 \
        -wrap word 
    vTcl:DefineAlias "$site_5_0.03" "q_tl1" vTcl:WidgetProc "property,q" 0
    pack $site_5_0.fra30 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    pack $site_5_0.03 \
        -in $site_5_0 -anchor center -expand 1 -fill both -side top 
    frame $site_4_0.04 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_5_0 $site_4_0.04
    frame $site_5_0.fra31 \
        -height 75 -width 125 
    vTcl:DefineAlias "$site_5_0.fra31" "label,q_tl3" vTcl:WidgetProc "property,q" 0
    set site_6_0 $site_5_0.fra31
    checkbutton $site_6_0.cpd82 \
        -background #ffa850 -borderwidth 0 -text B-In \
        -textvariable label(barge_in) -variable curobj(q_tl2_bi) 
    label $site_6_0.lab40 \
        -background #ffa850 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text {ZL 2} -textvariable label(opt,q_tl2) 
    vTcl:DefineAlias "$site_6_0.lab40" "label,q_tl2" vTcl:WidgetProc "property,q" 0
    button $site_6_0.but51 \
        -borderwidth 1 -highlightthickness 0 -padx 0 -pady 0 
    vTcl:DefineAlias "$site_6_0.but51" "ls,q_tl2" vTcl:WidgetProc "property,q" 0
    pack $site_6_0.cpd82 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side left 
    pack $site_6_0.lab40 \
        -in $site_6_0 -anchor center -expand 1 -fill x -side left 
    pack $site_6_0.but51 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side right 
    text $site_5_0.06 \
        -background #ffffff -foreground #000000 -height 1 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff -width 1 \
        -wrap word 
    vTcl:DefineAlias "$site_5_0.06" "q_tl2" vTcl:WidgetProc "property,q" 0
    pack $site_5_0.fra31 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    pack $site_5_0.06 \
        -in $site_5_0 -anchor center -expand 1 -fill both -side top 
    frame $site_4_0.07 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_5_0 $site_4_0.07
    frame $site_5_0.fra32 \
        -height 75 -width 125 
    set site_6_0 $site_5_0.fra32
    checkbutton $site_6_0.cpd83 \
        -background #ffa850 -borderwidth 0 -text B-In \
        -textvariable label(barge_in) -variable curobj(q_tl3_bi) 
    button $site_6_0.but52 \
        -borderwidth 1 -highlightthickness 0 -padx 0 -pady 0 
    vTcl:DefineAlias "$site_6_0.but52" "ls,q_tl3" vTcl:WidgetProc "property,q" 0
    label $site_6_0.lab41 \
        -background #ffa850 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text {ZL 3} -textvariable label(opt,q_tl3) 
    vTcl:DefineAlias "$site_6_0.lab41" "label,q_tl4" vTcl:WidgetProc "property,q" 0
    pack $site_6_0.cpd83 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side left 
    pack $site_6_0.but52 \
        -in $site_6_0 -anchor center -expand 0 -fill none -side right 
    pack $site_6_0.lab41 \
        -in $site_6_0 -anchor center -expand 1 -fill x -side left 
    text $site_5_0.09 \
        -background #ffffff -foreground #000000 -height 1 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff -width 1 \
        -wrap word 
    vTcl:DefineAlias "$site_5_0.09" "q_tl3" vTcl:WidgetProc "property,q" 0
    pack $site_5_0.fra32 \
        -in $site_5_0 -anchor center -expand 0 -fill x -side top 
    pack $site_5_0.09 \
        -in $site_5_0 -anchor center -expand 1 -fill both -side top 
    pack $site_4_0.01 \
        -in $site_4_0 -anchor center -expand 1 -fill both -side left 
    pack $site_4_0.04 \
        -in $site_4_0 -anchor center -expand 1 -fill both -side left 
    pack $site_4_0.07 \
        -in $site_4_0 -anchor center -expand 1 -fill both -side left 
    frame $site_3_0.fra17 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.fra17
    frame $site_4_0.01 \
        -height 75 -width 125 
    set site_5_0 $site_4_0.01
    checkbutton $site_5_0.cpd84 \
        -background #ffa850 -borderwidth 0 -text B-In \
        -textvariable label(barge_in) -variable curobj(q_ctxhelp_bi) 
    label $site_5_0.02 \
        -background #ffa850 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text Kontexthilfe -textvariable label(opt,q_ctxhelp) 
    vTcl:DefineAlias "$site_5_0.02" "label,q_ctxhelp" vTcl:WidgetProc "property,q" 0
    button $site_5_0.03 \
        -borderwidth 1 -highlightthickness 0 -padx 0 -pady 0 
    vTcl:DefineAlias "$site_5_0.03" "ls,q_ctxhelp" vTcl:WidgetProc "property,q" 0
    pack $site_5_0.cpd84 \
        -in $site_5_0 -anchor center -expand 0 -fill none -side left 
    pack $site_5_0.02 \
        -in $site_5_0 -anchor center -expand 1 -fill x -side left 
    pack $site_5_0.03 \
        -in $site_5_0 -anchor center -expand 0 -fill none -side right 
    text $site_4_0.04 \
        -background #ffffff -foreground #000000 -height 1 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff -width 1 \
        -wrap word 
    vTcl:DefineAlias "$site_4_0.04" "q_ctxhelp" vTcl:WidgetProc "property,q" 0
    pack $site_4_0.01 \
        -in $site_4_0 -anchor center -expand 0 -fill x -side top 
    pack $site_4_0.04 \
        -in $site_4_0 -anchor center -expand 1 -fill both -side top 
    frame $site_3_0.fra18 \
        -background #dcdcdc -height 75 -highlightbackground #dcdcdc \
        -highlightcolor #000000 -width 125 
    set site_4_0 $site_3_0.fra18
    frame $site_4_0.01 \
        -height 75 -width 125 
    set site_5_0 $site_4_0.01
    label $site_5_0.02 \
        -background #ffa850 -borderwidth 1 -foreground #000000 \
        -highlightbackground #dcdcdc -highlightcolor #000000 -relief raised \
        -text Kommentar -textvariable label(opt,q_memo) 
    vTcl:DefineAlias "$site_5_0.02" "label,q_memo" vTcl:WidgetProc "property,q" 0
    checkbutton $site_5_0.cpd73 \
        -background #ffa850 -borderwidth 0 -textvariable label(vxml) \
        -variable curobj(q_vxml) 
    pack $site_5_0.02 \
        -in $site_5_0 -anchor center -expand 1 -fill x -side left 
    pack $site_5_0.cpd73 \
        -in $site_5_0 -anchor center -expand 0 -fill none -side top 
    text $site_4_0.03 \
        -background #ffffff -foreground #000000 -height 1 \
        -highlightbackground #ffffff -highlightcolor #000000 \
        -selectbackground #547098 -selectforeground #ffffff -width 1 \
        -wrap word 
    vTcl:DefineAlias "$site_4_0.03" "q_memo" vTcl:WidgetProc "property,q" 0
    pack $site_4_0.01 \
        -in $site_4_0 -anchor center -expand 0 -fill x -side top 
    pack $site_4_0.03 \
        -in $site_4_0 -anchor center -expand 1 -fill both -side top 
    button $site_3_0.cpd84 \
        \
        -command {valCurobj2List
::vTcl:Toplevel:WidgetProc $widget(property,q) endmodal} \
        -pady 0 -text Anwenden -textvariable label(apply) 
    vTcl:DefineAlias "$site_3_0.cpd84" "Button1" vTcl:WidgetProc "property,q" 0
    button $site_3_0.cpd85 \
        \
        -command {if {![valCurobj2List 1] ||
     [tk_messageBox -message [res::Str leaveprop] -type yesno]=="yes"} {
    ::vTcl:Toplevel:WidgetProc $widget(property,q) endmodal
}} \
        -pady 0 -text Abbrechen -textvariable label(cancel) 
    vTcl:DefineAlias "$site_3_0.cpd85" "property,q,cancel" vTcl:WidgetProc "property,q" 0
    pack $site_3_0.01 \
        -in $site_3_0 -anchor center -expand 0 -fill x -side top 
    pack $site_3_0.fra25 \
        -in $site_3_0 -anchor center -expand 0 -fill x -side top 
    pack $site_3_0.04 \
        -in $site_3_0 -anchor center -expand 1 -fill both -side top 
    pack $site_3_0.fra20 \
        -in $site_3_0 -anchor center -expand 1 -fill both -side top 
    pack $site_3_0.fra24 \
        -in $site_3_0 -anchor center -expand 1 -fill both -side top 
    pack $site_3_0.fra29 \
        -in $site_3_0 -anchor center -expand 1 -fill both -side top 
    pack $site_3_0.fra17 \
        -in $site_3_0 -anchor center -expand 1 -fill both -side top 
    pack $site_3_0.fra18 \
        -in $site_3_0 -anchor center -expand 1 -fill both -side top 
    pack $site_3_0.cpd84 \
        -in $site_3_0 -anchor center -expand 1 -fill none -side left 
    pack $site_3_0.cpd85 \
        -in $site_3_0 -anchor center -expand 1 -fill none -side left 
    ###################
    # SETTING GEOMETRY
    ###################
    pack $top.cpd82 \
        -in $top -anchor center -expand 1 -fill both -side top 

    vTcl:FireEvent $base <<Ready>>
}

#############################################################################
## Binding tag:  _TopLevel

bind "_TopLevel" <<Create>> {
    if {![info exists _topcount]} {set _topcount 0}; incr _topcount
}
bind "_TopLevel" <<DeleteWindow>> {
    if {[set ::%W::_modal]} {
                vTcl:Toplevel:WidgetProc %W endmodal
            } else {
                destroy %W; if {$_topcount == 0} {exit}
            }
}
bind "_TopLevel" <Destroy> {
    if {[winfo toplevel %W] == "%W"} {incr _topcount -1}
}
#############################################################################
## Binding tag:  <Key-Escape>


Window show .
Window show .top17
Window show .top30
Window show .top35
Window show .top71
Window show .top72
Window show .top73
Window show .top74
Window show .top75
Window show .top76
Window show .top81

main $argc $argv
